#!/usr/local/bin/ruby
# -
# $Id: install.rb,v 1.8.2.1 2003/04/11 14:08:51 toki Exp $

require 'getoptlong'

BIN_MODE = 0555
LIB_MODE = 0444

ruby_scripts = []
ruby_libraries = []

File::open(File::join(File::dirname($0), '.' + File::basename($0, '.rb'))) do |input|
  state = nil
  for line in input
    line.chomp!("\n")
    line.chomp!("\r")

    case (line)
    when /^\s*__SCRIPT__\s*$/
      state = :script
      next
    when /^\s*__LIBRARY__\s*$/
      state = :library
      next
    end

    state or next
    (line =~ /^\s*$/) and next

    src, dst = line.split(/\s+/, 2)
    dst = src unless dst
    src.gsub!(%r"/", File::SEPARATOR)
    dst.gsub!(%r"/", File::SEPARATOR)

    case (state)
    when :script
      ruby_scripts.push([ src, dst ])
    when :library
      ruby_libraries.push([ src, dst ])
    else
      raise "internal state error: #{state}"
    end
  end
end

options = {}
def options.method_missing(id, *args)
  case (args.length)
  when 0
    if (self.include? id) then
      return self[id]
    else
      raise "undefined option: #{id}"
    end
  when 1
    self[id.id2name.sub(/=$/, '').intern] = args.shift
  else
    raise 'wrong arguments: ' + args.map{ |arg| arg.inspect }.join(', ')
  end
end

options.debug = false
options.ruby_path =  nil
options.ruby_opts = nil
options.bin_dir = nil
options.lib_dir = nil
options.install_query = true

getoptlong = GetoptLong::new
getoptlong.set_options([ '-d', '--debug', GetoptLong::NO_ARGUMENT     ],
		       [ '-h', '--help',  GetoptLong::NO_ARGUMENT     ],
		       [ '--ruby-path', GetoptLong::REQUIRED_ARGUMENT ],
		       [ '--ruby-opts', GetoptLong::REQUIRED_ARGUMENT ],
		       [ '--bin-dir',   GetoptLong::REQUIRED_ARGUMENT ],
		       [ '--lib-dir',   GetoptLong::REQUIRED_ARGUMENT ])

help_msg = <<EOF
usage: ruby #{File::basename($0)} ([options])
options:
  -d, --debug  (not install)
  -h, --help
      --ruby-path=[ruby interpreter path]
      --ruby-opts=[ruby interpreter options]
      --bin-dir=[path to install ruby script]
      --lib-dir=[path to install ruby library]
EOF

for opt_name, opt_arg in getoptlong
  case (opt_name)
  when '-d', '--debug'
    options.debug = true
  when '-h', '--help'
    print help_msg
    exit 1
  when '--ruby-path'
    options.ruby_path = opt_arg
  when '--ruby-opts'
    options.ruby_opts = opt_arg
  when '--bin-dir'
    options.bin_dir = opt_arg
  when '--lib-dir'
    options.lib_dir = opt_arg
  end
end

STDOUT.sync = true

class List
  def initialize(title)
    @title = title
    @list = []
  end

  def push(item)
    @list.push(item)
  end

  def empty?
    @list.empty?
  end

  def include?(elem)
    @list.include? elem
  end

  def select
    loop do
      puts @title
      @list.each_with_index do |item, i|
	print i + 1, '. ', item, "\n"
      end

      case (@list.length)
      when 0
	print "not found.\n"
	print "input: "
      when 1
	print "number [1] or another input: "
      else
	print "number [1-#{@list.length}] or another input: "
      end

      if (answer = gets) then
	answer.chomp!("\n")
	answer.chomp!("\r")
	answer.strip!

	if (answer =~ /^[0-9]+$/) then
	  n = answer.to_i
	  if (0 < n && n <= @list.length) then
	    return @list[n - 1]
	  end
	elsif (! answer.empty?) then
	  return answer
	end
      else
	return nil
      end
    end
  end
end

if (options.ruby_path && options.bin_dir && options.lib_dir) then
  options.install_query = false
else
  print help_msg
end

path_separator = Regexp::compile(Regexp::quote(File::PATH_SEPARATOR))
path_reject = [
  /\./,
  /#{Regexp::quote(RUBY_PLATFORM)}/,
  /X11/,
  /WINDOWS/i,
  /WINNT/i,
  /SYSTEM(32)?/i
]

if (options.install_query && ! options.ruby_path) then
  list = List::new('**** select a path of ruby interpreter ****')
  for path in ENV['PATH'].split(path_separator)
    (File::directory? path) or next
    (path_reject.find{ |pattern| pattern =~ path }) and next
    Dir::foreach(path) do |filename|
      (filename =~ /^(ruby)(\.\S+)?$/) or next
      ruby_name = $1
      abs_path = File::expand_path(File::join(path, ruby_name))
      list.push(abs_path) unless (list.include? abs_path)
    end
  end
  list.empty? and raise "not found a ruby interpreter."
  options.ruby_path = list.select
end

if (options.install_query && ! options.ruby_opts) then
  puts '**** input options of ruby interpreter ****'
  print 'input ruby options (default: none): '
  ruby_opts = gets
  ruby_opts.chomp!("\n")
  ruby_opts.chomp!("\r")
  if (ruby_opts && ruby_opts !~ /^\s*$/) then
    options.ruby_opts = ruby_opts
  end
end

if (options.install_query && ! ruby_scripts.empty? && ! options.bin_dir) then
  list = List::new('**** select a path to install ruby scripts ****')
  for path in ENV['PATH'].split(path_separator)
    next if (path_reject.find{ |pattern| pattern =~ path })
    list.push(path) unless (list.include? path)
  end
  list.empty? and raise "not found a script path." 
  options.bin_dir = list.select
end

if (options.install_query && ! ruby_libraries.empty? && ! options.lib_dir) then
  list = List::new('**** select a path to install ruby libraries ****')
  for path in $:
    next if (path_reject.find{ |pattern| pattern =~ path })
    list.push(path) unless (list.include? path)
  end
  list.empty? and raise "not found a ruby library path."
  options.lib_dir = list.select
end

options.ruby_path = File::expand_path(options.ruby_path)
options.bin_dir   = File::expand_path(options.bin_dir)
options.lib_dir   = File::expand_path(options.lib_dir)

puts '**** settings ****'

puts "ruby path: #{options.ruby_path}"
puts "ruby options: #{options.ruby_opts}"
puts "script path: #{options.bin_dir}"
puts "library path: #{options.lib_dir}" 

if (options.install_query) then
  print "install? [Yes/No]: "
  unless ((answer = gets) && answer =~ /^Y/i) then
    exit 1
  end
end

puts '**** install ****'

module Install
  def mkdir(path)
    unless (File::directory? path) then
      mkdir(File::dirname(path))
      Dir::mkdir(path)
    end
  end
  module_function :mkdir

  def install(src, dst, mode, ruby_path)
    File::unlink(dst) if (File::exist? dst)
    mkdir(File::dirname(dst))
    File::open(src, 'r') do |input|
      File::open(dst, 'w') do |output|
	for line in input
	  if (input.lineno == 1 && line =~ /^\#!/) then
	    line.sub!(/^#!\S*ruby\S*/, '#!' + ruby_path)
	  end
	  output.print line
	end
	output.chmod(mode)
      end
    end
  end
  module_function :install
end

ruby_path_with_opts = options.ruby_path
ruby_path_with_opts += ' ' + options.ruby_opts if options.ruby_opts

[ [ ruby_scripts,   options.bin_dir, BIN_MODE ],
  [ ruby_libraries, options.lib_dir, LIB_MODE ]
].each do |install_list, install_dir, install_mode|
  for src, dst in install_list
    dst = File::join(install_dir, dst)
    print src, ' -> ', dst, "\n"
    Install::install(src, dst, install_mode, ruby_path_with_opts) unless options.debug
  end
end
