#!/usr/local/bin/ruby
# $Id: pseudo_io.rb,v 1.13.2.4 2003/04/11 14:08:55 toki Exp $
# -
# Pseudo I/O for WWW server test

require 'monitor'
require 'runit/testcase'
require 'runit/cui/testrunner'

module TestWWWsrv
  class PseudoIO
    include Enumerable
    include MonitorMixin

    def initialize
      super
      @data = nil
    end

    attr_reader :data

    def write(fragment)
      synchronize{
	if (@data) then
	  @data.concat(fragment)
	else
	  @data = fragment.dup
	end
      }

      fragment.length
    end

    def print(*args)
      write(args.join)
      nil
    end

    def printf(*args)
      write(format(*args))
      nil
    end

    def read(size=nil)
      synchronize{
	if (size) then
	  if (@data) then
	    if (@data.length > size) then
	      fragment = @data[0...size]
	      @data = @data[size..-1]
	      return fragment
	    else
	      fragment = @data
	      @data = nil
	      return fragment
	    end
	  else
	    return nil
	  end
	else
	  fragment = @data
	  @data = nil
	  return fragment
	end
      }
    end

    def gets(separator=$/)
      synchronize{
	if (@data) then
	  pattern = Regexp.compile('^.*?' + Regexp.quote(separator))
	  if (@data.sub!(pattern, '')) then
	    @data = nil if @data.empty?
	    return $&
	  else
	    fragment = @data
	    @data = nil
	    return fragment
	  end
	else
	  return nil
	end
      }
    end

    def each_line(separator=$/)
      while (line = gets(separator))
	yield(line)
      end

      nil
    end

    alias each each_line

    def flush
      self
    end

    def clear
      synchronize{
	@data = nil
      }
      self
    end

    def to_io
      self
    end

    def close
      nil
    end
  end
end

if ($0 == __FILE__) then
  module TestWWWsrv
    class TestPseudoIO < RUNIT::TestCase
      def setup
	@io = PseudoIO::new
      end

      def test_new
	assert_instance_of(PseudoIO, @io)
      end

      def test_write
	assert_nil(@io.data)
	@io.write("hello")
	@io.write(" ")
	@io.write("world")
	assert_equal("hello world", @io.data)
      end

      def test_print
	assert_nil(@io.data)
	@io.print("hello", " ", "world")
	assert_equal("hello world", @io.data)
      end

      def test_read
	assert_nil(@io.read)
	assert_nil(@io.read(256))

	@io.write('Hello world')
	assert_equal('Hello world', @io.read)
	@io.clear

	@io.write('Hello world')
	assert_equal('Hello ', @io.read(6))
	assert_equal('world', @io.read(256))
      end

      def test_gets
	@io.write("hello\n")
	@io.write("world\n")
	assert_equal("hello\n", @io.gets)
	assert_equal("world\n", @io.gets)
	assert_nil(@io.gets)
      end

      def test_clear
	@io.write("hello world\n")
	@io.clear
	assert_nil(@io.data)
      end
    end
  end

  if (ARGV.empty?) then
    TestWWWsrv::constants.map{ |name|
      TestWWWsrv::const_get(name)
    }.select{ |const|
      (const.kind_of? Class) && (const < RUNIT::TestCase)
    }.each do |test_case|
      RUNIT::CUI::TestRunner::run(test_case::suite)
    end
  else
    for class_name in ARGV
      test_case = TestWWWsrv.const_get("Test#{class_name}")
      RUNIT::CUI::TestRunner.run(test_case.suite)
    end
  end
end
