#!/usr/local/bin/ruby
# $Id: test_document.rb,v 1.24.2.1 2003/01/05 13:06:27 toki Exp $
# -
# Test of WWWsrv::Document class
# Test of WWWsrv::Page class
# Test of WWWsrv::DocumentFolder class
# Test of WWWsrv::DocumentPart class
# Test of WWWsrv::DocumentDispatcher class
# Test of WWWsrv::Redirect class
# Test of WWWsrv::AccessControl class
# Test of WWWsrv::PartialContent class

require 'docutils'
require 'pseudo_io'
require 'pseudo_doc'
require 'wwwsrv/logger'
require 'wwwsrv/request'
require 'wwwsrv/response'
require 'wwwsrv/document'
require 'runit/testcase'
require 'runit/cui/testrunner'

module TestWWWsrv
  class TestHTTPError < RUNIT::TestCase
    include WWWsrv::HTTPerror

    def test_http_error_with_status
      doc_status, err_args = WWWsrv::Document::catch_status{
	http_error(404)
      }
      assert_equal(:error, doc_status)
      status, message, header, version = err_args
      assert_equal(404, status)
      assert_nil(message)
      assert_instance_of(Hash, header)
      assert(header.empty?)
      assert_equal(WWWsrv::HTTP_1_1, version)

      doc_status, err_args = WWWsrv::Document::catch_status{
	WWWsrv::HTTPerror::http_error(404)
      }
      assert_equal(:error, doc_status)
      status, message, header, version = err_args
      assert_equal(404, status)
      assert_nil(message)
      assert_instance_of(Hash, header)
      assert(header.empty?)
      assert_equal(WWWsrv::HTTP_1_1, version)
    end

    def test_http_error_with_message
      doc_status, err_args = WWWsrv::Document::catch_status{
	http_error(404, 'Not Found')
      }
      assert_equal(:error, doc_status)
      status, message, header, version = err_args
      assert_equal(404, status)
      assert_equal('Not Found', message)
      assert_instance_of(Hash, header)
      assert(header.empty?)
      assert_equal(WWWsrv::HTTP_1_1, version)

      doc_status, err_args = WWWsrv::Document::catch_status{
	WWWsrv::HTTPerror::http_error(404, 'Not Found')
      }
      assert_equal(:error, doc_status)
      status, message, header, version = err_args
      assert_equal(404, status)
      assert_equal('Not Found', message)
      assert_instance_of(Hash, header)
      assert(header.empty?)
      assert_equal(WWWsrv::HTTP_1_1, version)
    end

    def test_http_error_with_header
      doc_status, err_args = WWWsrv::Document::catch_status{
	http_error(405, 'Method Not Allowed', 'Allow' => 'GET, HEAD')
      }
      assert_equal(:error, doc_status)
      status, message, header, version = err_args
      assert_equal(405, status)
      assert_equal('Method Not Allowed', message)
      assert_instance_of(Hash, header)
      assert_equal(1, header.length)
      assert_equal('GET, HEAD', header['Allow'])
      assert_equal(WWWsrv::HTTP_1_1, version)

      doc_status, err_args = WWWsrv::Document::catch_status{
	WWWsrv::HTTPerror::http_error(405, 'Method Not Allowed', 'Allow' => 'GET, HEAD')
      }
      assert_equal(:error, doc_status)
      status, message, header, version = err_args
      assert_equal(405, status)
      assert_equal('Method Not Allowed', message)
      assert_instance_of(Hash, header)
      assert_equal(1, header.length)
      assert_equal('GET, HEAD', header['Allow'])
      assert_equal(WWWsrv::HTTP_1_1, version)
    end

    def test_http_error_with_version
      doc_status, err_args = WWWsrv::Document::catch_status{
	http_error(405, 'Method Not Allowed', 'Allow' => 'GET, HEAD', 'X-HTTP-Version' => WWWsrv::HTTP_1_0)
      }
      assert_equal(:error, doc_status)
      status, message, header, version = err_args
      assert_equal(405, status)
      assert_equal('Method Not Allowed', message)
      assert_instance_of(Hash, header)
      assert_equal(1, header.length)
      assert_equal('GET, HEAD', header['Allow'])
      assert_equal(WWWsrv::HTTP_1_0, version)

      doc_status, err_args = WWWsrv::Document::catch_status{
	http_error(405, 'Method Not Allowed', 'Allow' => 'GET, HEAD', 'X-HTTP-Version' => 'HTTP/1.0')
      }
      assert_equal(:error, doc_status)
      status, message, header, version = err_args
      assert_equal(405, status)
      assert_equal('Method Not Allowed', message)
      assert_instance_of(Hash, header)
      assert_equal(1, header.length)
      assert_equal('GET, HEAD', header['Allow'])
      assert_equal(WWWsrv::HTTP_1_0, version)

      doc_status, err_args = WWWsrv::Document::catch_status{
	WWWsrv::HTTPerror::http_error(405, 'Method Not Allowed', 'Allow' => 'GET, HEAD', 'X-HTTP-Version' => WWWsrv::HTTP_1_0)
      }
      assert_equal(:error, doc_status)
      status, message, header, version = err_args
      assert_equal(405, status)
      assert_equal('Method Not Allowed', message)
      assert_instance_of(Hash, header)
      assert_equal(1, header.length)
      assert_equal('GET, HEAD', header['Allow'])
      assert_equal(WWWsrv::HTTP_1_0, version)

      doc_status, err_args = WWWsrv::Document::catch_status{
	WWWsrv::HTTPerror::http_error(405, 'Method Not Allowed', 'Allow' => 'GET, HEAD', 'X-HTTP-Version' => 'HTTP/1.0')
      }
      assert_equal(:error, doc_status)
      status, message, header, version = err_args
      assert_equal(405, status)
      assert_equal('Method Not Allowed', message)
      assert_instance_of(Hash, header)
      assert_equal(1, header.length)
      assert_equal('GET, HEAD', header['Allow'])
      assert_equal(WWWsrv::HTTP_1_0, version)
    end
  end

  class TestDocument < RUNIT::TestCase
    include DocumentUtils

    def setup
      @document = WWWsrv::Document::new
      @root = PseudoDocument::new('text/plain', [ "Hello world\n" ])
    end

    def test_new
      assert_instance_of(WWWsrv::Document, @document)
    end

    def test_respond
      srv_opt, log_output = make_srv_opt
      request, body_input = make_request('GET', '/')
      response = make_response
      respondent = try_to_respond(@document, srv_opt, '', request, response)
      assert_equal(:error, respondent[:doc_status])
      err_status, err_message = respondent[:err_args]
      assert_equal(404, err_status)
      assert_not_nil(err_message)
      assert(! err_message.empty?)
      assert_nil(respondent[:body])
      assert(respondent[:body_alist].empty?)
    end
  end

  class TestPage < RUNIT::TestCase
    include DocumentUtils

    def setup
      @content = "Hello world.\n"
      @type = 'text/html'
      @mtime = Time::now
      @page = WWWsrv::Page::new(@content, @type, @mtime)
    end

    def test_head
      srv_opt, log_output = make_srv_opt
      request, body_input = make_request('HEAD', '/')
      response = make_response
      respondent = try_to_respond(@page, srv_opt, '', request, response)
      assert_equal(:success, respondent[:doc_status])
      assert_equal(200, respondent[:response].status)
      assert_match(/^".+"$/, respondent[:response].header.etag)
      assert_equal(@type, respondent[:response].header.content_type)
      assert_equal(@content.length.to_s, respondent[:response].header.content_length)
      assert_nil(respondent[:body])
      assert(respondent[:body_alist].empty?)
      assert_equal(WWWsrv::HTMLconv::http_date(@mtime), respondent[:response].header.last_modified)
      assert_nil(respondent[:response].header.content_md5)

      page_with_md5 = WWWsrv::Page::new(@content, @type, @mtime, true)
      srv_opt, log_output = make_srv_opt
      request, body_input = make_request('HEAD', '/')
      response = make_response
      respondent = try_to_respond(page_with_md5, srv_opt, '', request, response)
      assert_equal(:success, respondent[:doc_status])
      assert_equal(200, respondent[:response].status)
      assert_match(/^".+"$/, respondent[:response].header.etag)
      assert_equal(@type, respondent[:response].header.content_type)
      assert_equal(@content.length.to_s, respondent[:response].header.content_length)
      assert_nil(respondent[:body])
      assert(respondent[:body_alist].empty?)
      assert_equal(WWWsrv::HTMLconv::http_date(@mtime), respondent[:response].header.last_modified)
      md5 = Digest::MD5::new
      md5.update(@content)
      assert_equal(md5.digest, respondent[:response].header.content_md5.unpack('m').first)
    end

    def test_get
      srv_opt, log_output = make_srv_opt
      request, body_input = make_request('GET', '/')
      response = make_response
      respondent = try_to_respond(@page, srv_opt, '', request, response)
      assert_equal(:success, respondent[:doc_status])
      assert_equal(200, respondent[:response].status)
      assert_match(/^".+"$/, respondent[:response].header.etag)
      assert_equal(@type, respondent[:response].header.content_type)
      assert_equal(@content.length.to_s, respondent[:response].header.content_length)
      assert_equal(@content, respondent[:body])
      assert(! respondent[:body_alist].empty?)
      assert_equal(WWWsrv::HTMLconv::http_date(@mtime), respondent[:response].header.last_modified)
      assert_nil(respondent[:response].header.content_md5)

      page_with_md5 = WWWsrv::Page::new(@content, @type, @mtime, true)
      srv_opt, log_output = make_srv_opt
      request, body_input = make_request('GET', '/')
      response = make_response
      respondent = try_to_respond(page_with_md5, srv_opt, '', request, response)
      assert_equal(:success, respondent[:doc_status])
      assert_equal(200, respondent[:response].status)
      assert_match(/^".+"$/, respondent[:response].header.etag)
      assert_equal(@type, respondent[:response].header.content_type)
      assert_equal(@content.length.to_s, respondent[:response].header.content_length)
      assert_equal(@content, respondent[:body])
      assert(! respondent[:body_alist].empty?)
      assert_equal(WWWsrv::HTMLconv::http_date(@mtime), respondent[:response].header.last_modified)
      md5 = Digest::MD5::new
      md5.update(@content)
      assert_equal(md5.digest, respondent[:response].header.content_md5.unpack('m').first)
    end
  end

  class TestDocumentFolder < RUNIT::TestCase
    def setup
      @folder = WWWsrv::DocumentFolder::new
      @mount_alist = [
	[ '/',                [ '/',     WWWsrv::Document::new, nil                       ] ],
	[ '/test',            [ '/test', WWWsrv::Document::new, nil                       ] ],
	[ '/test/index.html', [ '/test', WWWsrv::Document::new, { :page => 'index.html' } ] ],
	[ '/data',            [ '/data', WWWsrv::Document::new, nil                       ] ],
	[ '/data/index.html', [ '/data', WWWsrv::Document::new, { :page => 'index.html' } ] ],
	[ '/cgi/index.cgi',   [ '/cgi',  WWWsrv::Document::new, { :mask => /\.cgi$/     } ] ],
	[ '/ssi/index.shtml', [ '/ssi',  WWWsrv::Document::new, { :mask => /\.shtml/    } ] ]
      ]
    end

    def test_new
      assert_instance_of(WWWsrv::DocumentFolder, @folder)
    end

    def test_mount
      for path, (mount_point, document, mount_opts) in @mount_alist
	if (mount_opts) then
	  assert_equal(document, @folder.mount(document, mount_point, mount_opts))
	else
	  assert_equal(document, @folder.mount(document, mount_point))
	end
      end

      for path, (mount_point, document, mount_opts) in @mount_alist
	mount_pair = @folder.find(path)
	assert_equal(document, mount_pair.first)
	assert_equal(mount_point, mount_pair.last)
      end
    end

    def test_umount
      for path, (mount_point, document, mount_opts) in @mount_alist
	if (mount_opts) then
	  assert_equal(document, @folder.mount(document, mount_point, mount_opts))
	else
	  assert_equal(document, @folder.mount(document, mount_point))
	end
      end

      assert_nil(@folder.umount('/detarame'))

      for path, (mount_point, document, mount_opts) in @mount_alist
	if (mount_opts) then
	  assert_equal(document, @folder.umount(mount_point, mount_opts))
	else
	  assert_equal(document, @folder.umount(mount_point))
	end
      end
    end

    def test_find
      for path, (mount_point, document, mount_opts) in @mount_alist
	if (mount_opts) then
	  assert_equal(document, @folder.mount(document, mount_point, mount_opts))
	else
	  assert_equal(document, @folder.mount(document, mount_point))
	end
      end

      [ %w[ /                /                ],
	%w[ /                /index.html      ],
	%w[ /test            /test            ],
	%w[ /test            /test/test.html  ],
	%w[ /test/index.html /test/index.html ],
	%w[ /data            /data            ],
	%w[ /data            /data/test.html  ],
	%w[ /data/index.html /data/index.html ],
	%w[ /cgi/index.cgi   /cgi/test.cgi    ],
	%w[ /cgi/index.cgi   /cgi/index.cgi   ],
	%w[ /                /cgi/index.html  ],
	%w[ /ssi/index.shtml /ssi/test.shtml  ],
	%w[ /ssi/index.shtml /ssi/index.shtml ],
	%w[ /                /ssi/index.html  ]
      ].each do |prefix, page|
	assert_not_nil(mount_pair = @folder.find(page))
	assert_equal(@mount_alist.assoc(prefix)[1][1], mount_pair.first)
	assert_equal(@mount_alist.assoc(prefix)[1][0], mount_pair.last)
      end
    end
  end

  class TestDocumentPart < RUNIT::TestCase
    include DocumentUtils

    def setup
      @path = 'test'
      @name = 'index.html'
      @content = <<-EOF
<html>
<head>
<title>Test</title>
</head>
<body>
<h1>Test</h1>
<p>Hello world.</p>
</body>
</html>
      EOF

      @root = WWWsrv::DocumentFolder::new
      @root.mount(WWWsrv::Page::new(@content), '/' + @path, :page => @name)
      @document = WWWsrv::DocumentPart::new(@root, '/' + @path)
    end

    def test_new
      assert_instance_of(WWWsrv::DocumentPart, @document)
    end

    def test_respond
      prefix = '/' + @path
      srv_opt, log_output = make_srv_opt
      request, body_input = make_request('GET', [ prefix, @name ].join('/'))
      response = make_response
      respondent = try_to_respond(@document, srv_opt, prefix, request, response)
      assert_equal(:success, respondent[:doc_status])
      assert_instance_of(WWWsrv::Response, respondent[:response])
      assert_equal(WWWsrv::HTTP_1_1, respondent[:response].version)
      assert_equal(200, respondent[:response].status)
      assert_equal('OK', respondent[:response].reason)
      assert_equal(@content, respondent[:body])
      assert(! respondent[:body_alist].empty?)

      prefix = '/detarame'
      srv_opt, log_output = make_srv_opt
      request, body_input = make_request('GET', [ prefix, @name ].join('/'))
      response = make_response
      respondent = try_to_respond(@document, srv_opt, prefix, request, response)
      assert_equal(:success, respondent[:doc_status])
      assert_instance_of(WWWsrv::Response, respondent[:response])
      assert_equal(WWWsrv::HTTP_1_1, respondent[:response].version)
      assert_equal(200, respondent[:response].status)
      assert_equal('OK', respondent[:response].reason)
      assert_equal(@content, respondent[:body])
      assert(! respondent[:body_alist].empty?)
    end
  end

  WWWsrv::Header::define_field('X-Select')

  class TestDocumentDispatcher < RUNIT::TestCase
    include DocumentUtils

    def setup
      @dispatcher = WWWsrv::DocumentDispatcher::new(WWWsrv::Page::new('default', 'text/plain'))
      @dispatcher.add_cond(WWWsrv::Page::new('foo', 'text/plain')) { |srv_opt, prefix, request|
	request.header.x_select == 'foo'
      }
      @dispatcher.add_cond(WWWsrv::Page::new('bar', 'text/plain')) { |srv_opt, prefix, request|
	request.header.x_select == 'bar'
      }
    end

    def test_dispatch_default
      srv_opt, log_output = make_srv_opt
      request, body_input = make_request('GET', '/')
      response = make_response
      respondent = try_to_respond(@dispatcher, srv_opt, '', request, response)
      assert_equal('default', respondent[:body])
    end

    def test_dispatch_foo
      srv_opt, log_output = make_srv_opt
      request, body_input = make_request('GET', '/')
      request.header.x_select = 'foo'
      response = make_response
      respondent = try_to_respond(@dispatcher, srv_opt, '', request, response)
      assert_equal('foo', respondent[:body])
    end

    def test_dispatch_bar
      srv_opt, log_output = make_srv_opt
      request, body_input = make_request('GET', '/')
      request.header.x_select = 'bar'
      response = make_response
      respondent = try_to_respond(@dispatcher, srv_opt, '', request, response)
      assert_equal('bar', respondent[:body])
    end

    def test_not_found
      @dispatcher.default_document = nil
      srv_opt, log_output = make_srv_opt
      request, body_input = make_request('GET', '/')
      response = make_response
      respondent = try_to_respond(@dispatcher, srv_opt, '', request, response)
      assert_equal(:error, respondent[:doc_status])
      err_status, err_message = respondent[:err_args]
      assert_equal(404, err_status)
      assert_not_nil(err_message)
    end
  end

  class TestRedirect < RUNIT::TestCase
    include DocumentUtils

    def test_redirect_301
      redirect = WWWsrv::Redirect::new{ |srv_opt, prefix, request| '/foo' }
      srv_opt, log_output = make_srv_opt
      request, body_input = make_request('GET', '/')
      response = make_response
      respondent = try_to_respond(redirect, srv_opt, '', request, response)
      assert_equal(301, respondent[:response].status)
      assert_equal('http://localhost:80/foo', respondent[:response].header.location)
    end

    def test_redirect_307
      redirect = WWWsrv::Redirect::new(307) { |srv_opt, prefix, request| '/bar' }
      srv_opt, log_output = make_srv_opt
      request, body_input = make_request('GET', '/')
      response = make_response
      respondent = try_to_respond(redirect, srv_opt, '', request, response)
      assert_equal(307, respondent[:response].status)
      assert_equal('http://localhost:80/bar', respondent[:response].header.location)
    end
  end

  class TestAccessControl < RUNIT::TestCase
    include DocumentUtils

    def setup
      @root = WWWsrv::DocumentFolder::new
      @root.mount(WWWsrv::Page::new("Hello world.\n", 'text/plain'), '/', :page => '')
      @access = WWWsrv::AccessControl::new(@root)
    end

    def test_new
      assert_instance_of(WWWsrv::AccessControl, @access)
    end

    def test_allow
      @access.add_access_control(:allow) { |srv_opt, prefix, request|
	request.client.address == '127.0.0.1'
      }
      @access.add_access_control(:allow, &proc{ |srv_opt, prefix, request|
        request.client.address == '192.168.0.1'
      })

      srv_opt, log_output = make_srv_opt(WWWsrv::Logger::SYSTEM)
      request, body_input =
	make_request('GET', '/',
		     :client_name => 'localhost',
		     :client_address => '127.0.0.1')
      response = make_response
      respondent = try_to_respond(@access, srv_opt, '', request, response)
      assert_equal(:success, respondent[:doc_status])
      assert_equal(200, respondent[:response].status)
      assert_nil(log_output.data)

      srv_opt, log_output = make_srv_opt(WWWsrv::Logger::SYSTEM)
      request, body_input =
	make_request('GET', '/',
		     :client_name => 'private',
		     :client_address => '192.168.0.1')
      response = make_response
      respondent = try_to_respond(@access, srv_opt, '', request, response)
      assert_equal(:success, respondent[:doc_status])
      assert_equal(200, respondent[:response].status)
      assert_nil(log_output.data)
    end

    def test_allow_with_logging
      @access.add_access_control(:allow, true) { |srv_opt, prefix, request|
	request.client.address == '127.0.0.1'
      }
      @access.add_access_control(:allow, true, &proc{ |srv_opt, prefix, request|
        request.client.address == '192.168.0.1'
      })

      srv_opt, log_output = make_srv_opt(WWWsrv::Logger::SYSTEM)
      request, body_input =
	make_request('GET', '/',
		     :client_name => 'localhost',
		     :client_address => '127.0.0.1')
      response = make_response
      respondent = try_to_respond(@access, srv_opt, '', request, response)
      assert_equal(:success, respondent[:doc_status])
      assert_equal(200, respondent[:response].status)
      assert_not_nil(log_output.data)

      srv_opt, log_output = make_srv_opt(WWWsrv::Logger::SYSTEM)
      request, body_input =
	make_request('GET', '/',
		     :client_name => 'private',
		     :client_address => '192.168.0.1')
      response = make_response
      respondent = try_to_respond(@access, srv_opt, '', request, response)
      assert_equal(:success, respondent[:doc_status])
      assert_equal(200, respondent[:response].status)
      assert_not_nil(log_output.data)
    end

    def test_deny
      @access.add_access_control(:deny) { |srv_opt, prefix, request|
	request.client.address == '127.0.0.1'
      }
      @access.add_access_control(:deny, &proc{ |srv_opt, prefix, request|
        request.client.address == '192.168.0.1'
      })

      srv_opt, log_output = make_srv_opt(WWWsrv::Logger::SYSTEM)
      request, body_input =
	make_request('GET', '/',
		     :client_name => 'localhost',
		     :client_address => '127.0.0.1')
      response = make_response
      respondent = try_to_respond(@access, srv_opt, '', request, response)
      assert_equal(:error, respondent[:doc_status])
      err_status, message = respondent[:err_args]
      assert_equal(403, err_status)
      assert_nil(log_output.data)

      srv_opt, log_output = make_srv_opt(WWWsrv::Logger::SYSTEM)
      request, body_input =
	make_request('GET', '/',
		     :client_name => 'private',
		     :client_address => '192.168.0.1')
      response = make_response
      respondent = try_to_respond(@access, srv_opt, '', request, response)
      assert_equal(:error, respondent[:doc_status])
      err_status, message = respondent[:err_args]
      assert_equal(403, err_status)
      assert_nil(log_output.data)
    end

    def test_deny_with_logging
      @access.add_access_control(:deny, true) { |srv_opt, prefix, request|
	request.client.address == '127.0.0.1'
      }
      @access.add_access_control(:deny, true, &proc{ |srv_opt, prefix, request|
        request.client.address == '192.168.0.1'
      })

      srv_opt, log_output = make_srv_opt(WWWsrv::Logger::SYSTEM)
      request, body_input =
	make_request('GET', '/',
		     :client_name => 'localhost',
		     :client_address => '127.0.0.1')
      response = make_response
      respondent = try_to_respond(@access, srv_opt, '', request, response)
      assert_equal(:error, respondent[:doc_status])
      err_status, err_message = respondent[:err_args]
      assert_equal(403, err_status)
      assert_nil(respondent[:response])
      assert_nil(respondent[:body])
      assert(respondent[:body_alist].empty?)
      assert_not_nil(log_output.data)

      srv_opt, log_output = make_srv_opt(WWWsrv::Logger::SYSTEM)
      request, body_input =
	make_request('GET', '/',
		     :client_name => 'private',
		     :client_address => '192.168.0.1')
      response = make_response
      respondent = try_to_respond(@access, srv_opt, '', request, response)
      assert_equal(:error, respondent[:doc_status])
      err_status, err_message = respondent[:err_args]
      assert_equal(403, err_status)
      assert_nil(respondent[:response])
      assert_nil(respondent[:body])
      assert(respondent[:body_alist].empty?)
      assert_not_nil(log_output.data)
    end

    def test_allow_deny
      @access.add_access_control(:allow) { |srv_opt, prefix, request|
        request.client.address == '127.0.0.1'
      }
      @access.add_access_control(:deny, &proc{ |srv_opt, prefix, request|
        request.client.address == '192.168.0.1'
      })

      srv_opt, log_output = make_srv_opt(WWWsrv::Logger::SYSTEM)
      request, body_input =
	make_request('GET', '/',
		     :client_name => 'localhost',
		     :client_address => '127.0.0.1')
      response = make_response
      respondent = try_to_respond(@access, srv_opt, '', request, response)
      assert_equal(:success, respondent[:doc_status])
      assert_equal(200, respondent[:response].status)
      assert_nil(log_output.data)

      srv_opt, log_output = make_srv_opt(WWWsrv::Logger::SYSTEM)
      request, body_input =
	make_request('GET', '/',
		     :client_name => 'private',
		     :client_address => '192.168.0.1')
      response = make_response
      respondent = try_to_respond(@access, srv_opt, '', request, response)
      assert_equal(:error, respondent[:doc_status])
      err_status, message = respondent[:err_args]
      assert_equal(403, err_status)
      assert_nil(respondent[:response])
      assert_nil(respondent[:body])
      assert(respondent[:body_alist].empty?)
      assert_nil(log_output.data)
    end
  end

  class TestPartialContent < RUNIT::TestCase
    include DocumentUtils

    def setup
      @content = %w[
        01234
        56789
        ABCDE
      ]
      @content_length = @content.join.length
      @source = PseudoDocument::new('text/plain', @content)
      @source.set_content_length
      @partial = WWWsrv::PartialContent::new(@source)
    end

    def test_new
      assert_instance_of(WWWsrv::PartialContent, @partial)
    end

    def test_without_range
      srv_opt, log_output = make_srv_opt
      request, body_input = make_request('GET', '/')
      response = make_response
      respondent = try_to_respond(@partial, srv_opt, '', request, response)
      assert_equal(:success, respondent[:doc_status])
      assert_equal(200, respondent[:response].status)
      assert_equal('bytes', respondent[:response].header.accept_ranges)
      assert_nil(respondent[:response].header.content_range)
      assert_equal(@content.join, respondent[:body])
    end

    def test_without_range_and_content_length
      @source.unset_content_length
      srv_opt, log_output = make_srv_opt
      request, body_input = make_request('GET', '/')
      response = make_response
      respondent = try_to_respond(@partial, srv_opt, '', request, response)
      assert_equal(:success, respondent[:doc_status])
      assert_equal(200, respondent[:response].status)
      assert_nil(respondent[:response].header.accept_ranges)
      assert_nil(respondent[:response].header.content_range)
      assert_equal(@content.join, respondent[:body])
    end

    def test_range_not_200
      srv_opt, log_output = make_srv_opt
      request, body_input = make_request('GET', '/')
      response = make_response(201)
      respondent = try_to_respond(@partial, srv_opt, '', request, response)
      assert_equal(:success, respondent[:doc_status])
      assert_equal(201, respondent[:response].status)
      assert_equal('bytes', respondent[:response].header.accept_ranges)
      assert_nil(respondent[:response].header.content_range)
      assert_equal(@content.join, respondent[:body])
    end

    def test_range_not_200_without_content_length
      @source.unset_content_length
      srv_opt, log_output = make_srv_opt
      request, body_input = make_request('GET', '/')
      response = make_response(201)
      respondent = try_to_respond(@partial, srv_opt, '', request, response)
      assert_equal(:success, respondent[:doc_status])
      assert_equal(201, respondent[:response].status)
      assert_nil(respondent[:response].header.accept_ranges)
      assert_nil(respondent[:response].header.content_range)
      assert_equal(@content.join, respondent[:body])
    end

    def test_range
      # first fragment
      srv_opt, log_output = make_srv_opt
      request, body_input = make_request('GET', '/')
      request.header.range = 'bytes=0-4'
      response = make_response
      respondent = try_to_respond(@partial, srv_opt, '', request, response)
      assert_equal(:success, respondent[:doc_status])
      assert_equal(206, respondent[:response].status)
      assert_equal('bytes', respondent[:response].header.accept_ranges)
      assert_equal("bytes 0-4/#{@content_length}", respondent[:response].header.content_range)
      assert_equal('5', respondent[:response].header.content_length)
      assert_equal(@content[0], respondent[:body])

      # second fragment
      srv_opt, log_output = make_srv_opt
      request, body_input = make_request('GET', '/')
      request.header.range = 'bytes=5-9'
      response = make_response
      respondent = try_to_respond(@partial, srv_opt, '', request, response)
      assert_equal(:success, respondent[:doc_status])
      assert_equal(206, respondent[:response].status)
      assert_equal('bytes', respondent[:response].header.accept_ranges)
      assert_equal("bytes 5-9/#{@content_length}", respondent[:response].header.content_range)
      assert_equal('5', respondent[:response].header.content_length)
      assert_equal(@content[1], respondent[:body])

      # third fragment
      srv_opt, log_output = make_srv_opt
      request, body_input = make_request('GET', '/')
      request.header.range = 'bytes=10-14'
      response = make_response
      respondent = try_to_respond(@partial, srv_opt, '', request, response)
      assert_equal(:success, respondent[:doc_status])
      assert_equal(206, respondent[:response].status)
      assert_equal('bytes', respondent[:response].header.accept_ranges)
      assert_equal("bytes 10-14/#{@content_length}", respondent[:response].header.content_range)
      assert_equal('5', respondent[:response].header.content_length)
      assert_equal(@content[2], respondent[:body])

      # fragments over first to second
      srv_opt, log_output = make_srv_opt
      request, body_input = make_request('GET', '/')
      request.header.range = 'bytes=3-7'
      response = make_response
      respondent = try_to_respond(@partial, srv_opt, '', request, response)
      assert_equal(:success, respondent[:doc_status])
      assert_equal(206, respondent[:response].status)
      assert_equal('bytes', respondent[:response].header.accept_ranges)
      assert_equal("bytes 3-7/#{@content_length}", respondent[:response].header.content_range)
      assert_equal(@content.join[3..7], respondent[:body])

      # fragments over second to third
      srv_opt, log_output = make_srv_opt
      request, body_input = make_request('GET', '/')
      request.header.range = 'bytes=7-'
      response = make_response
      respondent = try_to_respond(@partial, srv_opt, '', request, response)
      assert_equal(:success, respondent[:doc_status])
      assert_equal(206, respondent[:response].status)
      assert_equal('bytes', respondent[:response].header.accept_ranges)
      assert_equal("bytes 7-14/#{@content_length}", respondent[:response].header.content_range)
      assert_equal('8', respondent[:response].header.content_length)
      assert_equal(@content.join[7..14], respondent[:body])

      # part of last fragment
      srv_opt, log_output = make_srv_opt
      request, body_input = make_request('GET', '/')
      request.header.range = 'bytes=-3'
      response = make_response
      respondent = try_to_respond(@partial, srv_opt, '', request, response)
      assert_equal(:success, respondent[:doc_status])
      assert_equal(206, respondent[:response].status)
      assert_equal('bytes', respondent[:response].header.accept_ranges)
      assert_equal("bytes 12-14/#{@content_length}", respondent[:response].header.content_range)
      assert_equal('3', respondent[:response].header.content_length)
      assert_equal(@content.join[12..14], respondent[:body])

      # overall of last part
      srv_opt, log_output = make_srv_opt
      request, body_input = make_request('GET', '/')
      request.header.range = 'bytes=-1024'
      response = make_response
      respondent = try_to_respond(@partial, srv_opt, '', request, response)
      assert_equal(:success, respondent[:doc_status])
      assert_equal(206, respondent[:response].status)
      assert_equal('bytes', respondent[:response].header.accept_ranges)
      assert_equal("bytes 0-14/#{@content_length}", respondent[:response].header.content_range)
      assert_equal('15', respondent[:response].header.content_length)
      assert_equal(@content.join, respondent[:body])
    end

    def test_range_without_content_length
      @source.unset_content_length

      # first fragment
      srv_opt, log_output = make_srv_opt
      request, body_input = make_request('GET', '/')
      request.header.range = 'bytes=0-4'
      response = make_response
      respondent = try_to_respond(@partial, srv_opt, '', request, response)
      assert_equal(:success, respondent[:doc_status])
      assert_equal(200, respondent[:response].status)
      assert_nil(respondent[:response].header.accept_ranges)
      assert_nil(respondent[:response].header.content_range)
      assert_equal(@content.join, respondent[:body])

      # second fragment
      srv_opt, log_output = make_srv_opt
      request, body_input = make_request('GET', '/')
      request.header.range = 'bytes=5-9'
      response = make_response
      respondent = try_to_respond(@partial, srv_opt, '', request, response)
      assert_equal(:success, respondent[:doc_status])
      assert_equal(200, respondent[:response].status)
      assert_nil(respondent[:response].header.accept_ranges)
      assert_nil(respondent[:response].header.content_range)
      assert_equal(@content.join, respondent[:body])

      # third fragment
      srv_opt, log_output = make_srv_opt
      request, body_input = make_request('GET', '/')
      request.header.range = 'bytes=10-14'
      response = make_response
      respondent = try_to_respond(@partial, srv_opt, '', request, response)
      assert_equal(:success, respondent[:doc_status])
      assert_equal(200, respondent[:response].status)
      assert_nil(respondent[:response].header.accept_ranges)
      assert_nil(respondent[:response].header.content_range)
      assert_equal(@content.join, respondent[:body])

      # fragments over first to second
      srv_opt, log_output = make_srv_opt
      request, body_input = make_request('GET', '/')
      request.header.range = 'bytes=3-7'
      response = make_response
      respondent = try_to_respond(@partial, srv_opt, '', request, response)
      assert_equal(:success, respondent[:doc_status])
      assert_equal(200, respondent[:response].status)
      assert_nil(respondent[:response].header.accept_ranges)
      assert_nil(respondent[:response].header.content_range)
      assert_equal(@content.join, respondent[:body])

      # fragments over second to third
      srv_opt, log_output = make_srv_opt
      request, body_input = make_request('GET', '/')
      request.header.range = 'bytes=7-'
      response = make_response
      respondent = try_to_respond(@partial, srv_opt, '', request, response)
      assert_equal(:success, respondent[:doc_status])
      assert_equal(200, respondent[:response].status)
      assert_nil(respondent[:response].header.accept_ranges)
      assert_nil(respondent[:response].header.content_range)
      assert_equal(@content.join, respondent[:body])

      # part of last fragment
      srv_opt, log_output = make_srv_opt
      request, body_input = make_request('GET', '/')
      request.header.range = 'bytes=-3'
      response = make_response
      respondent = try_to_respond(@partial, srv_opt, '', request, response)
      assert_equal(:success, respondent[:doc_status])
      assert_equal(200, respondent[:response].status)
      assert_nil(respondent[:response].header.accept_ranges)
      assert_nil(respondent[:response].header.content_range)
      assert_equal(@content.join, respondent[:body])

      # overall of last part
      srv_opt, log_output = make_srv_opt
      request, body_input = make_request('GET', '/')
      request.header.range = 'bytes=-1024'
      response = make_response
      respondent = try_to_respond(@partial, srv_opt, '', request, response)
      assert_equal(:success, respondent[:doc_status])
      assert_nil(respondent[:response].header.accept_ranges)
      assert_nil(respondent[:response].header.content_range)
      assert_equal(@content.join, respondent[:body])
    end

    def test_out_of_range
      srv_opt, log_output = make_srv_opt
      request, body_input = make_request('GET', '/')
      request.header.range = 'bytes=1024-'
      response = make_response
      respondent = try_to_respond(@partial, srv_opt, '', request, response)
      assert_equal(:error, respondent[:doc_status])
      err_status, err_message = respondent[:err_args]
      assert_equal(416, err_status)

      srv_opt, log_output = make_srv_opt
      request, body_input = make_request('GET', '/')
      request.header.range = 'bytes=10-3'
      response = make_response
      respondent = try_to_respond(@partial, srv_opt, '', request, response)
      assert_equal(:error, respondent[:doc_status])
      err_status, err_message = respondent[:err_args]
      assert_equal(416, err_status)
    end

    def test_out_of_range_without_content_length
      @source.unset_content_length

      srv_opt, log_output = make_srv_opt
      request, body_input = make_request('GET', '/')
      request.header.range = 'bytes=1024-'
      response = make_response
      respondent = try_to_respond(@partial, srv_opt, '', request, response)
      assert_equal(:success, respondent[:doc_status])
      assert_equal(200, respondent[:response].status)
      assert_nil(respondent[:response].header.accept_ranges)
      assert_nil(respondent[:response].header.content_range)
      assert_equal(@content.join, respondent[:body])

      srv_opt, log_output = make_srv_opt
      request, body_input = make_request('GET', '/')
      request.header.range = 'bytes=10-3'
      response = make_response
      respondent = try_to_respond(@partial, srv_opt, '', request, response)
      assert_equal(:success, respondent[:doc_status])
      assert_equal(200, respondent[:response].status)
      assert_nil(respondent[:response].header.accept_ranges)
      assert_nil(respondent[:response].header.content_range)
      assert_equal(@content.join, respondent[:body])
    end
  end
end

if ($0 == __FILE__) then
  if (ARGV.empty?) then
    TestWWWsrv::constants.map{ |name|
      TestWWWsrv::const_get(name)
    }.select{ |const|
      (const.kind_of? Class) && (const < RUNIT::TestCase)
    }.each do |test_case|
      RUNIT::CUI::TestRunner::run(test_case::suite)
    end
  else
    for class_name in ARGV
      test_case = TestWWWsrv.const_get("Test#{class_name}")
      RUNIT::CUI::TestRunner.run(test_case.suite)
    end
  end
end
