#!/usr/local/bin/ruby
# $Id: test_erbdoc.rb,v 1.17.2.1 2003/01/05 13:06:28 toki Exp $
# -
# Test of eRuby document for WWW server

require 'docutils'
require 'pseudo_io'
require 'wwwsrv/htmlconv'
require 'wwwsrv/srvopt'
require 'wwwsrv/request'
require 'wwwsrv/response'
require 'wwwsrv/reader'
require 'wwwsrv/logger'
require 'wwwsrv/erbdoc'
require 'runit/testcase'
require 'runit/cui/testrunner'

module TestWWWsrv
  class TestERubyDocument < RUNIT::TestCase
    include DocumentUtils

    def setup
      @path = '/'
      @name = 'index.html'

      @local_dir = 'test'
      @local_path = File::join(@local_dir, @name)
      Dir::mkdir(@local_dir)
      File::open(@local_path, 'w') do |output|
	output.write('')
      end
      @document = WWWsrv::ERubyDocument::new(@local_dir)
    end

    def teardown
      File::unlink(@local_path)
      Dir::rmdir(@local_dir)
    end

    def test_new
      assert_instance_of(WWWsrv::ERubyDocument, @document)
    end

    WWWsrv::Header::define_field('X-Test-eRuby')

    def test_respond
      File::open(@local_path, 'w') do |output|
	output.write(<<-EOF)
<% include RUNIT::Assert
%><%# assert_equal('1', SID) # obsolete feature!
%><%# assert_equal('1', __sid__) # obsolete feature!
%><%# assert_equal(SID, __sid__) # obsolete feature!
%><% assert_instance_of(WWWsrv::ServerOption, SRV_OPT)
%><% assert_instance_of(WWWsrv::ServerOption, __srv_opt__)
%><% assert_equal(SRV_OPT, __srv_opt__)
%><% assert_instance_of(WWWsrv::Request, REQUEST)
%><% assert_instance_of(WWWsrv::Request, __request__)
%><% assert_equal(REQUEST, __request__)
%><% assert_instance_of(WWWsrv::Response, RESPONSE)
%><% assert_instance_of(WWWsrv::Response, __response__)
%><% assert_equal(RESPONSE, __response__)
%><% assert_equal('text/html', RESPONSE.header.content_type)
%><% RESPONSE.header.content_type = 'text/plain'
%><% __response__.header.x_test_eruby = 'foo'
%><%= 'SnVzdCBhbm90aGVyIHJ1YnkgaGFja2VyLg=='.unpack('m').first %>
	EOF
      end

      srv_opt, log_output = make_srv_opt
      request, body_input = make_request('GET', @path + @name)
      response = make_response
      respondent = try_to_respond(@document, srv_opt, '', request, response)
      assert_equal(:success, respondent[:doc_status])
      assert_instance_of(WWWsrv::Response, respondent[:response])
      assert_equal(200, respondent[:response].status)
      assert_equal('foo', respondent[:response].header.x_test_eruby)
      assert_equal('text/plain', respondent[:response].header.content_type)
      assert_equal(respondent[:body].length.to_s, respondent[:response].header.content_length)
      assert_equal("Just another ruby hacker.\n", respondent[:body])
      assert(! respondent[:body_alist].empty?)
    end

    def test_respond_with_not_allowed_method
      File::open(@local_path, 'w') do |output|
	output.write(<<-EOF)
<%= 'SnVzdCBhbm90aGVyIHJ1YnkgaGFja2VyLg=='.unpack('m').first %>
	EOF
      end

      srv_opt, log_output = make_srv_opt
      request, body_input = make_request('PUT', @path + @name)
      response = make_response
      respondent = try_to_respond(@document, srv_opt, '', request, response)
      assert_equal(:error, respondent[:doc_status])
      err_status, err_message, err_header, err_version = respondent[:err_args]
      assert_equal(405, err_status)
      assert(! (err_header['Allow'].include? 'PUT'))
    end
  end

  class TestERubyFilterDocument < RUNIT::TestCase
    include DocumentUtils

    def setup
      @path = '/'
      @name = 'index.html'
      @root = WWWsrv::DocumentFolder::new
      @document = WWWsrv::ERubyFilterDocument::new(@root)
    end

    def test_new
      assert_instance_of(WWWsrv::ERubyFilterDocument, @document)
    end

    def test_respond
      @root.mount(WWWsrv::Page::new(<<-EOF), @path, :page => @name)
<% include RUNIT::Assert
%><%# assert_equal('1', SID) # obsolete feature!
%><%# assert_equal('1', __sid__) # obsolete feature!
%><%# assert_equal(SID, __sid__) # obsolete feature!
%><% assert_instance_of(WWWsrv::ServerOption, SRV_OPT)
%><% assert_instance_of(WWWsrv::ServerOption, __srv_opt__)
%><% assert_equal(SRV_OPT, __srv_opt__)
%><% assert_instance_of(WWWsrv::Request, REQUEST)
%><% assert_instance_of(WWWsrv::Request, __request__)
%><% assert_equal(REQUEST, __request__)
%><% assert_instance_of(WWWsrv::Response, RESPONSE)
%><% assert_instance_of(WWWsrv::Response, __response__)
%><% assert_equal(RESPONSE, __response__)
%><% assert_equal('text/html', RESPONSE.header.content_type)
%><% RESPONSE.header.etag = 'detarame'           # disabled header field
%><% RESPONSE.header.last_modified = 'detarame'  # disabled header field
%><% RESPONSE.header.content_md5 = 'detarame'    # disabled header field
%><% RESPONSE.header.content_type = 'text/plain'
%><% __response__.header.x_test_eruby = 'foo'
%><%= 'SnVzdCBhbm90aGVyIHJ1YnkgaGFja2VyLg=='.unpack('m').first %>
      EOF

      srv_opt, log_output = make_srv_opt
      request, body_input = make_request('GET', @path + @name)
      response = make_response
      respondent = try_to_respond(@document, srv_opt, '', request, response)
      assert_equal(:success, respondent[:doc_status])
      assert_instance_of(WWWsrv::Response, respondent[:response])
      assert_equal(200, respondent[:response].status)
      assert_nil(respondent[:response].header.etag)
      assert_nil(respondent[:response].header.last_modified)
      assert_nil(respondent[:response].header.content_md5)
      assert_equal('foo', respondent[:response].header.x_test_eruby)
      assert_equal('text/plain', respondent[:response].header.content_type)
      assert_equal(respondent[:body].length.to_s, respondent[:response].header.content_length)
      assert_equal("Just another ruby hacker.\n", respondent[:body])
      assert(! respondent[:body_alist].empty?)
    end
  end
end

if ($0 == __FILE__) then
  if (ARGV.empty?) then
    TestWWWsrv::constants.map{ |name|
      TestWWWsrv::const_get(name)
    }.select{ |const|
      (const.kind_of? Class) && (const < RUNIT::TestCase)
    }.each do |test_case|
      RUNIT::CUI::TestRunner::run(test_case::suite)
    end
  else
    for class_name in ARGV
      test_case = TestWWWsrv.const_get("Test#{class_name}")
      RUNIT::CUI::TestRunner.run(test_case.suite)
    end
  end
end
