#!/usr/local/bin/ruby
# $Id: test_header.rb,v 1.9.2.2 2003/01/05 13:06:34 toki Exp $
# -
# Test of WWWsrv::Header class

require 'pseudo_io'
require 'wwwsrv/header'
require 'runit/testcase'
require 'runit/cui/testrunner'

module TestWWWsrv
  class TestHeader < RUNIT::TestCase
    def setup
      @header = WWWsrv::Header::new
    end

    def test_normalize
      [ %w[ Date           Date ],
	%w[ Server         SERVER ],
	%w[ Content-Type   content-type ],
	%w[ Content-Length CONTENT-length ],
	%w[ WWW-ETag-MD5-TE www-etag-md5-te ]
      ].each do |regular_name, field_name|
	assert_equal(regular_name, WWWsrv::Header::normalize(field_name))
      end
    end

    def test_new
      assert_instance_of(WWWsrv::Header, @header)
    end

    def test_push
      assert_nil(@header['Test-Head'])
      @header.push('Test-Head', 'Hello world')
      assert_equal('Hello world', @header['Test-Head'])
      @header.push('Test-Head', 'Hello Ruby')
      assert_equal('Hello world, Hello Ruby', @header['Test-Head'])
    end

    def test_pop
      assert_nil(@header.pop('Test-Head'))
      @header.push('Test-Head', 'Hello world')
      assert_equal('Hello world', @header.pop('Test-Head'))
    end

    def test_accessor
      assert_nil(@header['content-type'])
      assert_nil(@header['CONTENT-TYPE'])
      assert_nil(@header['Content-Type'])
      assert_nil(@header.content_type)
      @header['Content-Type'] = 'hello'
      assert_equal('hello', @header['content-type'])
      assert_equal('hello', @header['CONTENT-TYPE'])
      assert_equal('hello', @header['Content-Type'])
      assert_equal('hello', @header.content_type)
      @header.content_type = 'world'
      assert_equal('world', @header['content-type'])
      assert_equal('world', @header['CONTENT-TYPE'])
      assert_equal('world', @header['Content-Type'])
      assert_equal('world', @header.content_type)
    end

    def test_each
      header_alist = [
	[ 'Date', 'Date: Mon, 19 Nov 1975 00:00:00 GMT' ],
	[ 'Server', 'Test (with RUNIT)' ],
	[ 'Content-Type', 'text/html' ],
	[ 'Content-Length', '256' ]
      ]

      for field_name, field_value in header_alist
	@header[field_name] = field_value
      end

      @header.each_with_index do |(field_name, field_value), i|
	assert_equal(header_alist[i].first, field_name)
	assert_equal(header_alist[i].last, field_value)
      end
    end

    def test_multi_field
      @header.push('Content-Encoding', 'deflate')
      assert_equal('deflate', @header.content_encoding)
      @header.push('Content-Encoding', 'gzip')
      assert_equal('deflate, gzip', @header.content_encoding)
      @header.push('Content-Encoding', 'compress')
      assert_equal('deflate, gzip, compress', @header.content_encoding)

      for field_name, field_value in @header
	if (field_name == 'Content-Encoding') then
	  assert_equal('deflate, gzip, compress', field_value)
	end
      end
      assert_equal(1, @header.length)

      @header.set_multi_field('Content-Encoding')
      assert_equal('deflate, gzip, compress', @header.content_encoding)
      field_values = %w[ deflate gzip compress ]
      for field_name, field_value in @header
	if (field_name == 'Content-Encoding') then
	  assert_equal(field_values.shift, field_value)
	end
      end
      assert_equal(3, @header.length)

      @header.unset_multi_field('Content-Encoding')
      assert_equal('deflate, gzip, compress', @header.content_encoding)
      for field_name, field_value in @header
	if (field_name == 'Content-Encoding') then
	  assert_equal('deflate, gzip, compress', field_value)
	end
      end
      assert_equal(1, @header.length)
    end

    def test_fields
      @header.push('Content-Encoding', 'deflate')
      assert_equal(%w[ deflate ], @header.fields('Content-Encoding'))
      @header.push('Content-Encoding', 'gzip')
      assert_equal(%w[ deflate gzip ], @header.fields('Content-Encoding'))
      @header.push('Content-Encoding', 'compress')
      assert_equal(%w[ deflate gzip compress ], @header.fields('Content-Encoding'))
      @header.content_encoding = 'deflate, gzip, compress'
      assert_equal([ 'deflate, gzip, compress' ], @header.fields('Content-Encoding'))
    end

    def test_include
      @header.push('Content-Type', 'text/html')
      @header.push('Content-Length', '256')
      @header.push('Content-Encoding', 'deflate')
      @header.push('Content-Encoding', 'gzip')

      assert((@header.include? 'Content-Type'))
      assert((@header.include? 'Content-Length'))
      assert((@header.include? 'Content-Encoding'))

      @header.set_multi_field('Content-Encoding')
      assert((@header.include? 'Content-Encoding'))

      assert(! (@header.include? 'Accept-Encoding'))
    end

    def test_delete
      @header.push('Content-Type', 'text/html')
      @header.push('Content-Length', '256')
      @header.push('Content-Encoding', 'deflate')
      @header.push('Content-Encoding', 'gzip')
      assert_equal('text/html', @header.content_type)
      assert_equal('256', @header.content_length)
      assert_equal('deflate, gzip', @header.content_encoding)
      assert_equal(3, @header.length)

      @header.delete('Content-Length')
      @header.delete('Content-Encoding')
      assert_equal('text/html', @header.content_type)
      assert_nil(@header.content_length)
      assert_nil(@header.content_encoding)
      assert((@header.include? 'Content-Type'))
      assert(! (@header.include? 'Content-Length'))
      assert(! (@header.include? 'Content-Encoding'))
      assert_equal(1, @header.length)

      @header.push('Content-Length', '256')
      @header.push('Content-Encoding', 'deflate')
      @header.push('Content-Encoding', 'gzip')
      @header.set_multi_field('Content-Encoding')
      assert_equal('text/html', @header.content_type)
      assert_equal('256', @header.content_length)
      assert_equal('deflate, gzip', @header.content_encoding)
      assert_equal(4, @header.length)

      @header.delete('Content-Encoding')
      assert_equal('text/html', @header.content_type)
      assert_equal('256', @header.content_length)
      assert_nil(@header.content_encoding)
      assert((@header.include? 'Content-Type'))
      assert((@header.include? 'Content-Length'))
      assert(! (@header.include? 'Content-Encoding'))
      assert_equal(2, @header.length)
    end

    def test_delete_if
      @header.push('Content-Type', 'text/html')
      @header.push('Content-Length', '256')
      @header.push('Content-Encoding', 'deflate')
      @header.push('Content-Encoding', 'gzip')
      @header.set_multi_field('Content-Encoding')

      assert_equal('text/html', @header.content_type)
      assert_equal('256', @header.content_length)
      assert_equal('deflate, gzip', @header.content_encoding)
      assert_equal(4, @header.length)

      @header.delete_if{ |field_name, field_value|
	case (field_name)
	when 'Content-Length'
	  true
	when 'Content-Encoding'
	  field_value == 'gzip'
	else
	  false
	end
      }

      assert_equal('text/html', @header.content_type)
      assert_nil(@header.content_length)
      assert_equal('deflate', @header.content_encoding)
      assert((@header.include? 'Content-Type'))
      assert(! (@header.include? 'Content-Length'))
      assert((@header.include? 'Content-Encoding'))
      assert_equal(2, @header.length)

      @header.push('Content-Encoding', 'compress')
      assert_equal('text/html', @header.content_type)
      assert_equal('deflate, compress', @header.content_encoding)
      assert((@header.include? 'Content-Type'))
      assert((@header.include? 'Content-Encoding'))
      assert_equal(3, @header.length)

      @header.delete_if{ |field_name, field_value|
	if (field_name == 'Content-Encoding') then
	  case (field_value)
	  when 'deflate'
	    true
	  when 'compress'
	    true
	  else
	    false
	  end
	end
      }

      assert_equal('text/html', @header.content_type)
      assert_nil(@header.content_encoding)
      assert((@header.include? 'Content-Type'))
      assert(! (@header.include? 'Content-Encoding'))
      assert_equal(1, @header.length)
    end

    def test_clone
      @header.push('Content-Type', 'text/html')
      @header.push('Content-Length', '256')
      @header.push('Content-Encoding', 'deflate')
      @header.push('Content-Encoding', 'gzip')

      assert_equal('text/html', @header.content_type)
      assert_equal('256', @header.content_length)
      assert_equal('deflate, gzip', @header.content_encoding)
      assert_equal(3, @header.length)

      copy_header = @header.clone
      assert_equal('text/html', copy_header.content_type)
      assert_equal('256', copy_header.content_length)
      assert_equal('deflate, gzip', copy_header.content_encoding)
      assert_equal(3, copy_header.length)

      @header.set_multi_field('Content-Encoding')
      assert_equal(4, @header.length)
      assert_equal(3, copy_header.length)

      @header.content_type = 'text/plain'
      copy_header.content_length = '1024'
      assert(@header.content_type == 'text/plain')
      assert(copy_header.content_type != 'text/plain')
      assert(@header.content_length != '1024')
      assert(copy_header.content_length == '1024')

      @header.delete('Content-Length')
      copy_header.delete('Content-Encoding')
      assert(! (@header.include? 'Content-Length'))
      assert((copy_header.include? 'Content-Length'))
      assert((@header.include? 'Content-Encoding'))
      assert(! (copy_header.include? 'Content-Encoding'))

      @header.connection = 'close'
      assert_equal('close', @header.connection)
      assert_nil(copy_header.connection)
    end

    def test_to_s
      @header['Content-Type'] = 'text/html'
      @header['Content-Length'] = '256'
      @header.push('Content-Encoding', 'deflate')
      @header.push('Content-Encoding', 'gzip')
      @header.push('Content-Encoding', 'compress')

      header_message = "Content-Type: text/html\r\n"
      header_message += "Content-Length: 256\r\n"
      header_message += "Content-Encoding: deflate, gzip, compress\r\n"
      assert_equal(header_message, @header.to_s)

      @header.set_multi_field('Content-Encoding')
      header_message = "Content-Type: text/html\r\n"
      header_message += "Content-Length: 256\r\n"
      header_message += "Content-Encoding: deflate\r\n"
      header_message += "Content-Encoding: gzip\r\n"
      header_message += "Content-Encoding: compress\r\n"
      assert_equal(header_message, @header.to_s)
    end

    def test_build
      header = WWWsrv::Header::build('Content-Type' => 'text/html', 'Content-Length' => '256')
      assert_instance_of(WWWsrv::Header, header)
      assert_equal('text/html', header.content_type)
      assert_equal('256', header.content_length)
    end

    def test_equals
      header1 = WWWsrv::Header::build('Content-Type' => 'text/html', 'Content-Length' => '256')
      header2 = WWWsrv::Header::build('Content-Type' => 'text/html', 'Content-Length' => '256')
      assert(header1.__id__ != header2.__id__)
      assert(header1 == header2)
    end
  end

  class TestHeaderParse < RUNIT::TestCase
    def setup
      @input = PseudoIO::new
    end

    def test_parse_header
      @input.write("Host: localhost\r\n")
      @input.write("User-Agent: #{WWWsrv::product_token}/#{WWWsrv::VERSION} (with RUNIT)\r\n")
      @input.write("\r\n")
      @input.write("Content body...")
      header = WWWsrv::Header::parse(@input)
      assert_equal(2, header.length)
      assert_equal('localhost', header.host)
      assert_equal("#{WWWsrv::product_token}/#{WWWsrv::VERSION} (with RUNIT)", header.user_agent)
    end

    def test_parse_header_eol_spaces
      @input.write("Host: localhost \r\n")
      @input.write("User-Agent: #{WWWsrv::product_token}/#{WWWsrv::VERSION} (with RUNIT) \r\n")
      @input.write("\r\n")
      @input.write("Content body...")
      header = WWWsrv::Header::parse(@input)
      assert_equal(2, header.length)
      assert_equal('localhost', header.host)
      assert_equal("#{WWWsrv::product_token}/#{WWWsrv::VERSION} (with RUNIT)", header.user_agent)
    end

    def test_parse_header_continue
      @input.write("Host:\r\n")
      @input.write(" localhost\r\n")
      @input.write("User-Agent: \r\n")
      @input.write(" #{WWWsrv::product_token}/#{WWWsrv::VERSION} (with RUNIT)\r\n")
      @input.write("\r\n")
      @input.write("Content body...")
      header = WWWsrv::Header::parse(@input)
      assert_equal(2, header.length)
      assert_equal('localhost', header.host)
      assert_equal("#{WWWsrv::product_token}/#{WWWsrv::VERSION} (with RUNIT)", header.user_agent)
    end

    def test_parse_header_new_line_sep
      @input.write("Host: localhost\n")
      @input.write("User-Agent: #{WWWsrv::product_token}/#{WWWsrv::VERSION} (with RUNIT)\n")
      @input.write("\n")
      @input.write("Content body...")
      header = WWWsrv::Header::parse(@input)
      assert_equal(2, header.length)
      assert_equal('localhost', header.host)
      assert_equal("#{WWWsrv::product_token}/#{WWWsrv::VERSION} (with RUNIT)", header.user_agent)
    end

    def test_parse_header_invalid
      @input.write("Host; localhost\r\n")
      @input.write("User-Agent; #{WWWsrv::product_token}/#{WWWsrv::VERSION} (with RUNIT)\r\n")
      @input.write("\r\n")
      @input.write("Content body...")
      assert_exception(WWWsrv::HeaderParseError) { WWWsrv::Header::parse(@input) }
    end

    def test_parse_header_invalid_continue
      @input.write(" localhost\r\n")
      @input.write("Host:\r\n")
      @input.write("User-Agent: #{WWWsrv::product_token}/#{WWWsrv::VERSION} (with RUNIT)\r\n")
      @input.write("\r\n")
      @input.write("Content body...")
      assert_exception(WWWsrv::HeaderParseError) { WWWsrv::Header::parse(@input) }
    end

    def test_parse_header_empty
      @input.write("\r\n")
      @input.write("Content body...")
      header = WWWsrv::Header::parse(@input)
      assert(header.empty?)
    end
  end
end

if ($0 == __FILE__) then
  if (ARGV.empty?) then
    TestWWWsrv::constants.map{ |name|
      TestWWWsrv::const_get(name)
    }.select{ |const|
      (const.kind_of? Class) && (const < RUNIT::TestCase)
    }.each do |test_case|
      RUNIT::CUI::TestRunner::run(test_case::suite)
    end
  else
    for class_name in ARGV
      test_case = TestWWWsrv.const_get("Test#{class_name}")
      RUNIT::CUI::TestRunner.run(test_case.suite)
    end
  end
end
