#!/usr/local/bin/ruby
# $Id: test_local.rb,v 1.21.2.1 2003/01/05 13:06:37 toki Exp $
# -
# Test of WWWsrv::MIMEresolver class
# Test of WWWsrv::LocalDirIndex class
# Test of WWWsrv::LocalFileDocument class

require 'docutils'
require 'pseudo_io'
require 'pseudo_doc'
require 'wwwsrv/logger'
require 'wwwsrv/request'
require 'wwwsrv/response'
require 'wwwsrv/local'
require 'runit/testcase'
require 'runit/cui/testrunner'

class Array
  def shuffle
    for i in (0...(self.length - 1))
      j = i + rand(self.length - i)
      self[i], self[j] = self[j], self[i]
    end

    self
  end
end

module TestWWWsrv
  class TestMIMEresolver < RUNIT::TestCase
    def setup
      @map = {
	'.htm'     => 'text/html',
	'.html'    => 'text/html',
	'_ja.html' => 'text/html; charset=EUC-JP',
	'.rb'      => 'text/plain',
	'.txt'     => 'text/plain'
      }

      @mime = WWWsrv::MIMEresolver::new
      for ext, mime_type in @map
	@mime.add_mime_type(ext, mime_type)
      end
    end

    def test_resolve
      assert_equal('text/html', @mime.resolve('/index.htm'))
      assert_equal('text/html', @mime.resolve('/index.html'))
      assert_equal('text/html; charset=EUC-JP', @mime.resolve('/index_ja.html'))
      assert_equal('text/plain', @mime.resolve('/index.rb'))
      assert_equal('text/plain', @mime.resolve('/index.txt'))
      assert_equal('application/octet-stream', @mime.resolve('/index.aaa'))
      assert_equal('application/octet-stream', @mime.resolve('/index.gtm'))
      assert_equal('application/octet-stream', @mime.resolve('/index.itm'))
      assert_equal('application/octet-stream', @mime.resolve('/index.zzz'))
    end

    def test_clone
      new_mime = @mime.clone
      new_mime.add_mime_type('_ja.html', 'text/html; charset=Shift_JIS')
      assert_equal('text/html; charset=EUC-JP', @mime.resolve('/index_ja.html'))
      assert_equal('text/html; charset=Shift_JIS', new_mime.resolve('/index_ja.html'))
    end

    def test_each
      for ext, mime_type in @mime
	assert_equal(@map[ext], mime_type)
	@map.delete(ext)
      end
      assert(@map.empty?)
    end
  end

  class TestLocalDirIndex < RUNIT::TestCase
    include DocumentUtils

    def setup
      @root_path = File::join(Dir::pwd, 'test')
      @first = 'first'
      @second = 'second'
      @third = 'third'
      @dot_test = '.test'
      @content = "Hello world.\n"

      Dir::mkdir(@root_path)
      [ @first,
	@second,
	@third,
	@dot_test
      ].each do |name|
	File::open(File::join(@root_path, name), 'w') do |output|
	  output.write(@content)
	end
      end

      @local_dir = WWWsrv::LocalDirIndex::new
    end

    def teardown
      [ @first,
	@second,
	@third,
	@dot_test
      ].each do |name|
	File::unlink(File::join(@root_path, name))
      end
      Dir::rmdir(@root_path)
    end


    def test_new
      assert_instance_of(WWWsrv::LocalDirIndex, @local_dir)
    end

    def test_respond
      srv_opt, log_output = make_srv_opt
      request, body_input = make_request('GET', '/')
      response = make_response
      response[:local_path] = @root_path

      dir_response = nil
      dir_content = nil
      @local_dir.respond(srv_opt, '', request, response) do |arg, *opts|
	unless (dir_response) then
	  dir_response = arg
	else
	  dir_content = '' unless dir_content
	  dir_content.concat(arg)
	end
      end

      assert_instance_of(WWWsrv::Response, dir_response)
      assert_equal(200, dir_response.status)
      assert_not_nil(dir_response.header.content_type)
      assert_match(/#{Regexp::quote(@first)}/, dir_content)
      assert_match(/#{Regexp::quote(@second)}/, dir_content)
      assert_match(/#{Regexp::quote(@third)}/, dir_content)
      assert_not_match(/#{Regexp::quote(@dot_test)}/, dir_content)
    end
  end

  class TestLocalFileDocument < RUNIT::TestCase
    include ParseDate
    include DocumentUtils

    def setup
      @root_local_dir = File::expand_path(File::join(Dir::pwd, 'test'))
      @doc_local_dir = File::join(@root_local_dir, 'doc')
      @content = <<-EOF
<html>
<head>
<title>Test</title>
</head>
<body>
<h1>Test</h1>
<p>Hello world.</p>
</body>
</html>
      EOF
      @root_index_local_path = File::join(@root_local_dir, 'index.html')
      @doc_index_local_path = File::join(@doc_local_dir, 'index.htm')
      @last_modified = Time::gm(2000, 11, 19, 12, 12, 12)
      @document = WWWsrv::LocalFileDocument::new(@root_local_dir)

      [ @root_local_dir,
	@doc_local_dir
      ].each do |path|
	Dir::mkdir(path)
      end

      [ @root_index_local_path,
	@doc_index_local_path,
      ].each do |path|
	File::open(path, 'w') do |index|
	  index.binmode
	  index.write(@content)
	end
	File::utime(@last_modified, @last_modified, path)
      end
    end

    def teardown
      [ @doc_index_local_path,
	@root_index_local_path
      ].each do |path|
	File::unlink(path)
      end

      [ @doc_local_dir,
	@root_local_dir
      ].each do |path|
	Dir::rmdir(path)
      end
    end

    def test_new
      assert_instance_of(WWWsrv::LocalFileDocument, @document)
    end

    def test_head
      srv_opt, log_output = make_srv_opt
      request, body_input = make_request('HEAD', '/index.html')
      response = make_response
      respondent = try_to_respond(@document, srv_opt, '', request, response)
      assert_equal(:success, respondent[:doc_status])
      assert_equal(@root_index_local_path, respondent[:response][:local_path])
      assert_equal(WWWsrv::HTTP_1_1, respondent[:response].version)
      assert_equal(200, respondent[:response].status)
      assert_equal('OK', respondent[:response].reason)
      assert_equal('text/html', respondent[:response].header.content_type)
      assert_not_nil(respondent[:response].header.content_length)
      assert_not_nil(respondent[:response].header.etag)
      assert_equal(@last_modified, Time::gm(*parsedate(respondent[:response].header.last_modified)[0, 6]))
      assert_nil(respondent[:body])
      assert(respondent[:body_alist].empty?)
    end

    def test_get
      srv_opt, log_output = make_srv_opt
      request, body_input = make_request('GET', '/index.html')
      response = make_response
      respondent = try_to_respond(@document, srv_opt, '', request, response)
      assert_equal(:success, respondent[:doc_status])
      assert_equal(@root_index_local_path, respondent[:response][:local_path])
      assert_equal(WWWsrv::HTTP_1_1, respondent[:response].version)
      assert_equal(200, respondent[:response].status)
      assert_equal('OK', respondent[:response].reason)
      assert_equal('text/html', respondent[:response].header.content_type)
      assert_equal(respondent[:body].length.to_s, respondent[:response].header.content_length)
      assert_not_nil(respondent[:response].header.etag)
      assert_equal(@last_modified, Time::gm(*parsedate(respondent[:response].header.last_modified)[0, 6]))
      assert_equal(@content, respondent[:body])
    end

    def test_get_root
      srv_opt, log_output = make_srv_opt
      request, body_input = make_request('GET', '/')
      response = make_response
      respondent = try_to_respond(@document, srv_opt, '', request, response)
      assert_equal(:success, respondent[:doc_status])
      assert_equal(@root_index_local_path, respondent[:response][:local_path])
      assert_equal(WWWsrv::HTTP_1_1, respondent[:response].version)
      assert_equal(200, respondent[:response].status)
      assert_equal('OK', respondent[:response].reason)
      assert_equal('text/html', respondent[:response].header.content_type)
      assert_equal(respondent[:body].length.to_s, respondent[:response].header.content_length)
      assert_not_nil(respondent[:response].header.etag)
      assert_equal(@last_modified, Time::gm(*parsedate(respondent[:response].header.last_modified)[0, 6]))
      assert_equal(@content, respondent[:body])
    end

    def test_dir
      srv_opt, log_output = make_srv_opt
      request, body_input = make_request('GET', '/doc/')
      response = make_response
      respondent = try_to_respond(@document, srv_opt, '', request, response)
      assert_equal(:success, respondent[:doc_status])
      assert_equal(@doc_index_local_path, respondent[:response][:local_path])
      assert_equal(WWWsrv::HTTP_1_1, respondent[:response].version)
      assert_equal(200, respondent[:response].status)
      assert_equal('OK', respondent[:response].reason)
      assert_equal('text/html', respondent[:response].header.content_type)
      assert_equal(respondent[:body].length.to_s, respondent[:response].header.content_length)
      assert_not_nil(respondent[:response].header.etag)
      assert_equal(@last_modified, Time::gm(*parsedate(respondent[:response].header.last_modified)[0, 6]))

      srv_opt, log_output = make_srv_opt
      request, body_input = make_request('GET', '/test/doc/')
      response = make_response
      respondent = try_to_respond(@document, srv_opt, '/test', request, response)
      assert_equal(:success, respondent[:doc_status])
      assert_equal(@doc_index_local_path, respondent[:response][:local_path])
      assert_equal(WWWsrv::HTTP_1_1, respondent[:response].version)
      assert_equal(200, respondent[:response].status)
      assert_equal('OK', respondent[:response].reason)
      assert_equal('text/html', respondent[:response].header.content_type)
      assert_equal(respondent[:body].length.to_s, respondent[:response].header.content_length)
      assert_not_nil(respondent[:response].header.etag)
      assert_equal(@last_modified, Time::gm(*parsedate(respondent[:response].header.last_modified)[0, 6]))

      srv_opt, log_output = make_srv_opt
      request, body_input = make_request('GET', '/test/doc/')
      response = make_response
      respondent = try_to_respond(@document, srv_opt, '/test/', request, response)
      assert_equal(:success, respondent[:doc_status])
      assert_equal(@doc_index_local_path, respondent[:response][:local_path])
      assert_equal(WWWsrv::HTTP_1_1, respondent[:response].version)
      assert_equal(200, respondent[:response].status)
      assert_equal('OK', respondent[:response].reason)
      assert_equal('text/html', respondent[:response].header.content_type)
      assert_equal(respondent[:body].length.to_s, respondent[:response].header.content_length)
      assert_not_nil(respondent[:response].header.etag)
      assert_equal(@last_modified, Time::gm(*parsedate(respondent[:response].header.last_modified)[0, 6]))
    end

    def test_dir_move
      srv_opt, log_output = make_srv_opt
      request, body_input = make_request('GET', '/doc')
      response = make_response
      respondent = try_to_respond(@document, srv_opt, '', request, response)
      assert_equal(:success, respondent[:doc_status])
      assert_nil(respondent[:response][:local_path])
      assert_equal(WWWsrv::HTTP_1_1, respondent[:response].version)
      assert_equal(301, respondent[:response].status)
      assert_equal('Moved Permanently', respondent[:response].reason)
      assert_equal('http://localhost:80/doc/', respondent[:response].header.location)
      assert_nil(respondent[:response].header.content_length)

      srv_opt, log_output = make_srv_opt
      request, body_input = make_request('GET', '/test')
      response = make_response
      respondent = try_to_respond(@document, srv_opt, '/test', request, response)
      assert_equal(:success, respondent[:doc_status])
      assert_nil(respondent[:response][:local_path])
      assert_equal(WWWsrv::HTTP_1_1, respondent[:response].version)
      assert_equal(301, respondent[:response].status)
      assert_equal('Moved Permanently', respondent[:response].reason)
      assert_equal('http://localhost:80/test/', respondent[:response].header.location)
      assert_nil(respondent[:response].header.content_length)
    end
  end

  class TestLocalFileDocumentWithRangeRequest < RUNIT::TestCase
    include DocumentUtils

    RANDOM_CONTENT = ''
    char_list = ('A'..'Z').to_a + ('a'..'z').to_a + ('0'..'9').to_a
    while (RANDOM_CONTENT.length < 1024*256)
      char_list.shuffle
      RANDOM_CONTENT.concat(char_list.join)
    end
    RANDOM_CONTENT.freeze

    def setup
      @content = RANDOM_CONTENT.dup
      @content_root = Dir::getwd
      @content_path = "test_partial_content.dat"
      File::open(@content_path, 'w') do |output|
	output.write(@content)
      end

      @local_doc = WWWsrv::LocalFileDocument::new(@content_root)
    end

    def teardown
      File::delete(@content_path)
    end

    def test_without_range
      srv_opt, log_output = make_srv_opt
      request, body_input = make_request('GET', "/#{@content_path}")
      response = make_response
      respondent = try_to_respond(@local_doc, srv_opt, '', request, response)
      assert_equal(:success, respondent[:doc_status])
      assert_equal(200, respondent[:response].status)
      assert_equal('bytes', respondent[:response].header.accept_ranges)
      assert_nil(respondent[:response].header.content_range)
      assert_equal(@content, respondent[:body])
    end

    def test_range_trisected_pos1
      trisected_pos1 = @content.length / 3
      trisected_pos2 = @content.length / 3 * 2

      srv_opt, log_output = make_srv_opt
      request, body_input = make_request('GET', "/#{@content_path}")
      request.header.range = "bytes=0-#{trisected_pos1}"
      response = make_response
      respondent = try_to_respond(@local_doc, srv_opt, '', request, response)
      assert_equal(:success, respondent[:doc_status])
      assert_equal(206, respondent[:response].status)
      assert_equal('bytes', respondent[:response].header.accept_ranges)
      assert_equal("bytes 0-#{trisected_pos1}/#{@content.length}", respondent[:response].header.content_range)
      assert_equal(@content[0..trisected_pos1], respondent[:body])
    end

    def test_range_trisected_pos2
      trisected_pos1 = @content.length / 3
      trisected_pos2 = @content.length / 3 * 2

      srv_opt, log_output = make_srv_opt
      request, body_input = make_request('GET', "/#{@content_path}")
      request.header.range = "bytes=#{trisected_pos1}-#{trisected_pos2}"
      response = make_response
      respondent = try_to_respond(@local_doc, srv_opt, '', request, response)
      assert_equal(:success, respondent[:doc_status])
      assert_equal(206, respondent[:response].status)
      assert_equal('bytes', respondent[:response].header.accept_ranges)
      assert_equal("bytes #{trisected_pos1}-#{trisected_pos2}/#{@content.length}", respondent[:response].header.content_range)
      assert_equal(@content[trisected_pos1..trisected_pos2], respondent[:body])
    end

    def test_range_trisected_pos3
      trisected_pos1 = @content.length / 3
      trisected_pos2 = @content.length / 3 * 2

      srv_opt, log_output = make_srv_opt
      request, body_input = make_request('GET', "/#{@content_path}")
      request.header.range = "bytes=#{trisected_pos2}-#{@content.length - 1}"
      response = make_response
      respondent = try_to_respond(@local_doc, srv_opt, '', request, response)
      assert_equal(:success, respondent[:doc_status])
      assert_equal(206, respondent[:response].status)
      assert_equal('bytes', respondent[:response].header.accept_ranges)
      assert_equal("bytes #{trisected_pos2}-#{@content.length - 1}/#{@content.length}", respondent[:response].header.content_range)
      assert_equal(@content[trisected_pos2..(@content.length - 1)], respondent[:body])
      assert_equal(@content[trisected_pos2..-1], respondent[:body])
    end

    def test_range_overflow_right
      middle_pos = @content.length / 2
      last_pos = @content.length - 1
      overflow_pos = last_pos + 1

      srv_opt, log_output = make_srv_opt
      request, body_input = make_request('GET', "/#{@content_path}")
      request.header.range = "bytes=#{middle_pos}-#{overflow_pos}"
      response = make_response
      respondent = try_to_respond(@local_doc, srv_opt, '', request, response)
      assert_equal(:success, respondent[:doc_status])
      assert_equal(206, respondent[:response].status)
      assert_equal('bytes', respondent[:response].header.accept_ranges)
      assert_equal("bytes #{middle_pos}-#{last_pos}/#{@content.length}", respondent[:response].header.content_range)
      assert_equal(@content[middle_pos..last_pos], respondent[:body])
      assert_equal(@content[middle_pos..-1], respondent[:body])
    end

    def test_range_overflow_right_omit_left
      overflow_size = @content.length + 1

      srv_opt, log_output = make_srv_opt
      request, body_input = make_request('GET', "/#{@content_path}")
      request.header.range = "bytes=-#{overflow_size}"
      response = make_response
      respondent = try_to_respond(@local_doc, srv_opt, '', request, response)
      assert_equal(:success, respondent[:doc_status])
      assert_equal(206, respondent[:response].status)
      assert_equal('bytes', respondent[:response].header.accept_ranges)
      assert_equal("bytes 0-#{@content.length - 1}/#{@content.length}", respondent[:response].header.content_range)
      assert_equal(@content, respondent[:body])
    end

    def test_range_overflow_left
      overflow_left_pos = @content.length
      overflow_right_pos = overflow_left_pos + @content.length / 2

      srv_opt, log_output = make_srv_opt
      request, body_input = make_request('GET', "/#{@content_path}")
      request.header.range = "bytes=#{overflow_left_pos}-#{overflow_right_pos}"
      response = make_response
      respondent = try_to_respond(@local_doc, srv_opt, '', request, response)
      assert_equal(:error, respondent[:doc_status])
      assert_equal(416, respondent[:err_args].first)
    end

    def test_range_overflow_left_omit_right
      overflow_left_pos = @content.length

      srv_opt, log_output = make_srv_opt
      request, body_input = make_request('GET', "/#{@content_path}")
      request.header.range = "bytes=#{overflow_left_pos}-"
      response = make_response
      respondent = try_to_respond(@local_doc, srv_opt, '', request, response)
      assert_equal(:error, respondent[:doc_status])
      assert_equal(416, respondent[:err_args].first)
    end

    def test_range_half_omit_left
      half_size = @content.length / 2
      middle_pos = @content.length - half_size
      last_pos = @content.length - 1

      srv_opt, log_output = make_srv_opt
      request, body_input = make_request('GET', "/#{@content_path}")
      request.header.range = "bytes=-#{half_size}"
      response = make_response
      respondent = try_to_respond(@local_doc, srv_opt, '', request, response)
      assert_equal(:success, respondent[:doc_status])
      assert_equal(206, respondent[:response].status)
      assert_equal('bytes', respondent[:response].header.accept_ranges)
      assert_equal("bytes #{middle_pos}-#{last_pos}/#{@content.length}", respondent[:response].header.content_range)
      assert_equal(@content[middle_pos..last_pos], respondent[:body])
      assert_equal(@content[middle_pos..-1], respondent[:body])
    end

    def test_range_half_omit_right
      half_size = @content.length / 2
      middle_pos = @content.length - half_size
      last_pos = @content.length - 1

      srv_opt, log_output = make_srv_opt
      request, body_input = make_request('GET', "/#{@content_path}")
      request.header.range = "bytes=#{middle_pos}-"
      response = make_response
      respondent = try_to_respond(@local_doc, srv_opt, '', request, response)
      assert_equal(:success, respondent[:doc_status])
      assert_equal(206, respondent[:response].status)
      assert_equal('bytes', respondent[:response].header.accept_ranges)
      assert_equal("bytes #{middle_pos}-#{last_pos}/#{@content.length}", respondent[:response].header.content_range)
      assert_equal(@content[middle_pos..last_pos], respondent[:body])
      assert_equal(@content[middle_pos..-1], respondent[:body])
    end

    def test_range_invalid
      left_pos = @content.length / 3 * 2
      right_pos = @content.length / 3
      assert(left_pos > right_pos)

      srv_opt, log_output = make_srv_opt
      request, body_input = make_request('GET', "/#{@content_path}")
      request.header.range = "bytes=#{left_pos}-#{right_pos}"
      response = make_response
      respondent = try_to_respond(@local_doc, srv_opt, '', request, response)
      assert_equal(:error, respondent[:doc_status])
      assert_equal(416, respondent[:err_args].first)
    end
  end
end

if ($0 == __FILE__) then
  if (ARGV.empty?) then
    TestWWWsrv::constants.map{ |name|
      TestWWWsrv::const_get(name)
    }.select{ |const|
      (const.kind_of? Class) && (const < RUNIT::TestCase)
    }.each do |test_case|
      RUNIT::CUI::TestRunner::run(test_case::suite)
    end
  else
    for class_name in ARGV
      test_case = TestWWWsrv.const_get("Test#{class_name}")
      RUNIT::CUI::TestRunner.run(test_case.suite)
    end
  end
end
