#!/usr/local/bin/ruby
# $Id: test_reader.rb,v 1.9.2.2 2003/01/05 13:06:44 toki Exp $
# -
# Test of WWWsrv::LengthLimitedReader class
# Test of WWWsrv::LineReader class
# Test of WWWsrv::ContentReader class
# Test of WWWsrv::ChunkedContentReader class
# Test of WWWsrv::EmptyReader class

require 'pseudo_io'
require 'wwwsrv/logger'
require 'wwwsrv/reader'
require 'runit/testcase'
require 'runit/cui/testrunner'

module TestWWWsrv
  module TestReader
    def test_reader_new
      assert((make_reader('').kind_of? WWWsrv::Reader))
    end

    def test_read
      reader = make_reader('*' * 37)
      assert_equal('*' * 37, reader.read)

      read_size = 23
      for len in 1..256
	src = '*' * len
	reader = make_reader(src)
	0.step(len - 1, read_size) do |i|
	  assert_equal(src[i, read_size], reader.read(read_size))
	end
      end

      nil
    end

    def test_getc
      char_list = (0..255).to_a
      char_src = char_list.map{ |ch| ch.chr }.join

      reader = make_reader(char_src)
      for ch in char_list
	assert_equal(ch, reader.getc)
      end
      assert_nil(reader.getc)

      reader
    end

    def test_gets
      l = '*'
      line_buf = [ l + "\n" ]
      8.times do
	l *= 2
	line_buf.push(l + "\n")
      end
      line_buf.push('EOF')

      reader = make_reader(line_buf.join)
      until (line_buf.empty?)
	assert_equal(line_buf.shift, reader.gets)
      end
      assert_nil(reader.gets)

      reader
    end

    def test_readchar
      char_list = (0..255).to_a
      char_src = char_list.map{ |ch| ch.chr }.join

      reader = make_reader(char_src)
      for ch in char_list
	assert_equal(ch, reader.readchar)
      end
      assert_exception(EOFError) { reader.readchar }

      reader
    end

    def test_readline
      l = '*'
      line_buf = [ l + "\n" ]
      8.times do
	l *= 2
	line_buf.push(l + "\n")
      end
      line_buf.push('EOF')

      reader = make_reader(line_buf.join)
      until (line_buf.empty?)
	assert_equal(line_buf.shift, reader.readline)
      end
      assert_exception(EOFError) { reader.readline }

      reader
    end

    def test_readlines
      l = '*'
      line_buf = [ l + "\n" ]
      8.times do
	l *= 2
	line_buf.push(l + "\n")
      end
      line_buf.push('EOF')

      reader = make_reader(line_buf.join)
      assert_equal(line_buf, reader.readlines)
      assert_nil(reader.gets)

      reader
    end
  end

  class TestStringReader < RUNIT::TestCase
    include TestReader

    def make_reader(src)
      WWWsrv::StringReader::new(src.dup)
    end
    private :make_reader

    def setup
      @content = "Hello world.\n"
    end

    def test_new
      assert_instance_of(WWWsrv::StringReader, make_reader(''))
    end

    def test_read_all
      reader = make_reader(@content)
      assert_equal(@content.length, reader.length)
      assert_equal(@content, reader.read)
      assert_equal(0, reader.length)
      assert_nil(reader.read)
    end

    def test_read_all2
      reader = make_reader(@content)
      assert_equal(@content.length, reader.length)
      assert_equal(@content, reader.read(1024))
      assert_equal(0, reader.length)
      assert_nil(reader.read(1024))
    end

    def test_read_fragments
      reader = make_reader(@content)
      assert_equal('', reader.read(0))
      assert_equal(@content.length, reader.length)

      content_spool = ''
      rest_length = @content.length
      while (fragment = reader.read(4))
	rest_length -= fragment.length
	assert_equal(rest_length, reader.length)
	content_spool.concat(fragment)
      end

      assert_equal(@content, content_spool)
      assert_nil(reader.read(4))
      assert_nil(reader.read(0))
    end

    def test_concat
      append_message = "Bye.\n"
      reader = make_reader(@content)
      assert_equal(@content.length, reader.length)
      reader.concat(append_message)
      assert_equal(@content.length + append_message.length, reader.length)
      assert_equal(@content + append_message, reader.read)
      assert_equal(0, reader.length)
      assert_nil(reader.read)
    end

    def test_gets
      reader = super
      assert_equal(0, reader.length)
      assert_nil(reader.gets)
    end

    def test_unget
      preceding_message = 'preamble'
      reader = make_reader(@content)
      assert_equal(@content.length, reader.length)
      reader.unget(preceding_message)
      assert_equal(preceding_message.length + @content.length, reader.length)
      assert_equal(preceding_message + @content, reader.read)
      assert_equal(0, reader.length)
      assert_nil(reader.read)
      reader.unget(@content)
      assert_equal(@content.length, reader.length)
      assert_equal(@content, reader.read)
      assert_equal(0, reader.length)
      assert_nil(reader.read)
    end
  end

  class TestLengthLimitedReader < RUNIT::TestCase
    include TestReader

    def make_reader(src)
      input = PseudoIO::new
      input.write(src)
      input.write('EOF')
      WWWsrv::LengthLimitedReader::new(input, src.length)
    end
    private :make_reader

    def setup
      @content = "Hello world.\n"
      @content_tail = "EOF\n"
      @input = PseudoIO::new
      @input.write(@content + @content_tail)
    end

    def test_new
      assert_instance_of(WWWsrv::LengthLimitedReader,
			 WWWsrv::LengthLimitedReader::new(@input, @content.length))
    end

    def test_read_all
      reader = WWWsrv::LengthLimitedReader::new(@input, @content.length)
      assert_equal(@content, reader.read)
      assert_nil(reader.read)
      assert_equal(@content_tail, @input.read)
    end

    def test_read_all2
      reader = WWWsrv::LengthLimitedReader::new(@input, @content.length)
      assert_equal(@content, reader.read(1024))
      assert_nil(reader.read(1024))
      assert_equal(@content_tail, @input.read)
    end

    def test_read_fragments
      reader = WWWsrv::LengthLimitedReader::new(@input, @content.length)
      assert_equal('', reader.read(0))

      content_spool = ''
      while (fragment = reader.read(4))
	content_spool.concat(fragment)
      end

      assert_equal(@content, content_spool)
      assert_nil(reader.read(4))
      assert_nil(reader.read(0))
      assert_equal(@content_tail, @input.read)
    end

    def test_read_all_with_padding
      @input.write(@content_tail.succ)
      reader = WWWsrv::LengthLimitedReader::new(@input, @content.length, @content_tail.length)
      assert_equal(@content, reader.read)
      assert_nil(reader.read)
      assert_equal(@content_tail.succ, @input.read)
    end

    def test_read_all2_with_padding
      @input.write(@content_tail.succ)
      reader = WWWsrv::LengthLimitedReader::new(@input, @content.length, @content_tail.length)
      assert_equal(@content, reader.read(1024))
      assert_nil(reader.read)
      assert_equal(@content_tail.succ, @input.read)
    end

    def test_read_fragments_with_padding
      @input.write(@content_tail.succ)
      reader = WWWsrv::LengthLimitedReader::new(@input, @content.length, @content_tail.length)
      assert_equal('', reader.read(0))

      content_spool = ''
      while (fragment = reader.read(4))
	content_spool.concat(fragment)
      end

      assert_equal(@content, content_spool)
      assert_nil(reader.read(4))
      assert_nil(reader.read(0))
      assert_equal(@content_tail.succ, @input.read)
    end
  end

  class TestChunkedContentReader < RUNIT::TestCase
    def setup
      @chunk = [
	[ "Hello world.\n", [ 'debug', 'true' ] ],
	[ "Bye.\n", [] ]
      ]
      @trailer = [
	[ 'First', 'a' ],
	[ 'Second', 'b' ]
      ]

      @chunked_content = ''
      for chunk_data, chunk_extension in @chunk
	@chunked_content.concat(format('%x', chunk_data.length))
	for chunk_ext_name, chunk_ext_value in chunk_extension
	  @chunked_content.concat('; ')
	  @chunked_content.concat(chunk_ext_name)
	  if (chunk_ext_value) then
	    @chunked_content.concat('=')
	    @chunked_content.concat(chunk_ext_value)
	  end
	end
	@chunked_content.concat("\r\n")
	@chunked_content.concat(chunk_data)
	@chunked_content.concat("\r\n")
      end
      @chunked_content.concat("0\r\n")
      for field_name, field_value in @trailer
	@chunked_content.concat(field_name)
	@chunked_content.concat(': ')
	@chunked_content.concat(field_value)
	@chunked_content.concat("\r\n")
      end
      @chunked_content.concat("\r\n")

      @content = @chunk.map{ |chunk_data, chunk_extension|
	chunk_data
      }.join

      @input = PseudoIO::new
      @input.write(@chunked_content)
      @reader = WWWsrv::ChunkedTransferCodingReader::new(@input)
    end

    def test_new
      assert_instance_of(WWWsrv::ChunkedTransferCodingReader, @reader)
    end

    def test_read_all
      assert_equal(@content, @reader.read)
      assert_nil(@reader.read)
    end

    def test_read_all2
      assert_equal(@content, @reader.read(1024))
      assert_nil(@reader.read(1024))
    end

    def test_read_fragment
      assert_equal('', @reader.read(0))

      content_spool = ''
      while (fragment = @reader.read(4))
	content_spool.concat(fragment)
      end

      assert_equal(@content, content_spool)
      assert_nil(@reader.read(4))
      assert_nil(@reader.read(0))
    end

    def test_read_chunk_with_ext
      assert_nil(@reader.trailer)

      loop do
	chunk_data, chunk_extension = @reader.read_chunk_with_ext
	chunk_data or break
	chunk = @chunk.shift
	assert_equal(chunk[0], chunk_data)
	chunk[1].each_with_index do |(chunk_ext_name, chunk_ext_value), i|
	  assert_equal(chunk_ext_name, chunk_extension[i][0])
	  assert_equal(chunk_ext_value, chunk_extension[i][1])
	end
      end
      assert_nil(@reader.read_chunk_with_ext)
      assert_nil(@reader.read_chunk)

      @reader.trailer.each_with_index do |(field_name, field_value), i|
	assert_equal(@trailer[i][0].upcase, field_name.upcase)
	assert_equal(@trailer[i][1], field_value)
      end
    end

    include TestReader

    def make_reader(src)
      input = PseudoIO::new
      0.step(src.length - 1, 23) do |i|
	chunk = src[i, 23]
	input.write(format("%x\r\n", chunk.length))
	input.write(chunk + "\r\n")
      end
      input.write("0\r\n")
      input.write("\r\n")
      WWWsrv::ChunkedTransferCodingReader::new(input)
    end
    private :make_reader
  end

  class TestEmptyReader < RUNIT::TestCase
    def setup
      @content_reader = WWWsrv::EmptyReader::instance
    end

    def test_new
      assert_instance_of(WWWsrv::EmptyReader, @content_reader)
    end

    def test_read_all
      assert_nil(@content_reader.read)
    end

    def test_read_all2
      assert_nil(@content_reader.read(1024))
    end
  end
end

if ($0 == __FILE__) then
  if (ARGV.empty?) then
    TestWWWsrv::constants.map{ |name|
      TestWWWsrv::const_get(name)
    }.select{ |const|
      (const.kind_of? Class) && (const < RUNIT::TestCase)
    }.each do |test_case|
      RUNIT::CUI::TestRunner::run(test_case::suite)
    end
  else
    for class_name in ARGV
      test_case = TestWWWsrv.const_get("Test#{class_name}")
      RUNIT::CUI::TestRunner.run(test_case.suite)
    end
  end
end
