#!/usr/local/bin/ruby
# $Id: test_request.rb,v 1.13.2.2 2003/04/11 14:08:57 toki Exp $
# -
# Test of WWWsrv::Host class
# Test of WWWsrv::Client class
# Test of WWWsrv::Request class

require 'pseudo_io'
require 'wwwsrv/htmlconv'
require 'wwwsrv/request'
require 'runit/testcase'
require 'runit/cui/testrunner'

module TestWWWsrv
  class TestHost < RUNIT::TestCase
    def setup
      @host = WWWsrv::Host::fetch_cache('localhost')
      @host2 = WWWsrv::Host::fetch_cache('localhost', 8888)
    end

    def test_new
      assert_instance_of(WWWsrv::Host, @host)
      assert_instance_of(WWWsrv::Host, @host2)
    end

    def test_host
      assert_equal('localhost', @host.name)
      assert_equal('localhost', @host2.name)
    end

    def test_port
      assert_equal(80, @host.port)
      assert_equal(8888, @host2.port)
    end

    def test_to_s
      assert_equal('localhost:80', @host.to_s)
      assert_equal('localhost:8888', @host2.to_s)
    end

    def test_equal
      host = WWWsrv::Host::fetch_cache('localhost')
      assert(host == @host)
      assert(host != @host2)
    end
  end

  class TestClient < RUNIT::TestCase
    def setup
      @client = WWWsrv::Client::fetch_cache('localhost', '127.0.0.1')
      @client2 = WWWsrv::Client::fetch_cache('tester', '192.168.0.1')
    end

    def test_new
      assert_instance_of(WWWsrv::Client, @client)
      assert_instance_of(WWWsrv::Client, @client2)
    end

    def test_name
      assert_equal('localhost', @client.name)
      assert_equal('tester', @client2.name)
    end

    def test_address
      assert_equal('127.0.0.1', @client.address)
      assert_equal('192.168.0.1', @client2.address)
    end

    def test_to_s
      assert_equal('localhost(127.0.0.1)', @client.to_s)
      assert_equal('tester(192.168.0.1)', @client2.to_s)
    end

    def test_to_i
      assert_equal(0b01111111_00000000_00000000_00000001, @client.to_i)
      assert_equal(0b11000000_10101000_00000000_00000001, @client2.to_i)
    end

    def test_equal
      client = WWWsrv::Client::fetch_cache('localhost', '127.0.0.1')
      assert(client == @client)
      assert(client != @client2)
    end
  end

  class TestRequest < RUNIT::TestCase
    def setup
      @header = WWWsrv::Header::new
      @header.host = 'localhost'
      @header.user_agent = 'Test/0.9 (with RUNIT)'
      @content = PseudoIO::new
      @host = WWWsrv::Host::fetch_cache('localhost')
      @client = WWWsrv::Client::fetch_cache('localhost', '127.0.0.1')
      @request = WWWsrv::Request::new('GET', '/index.html', WWWsrv::HTTP_1_0, @header)
      @request.body = @content
      @request.host = @host
      @request.client = @client
      @request_null = WWWsrv::Request::new
    end

    def test_new
      assert_instance_of(WWWsrv::Request, @request)
      assert_instance_of(WWWsrv::Request, @request_null)
    end

    def test_method
      assert_equal('GET', @request.method)
      assert_nil(@request_null.method)
      assert_equal('GET', @request_null.method = 'GET')
      assert_equal('GET', @request_null.method)
    end

    def test_path
      assert_equal('/index.html', @request.path)
      assert_nil(@request_null.path)
      assert_equal('/index.html', @request_null.path = '/index.html')
      assert_equal('/index.html', @request_null.path)
    end

    def test_version
      assert_equal(WWWsrv::HTTP_1_0, @request.version)
      assert_nil(@request_null.version)
      assert_equal(WWWsrv::HTTP_1_0, @request_null.version = WWWsrv::HTTP_1_0)
      assert_equal(WWWsrv::HTTP_1_0, @request_null.version)
    end

    def test_header
      assert_equal(@header, @request.header)
      assert(@request_null.header.empty?)
      assert_equal(@header, @request_null.header = @header)
      assert_equal(@header, @request_null.header)
    end

    def test_host
      assert_equal(@host, @request.host)
      assert_nil(@request_null.host)
      assert_equal(@host, @request_null.host = @host)
      assert_equal(@host, @request_null.host)
    end

    def test_respond
      assert_equal(@content, @request.body)
      assert_nil(@request_null.body)
      assert_equal(@content, @request_null.body = @content)
      assert_equal(@content, @request_null.body)
    end

    def test_client
      assert_equal(@client, @request.client)
      assert_nil(@request_null.client)
      assert_equal(@client, @request_null.client = @client)
      assert_equal(@client, @request_null.client)
    end

    def test_normalized_path
      assert_equal('/index.html', @request.normalized_path)
      @request.path = '/foo/./bar'
      assert_equal('/foo/bar', @request.normalized_path)
      @request.path = '/foo/../bar'
      assert_equal('/bar', @request.normalized_path)
      @request.path = '/foo/.'
      assert_equal('/foo/', @request.normalized_path)
      @request.path = '/foo/bar/..'
      assert_equal('/foo/', @request.normalized_path)
      @request.path = '/../foo'
      assert_equal('/foo', @request.normalized_path)
      @request.path = '/foo/.'
      assert_equal('/foo/', @request.normalized_path)
      @request.path = '/../foo'
      assert_equal('/foo', @request.normalized_path)
      @request.path = '/..'
      assert_equal('/', @request.normalized_path)
      @request.path = '/foo?bar'
      assert_equal('/foo', @request.normalized_path)
      @request.path = '/foo+bar'
      assert_equal('/foo bar', @request.normalized_path)
      @request.path = '/foo/%7Ebar'
      assert_equal('/foo/~bar', @request.normalized_path)
      assert_nil(@request_null.normalized_path)
    end

    def test_condition
      etag = '"1234-5678"'
      mtime = Time::gm(2001, 11, 19, 0, 0, 0)

      @request.header = WWWsrv::Header::new
      @request.header.if_match = '*'
      assert_equal(200, @request.cond_if(mtime, etag))

      @request.header = WWWsrv::Header::new
      @request.header.if_match = '"1234-5678"'
      assert_equal(200, @request.cond_if(mtime, etag))

      @request.header = WWWsrv::Header::new
      @request.header.if_match = '"8765-4321"'
      assert_equal(412, @request.cond_if(mtime, etag))

      @request.header = WWWsrv::Header::new
      @request.header.if_none_match = '*'
      assert_equal(304, @request.cond_if(mtime, etag))

      @request.header = WWWsrv::Header::new
      @request.header.if_none_match = '"1234-5678"'
      assert_equal(304, @request.cond_if(mtime, etag))

      @request.header = WWWsrv::Header::new
      @request.header.if_none_match = '"8765-4321"'
      assert_equal(200, @request.cond_if(mtime, etag))

      @request.header = WWWsrv::Header::new
      @request.header.if_modified_since = WWWsrv::HTMLconv::http_date(mtime)
      assert_equal(304, @request.cond_if(mtime, etag))

      @request.header = WWWsrv::Header::new
      @request.header.if_modified_since = WWWsrv::HTMLconv::http_date(mtime)
      assert_equal(200, @request.cond_if(mtime + 1, etag))

      @request.header = WWWsrv::Header::new
      @request.header.if_unmodified_since = WWWsrv::HTMLconv::http_date(mtime)
      assert_equal(200, @request.cond_if(mtime, etag))

      @request.header = WWWsrv::Header::new
      @request.header.if_unmodified_since = WWWsrv::HTMLconv::http_date(mtime)
      assert_equal(412, @request.cond_if(mtime + 1, etag))

      @request.header = WWWsrv::Header::new
      @request.header.if_none_match = '"8765-4321"'
      @request.header.if_modified_since = WWWsrv::HTMLconv::http_date(mtime)
      assert_equal(200, @request.cond_if(mtime, etag))

      @request.header = WWWsrv::Header::new
      @request.header.if_match = '"1234-5678"'
      @request.header.if_unmodified_since = WWWsrv::HTMLconv::http_date(mtime - 1)
      assert_equal(200, @request.cond_if(mtime, etag))
    end

    def test_params
      assert(@request.params.empty?)

      param_data = [
	%w[ double first ],
	%w[ double second ],
	%w[ debug true ]
      ].map{ |param_name, param_value|
	WWWsrv::HTMLconv::encodeURL(param_name) + '=' +
	  WWWsrv::HTMLconv::encodeURL(param_value)
      }.join('&')

      request = WWWsrv::Request::new('GET', '/index.html?' + param_data, WWWsrv::HTTP_1_0, @header)
      request.body = @content
      request.host = @host
      request.client = @client
      assert_equal('first', request.params['double', 0])
      assert_equal('second', request.params['double', 1])
      assert_equal('true', request.params['debug', 0])

      @header.content_type = 'Application/X-WWW-Form-URLencoded'
      request = WWWsrv::Request::new('POST', '/index.html', WWWsrv::HTTP_1_0, @header)
      request.body = PseudoIO::new
      request.body.write(param_data)
      request.host = @host
      request.client = @client
      assert_equal('first', request.params['double', 0])
      assert_equal('second', request.params['double', 1])
      assert_equal('true', request.params['debug', 0])
    end

    def test_form_data
      assert(@request.form_data.empty?)

      param_data = [
	%w[ double first ],
	%w[ double second ],
	%w[ debug true ]
      ].map{ |param_name, param_value|
	WWWsrv::HTMLconv::encodeURL(param_name) + '=' +
	  WWWsrv::HTMLconv::encodeURL(param_value)
      }.join('&')

      request = WWWsrv::Request::new('GET', '/index.html?' + param_data, WWWsrv::HTTP_1_1, @header)
      request.body = @content
      request.host = @host
      request.client = @client
      assert_equal('form-data; name=double', request.form_data.head('double', 0).content_disposition)
      assert_equal('first', request.form_data['double', 0])
      assert_equal('form-data; name=double', request.form_data.head('double', 1).content_disposition)
      assert_equal('second', request.form_data['double', 1])
      assert_equal('form-data; name=debug', request.form_data.head('debug').content_disposition)
      assert_equal('true', request.form_data['debug'])

      @header.content_type = 'Application/X-WWW-Form-URLencoded'
      request = WWWsrv::Request::new('POST', '/index.html', WWWsrv::HTTP_1_1, @header)
      request.body = PseudoIO::new
      request.body.write(param_data)
      request.host = @host
      request.client = @client
      assert_equal('form-data; name=double', request.form_data.head('double', 0).content_disposition)
      assert_equal('first', request.form_data['double', 0])
      assert_equal('form-data; name=double', request.form_data.head('double', 1).content_disposition)
      assert_equal('second', request.form_data['double', 1])
      assert_equal('form-data; name=debug', request.form_data.head('debug').content_disposition)
      assert_equal('true', request.form_data['debug'])

      @header.content_type = 'multipart/form-data; boundary=AaB03x'
      content_reader = PseudoIO.new
      content_reader.print
      content_reader.print "--AaB03x\r\n"
      content_reader.print 'Content-Disposition: form-data; name="submit"' + "\r\n"
      content_reader.print "\r\n"
      content_reader.print "OK\r\n"
      content_reader.print "--AaB03x\r\n"
      content_reader.print 'Content-Disposition: form-data; name="date"' + "\r\n"
      content_reader.print "\r\n"
      content_reader.print "1975-11-19\r\n"
      content_reader.print "--AaB03x\r\n"
      content_reader.print 'Content-Disposition: form-data; name="files"' + "\r\n"
      content_reader.print 'Content-Type: multipart/mixed; boundary="Bb:C04y"' + "\r\n"
      content_reader.print "\r\n"
      content_reader.print "--Bb:C04y\r\n"
      content_reader.print 'Content-Disposition: file; filename="hello.html"' + "\r\n"
      content_reader.print "Content-Type: text/html\r\n"
      content_reader.print "\r\n"
      content_reader.print "<html>\r\n"
      content_reader.print "<body><p>Hello world.</p></body>\r\n"
      content_reader.print "</html>\r\n"
      content_reader.print "\r\n"
      content_reader.print "--Bb:C04y\r\n"
      content_reader.print 'Content-Disposition: file; filename="hello.css"' + "\r\n"
      content_reader.print "Content-Type: text/css\r\n"
      content_reader.print "\r\n"
      content_reader.print "p { font-weight: bold; font-size: large }\r\n"
      content_reader.print "\r\n"
      content_reader.print "--Bb:C04y--\r\n"
      content_reader.print "--AaB03x--\r\n"

      request = WWWsrv::Request::new('POST', '/index.html', WWWsrv::HTTP_1_1, @header)
      request.body = content_reader
      request.host = @host
      request.client = @client

      assert_equal('form-data; name="submit"', request.form_data.head('submit').content_disposition)
      assert_equal('OK', request.form_data['submit'])

      assert_equal('form-data; name="date"', request.form_data.head('date').content_disposition)
      assert_equal('1975-11-19', request.form_data['date'])

      assert_equal('form-data; name="files"', request.form_data.head('files', 0).content_disposition)
      assert_equal('multipart/mixed; boundary="Bb:C04y"', request.form_data.head('files', 0).content_type)

      assert_equal('file; filename="hello.html"', request.form_data.body_head('files', 0, 0).content_disposition)
      assert_equal('text/html', request.form_data.body_head('files', 0, 0).content_type)
      assert_equal("<html>\r\n<body><p>Hello world.</p></body>\r\n</html>\r\n", request.form_data['files', 0, 0])

      assert_equal('file; filename="hello.css"', request.form_data.body_head('files', 0, 1).content_disposition)
      assert_equal('text/css', request.form_data.body_head('files', 0, 1).content_type)
      assert_equal("p { font-weight: bold; font-size: large }\r\n", request.form_data['files', 0, 1])

      assert_nil(request.body.read)
    end

    def test_form_data_with_block
      @request.form_data do |tag, value|
	assert_fail('empty content is decoded.')
      end

      param_data = [
	%w[ double first ],
	%w[ double second ],
	%w[ debug true ]
      ].map{ |param_name, param_value|
	WWWsrv::HTMLconv::encodeURL(param_name) + '=' +
	  WWWsrv::HTMLconv::encodeURL(param_value)
      }.join('&')

      form_data_parsed_alist = [
	[ :form_data_name, 'double' ],
	[ :form_data_head, WWWsrv::Header.build('Content-Disposition' => 'form-data; name=double') ],
	[ :form_data_body, 'first' ],
	[ :form_data_end, nil ],
	# -
	[ :form_data_name, 'double' ],
	[ :form_data_head, WWWsrv::Header.build('Content-Disposition' => 'form-data; name=double') ],
	[ :form_data_body, 'second' ],
	[ :form_data_end, nil ],
	# -
	[ :form_data_name, 'debug' ],
	[ :form_data_head, WWWsrv::Header.build('Content-Disposition' => 'form-data; name=debug') ],
	[ :form_data_body, 'true' ],
	[ :form_data_end, nil ]
      ]

      @header.content_type = 'Application/X-WWW-Form-URLencoded'
      request = WWWsrv::Request::new('POST', '/index.html', WWWsrv::HTTP_1_1, @header)
      request.body = PseudoIO::new
      request.body.write(param_data)
      request.host = @host
      request.client = @client

      form_data_scan_alist = Array.new
      request.form_data do |tag, value|
	form_data_scan_alist.push([ tag, value ])
      end
      assert_equal(form_data_parsed_alist, form_data_scan_alist)

      @header.content_type = 'multipart/form-data; boundary=AaB03x'
      content_reader = PseudoIO.new
      content_reader.print "--AaB03x\r\n"
      content_reader.print 'Content-Disposition: form-data; name="submit"' + "\r\n"
      content_reader.print "\r\n"
      content_reader.print "OK\r\n"
      content_reader.print "--AaB03x\r\n"
      content_reader.print 'Content-Disposition: form-data; name="date"' + "\r\n"
      content_reader.print "\r\n"
      content_reader.print "1975-11-19\r\n"
      content_reader.print "--AaB03x\r\n"
      content_reader.print 'Content-Disposition: form-data; name="files"' + "\r\n"
      content_reader.print 'Content-Type: multipart/mixed; boundary="Bb:C04y"' + "\r\n"
      content_reader.print "\r\n"
      content_reader.print "--Bb:C04y\r\n"
      content_reader.print 'Content-Disposition: file; filename="hello.html"' + "\r\n"
      content_reader.print "Content-Type: text/html\r\n"
      content_reader.print "\r\n"
      content_reader.print "<html>\r\n"
      content_reader.print "<body><p>Hello world.</p></body>\r\n"
      content_reader.print "</html>\r\n"
      content_reader.print "\r\n"
      content_reader.print "--Bb:C04y\r\n"
      content_reader.print 'Content-Disposition: file; filename="hello.css"' + "\r\n"
      content_reader.print "Content-Type: text/css\r\n"
      content_reader.print "\r\n"
      content_reader.print "p { font-weight: bold; font-size: large }\r\n"
      content_reader.print "\r\n"
      content_reader.print "--Bb:C04y--\r\n"
      content_reader.print "--AaB03x--\r\n"

      form_data_parsed_alist = [
	[ :form_data_name, 'submit' ],
	[ :form_data_head, WWWsrv::Header.build('Content-Disposition' => 'form-data; name="submit"') ],
	[ :form_data_body, 'OK' ],
	[ :form_data_end, nil ],
	# -
	[ :form_data_name, 'date' ],
	[ :form_data_head, WWWsrv::Header.build('Content-Disposition' => 'form-data; name="date"') ],
	[ :form_data_body, '1975-11-19' ],
	[ :form_data_end, nil ],
	# -
	[ :form_data_name, 'files' ],
	[ :form_data_head, WWWsrv::Header.build(
            'Content-Disposition' => 'form-data; name="files"',
            'Content-Type' => 'multipart/mixed; boundary="Bb:C04y"'
          ),
	],
	[ :begin_mixed, nil ],
	# --
	[ :form_data_file, 'hello.html' ],
	[ :form_data_head, WWWsrv::Header.build(
            'Content-Disposition' => 'file; filename="hello.html"',
            'Content-Type' => 'text/html'
          ),
	],
	[ :form_data_body, "<html>\r\n" ],
	[ :form_data_body, "<body><p>Hello world.</p></body>\r\n" ],
	[ :form_data_body, "</html>\r\n" ],
	[ :form_data_end, nil ],
	# --
	[ :form_data_file, 'hello.css' ],
	[ :form_data_head, WWWsrv::Header.build(
            'Content-Disposition' => 'file; filename="hello.css"',
            'Content-Type' => 'text/css'
          )
        ],
	[ :form_data_body, "p { font-weight: bold; font-size: large }\r\n" ],
	[ :form_data_end, nil ],
	# --
	[ :end_mixed, nil ]
      ]

      request = WWWsrv::Request::new('POST', '/index.html', WWWsrv::HTTP_1_1, @header)
      request.body = content_reader
      request.host = @host
      request.client = @client

      form_data_scan_alist = Array.new
      request.form_data do |tag, value|
	form_data_scan_alist.push([ tag, value ])
      end
      assert_equal(form_data_parsed_alist, form_data_scan_alist)
    end

    def test_cgi_var
      request_content = "Hello world.\n"
      @request.method = 'GET'
      @request.path = '/test.cgi?1+2+3+%7E'
      @request.body.write(request_content)
      @request.header.content_type = 'text/html'
      @request.header.content_length =  request_content.length.to_s
      cgi_var = @request.cgi_var
      assert_equal('CGI/1.1', cgi_var['GATEWAY_INTERFACE'])
      assert_equal('GET', cgi_var['REQUEST_METHOD'])
      assert_equal('1+2+3+%7E', cgi_var['QUERY_STRING'])
      assert_equal('127.0.0.1', cgi_var['REMOTE_ADDR'])
      assert_equal('localhost', cgi_var['REMOTE_HOST'])
      assert_equal('/test.cgi', cgi_var['SCRIPT_NAME'])
      assert_equal('localhost', cgi_var['SERVER_NAME'])
      assert_equal('80', cgi_var['SERVER_PORT'])
      assert_equal('HTTP/1.0', cgi_var['SERVER_PROTOCOL'])
      assert_equal(WWWsrv::product_token, cgi_var['SERVER_SOFTWARE'])
      assert_equal('text/html', cgi_var['CONTENT_TYPE'])
      assert_equal(request_content.length.to_s, cgi_var['CONTENT_LENGTH'])
      assert_equal('Test/0.9 (with RUNIT)', cgi_var['HTTP_USER_AGENT'])
    end
  end

  class TestRequestParse < RUNIT::TestCase
    def setup
      @input = PseudoIO::new
    end

    def test_parse_request_line
      @input.write("GET / HTTP/1.0\r\n")
      @input.write("User-Agent: Test/0.9 (with RUNIT)\r\n")
      request_method, request_path, http_version = WWWsrv::Request::parse_line(@input)
      assert_equal("GET", request_method)
      assert_equal("/", request_path)
      assert_equal(WWWsrv::HTTP_1_0, http_version)
    end

    def test_parse_request_line_robustness
      @input.write("\r\n")
      @input.write("GET / HTTP/1.0\r\n")
      @input.write("Server: WWWsrv/0.0 (TEST)\r\n")
      request_method, request_path, http_version = WWWsrv::Request::parse_line(@input)
      assert_equal("GET", request_method)
      assert_equal("/", request_path)
      assert_equal(WWWsrv::HTTP_1_0, http_version)
    end

    def test_parse_request_line_new_line_sep
      @input.write("GET / HTTP/1.0\n")
      @input.write("User-Agent: Test/0.9 (with RUNIT)\n")
      request_method, request_path, http_version = WWWsrv::Request::parse_line(@input)
      assert_equal("GET", request_method)
      assert_equal("/", request_path)
      assert_equal(WWWsrv::HTTP_1_0, http_version)
    end

    def test_parse_request_line_old_http
      @input.write("GET /\r\n")
      @input.write("Content body...")
      request_method, request_path, http_version = WWWsrv::Request::parse_line(@input)
      assert_equal("GET", request_method)
      assert_equal("/", request_path)
      assert_equal(WWWsrv::HTTP_0_9, http_version)
    end

    def test_parse_request_line_invalid
      @input.write("DETARAME!\r\n")
      @input.write("User-Agent: Test/0.9 (with RUNIT)\r\n")
      assert_exception(WWWsrv::RequestLineParseError) { WWWsrv::Request::parse_line(@input) }
    end

    def test_parse_request_line_invalid_version
      @input.write("GET / HTTP/7.7.7\r\n")
      @input.write("User-Agent: Test/0.9 (with RUNIT)\r\n")
      assert_exception(WWWsrv::HTTPVersionFormatError) { WWWsrv::Request::parse_line(@input) }
    end

    def test_parse_request_line_empty
      request_method, request_path, http_version = WWWsrv::Request::parse_line(@input)
      assert_nil(request_method)
      assert_nil(request_path)
      assert_nil(http_version)
    end

    def test_parse_request
      @input.write("GET / HTTP/1.0\r\n")
      @input.write("Host: localhost\r\n")
      @input.write("User-Agent: Test/0.9 (with RUNIT)\r\n")
      @input.write("\r\n")
      @input.write("...body...")
      request = WWWsrv::Request::parse(@input)
      assert_instance_of(WWWsrv::Request, request)
      assert_equal('GET', request.method)
      assert_equal('/', request.path)
      assert_equal(WWWsrv::HTTP_1_0, request.version)
      assert_instance_of(WWWsrv::Header, request.header)
      assert_equal(2, request.header.length)
      assert_equal('localhost', request.header.host)
      assert_equal('Test/0.9 (with RUNIT)', request.header.user_agent)
    end
  end
end

if ($0 == __FILE__) then
  if (ARGV.empty?) then
    TestWWWsrv::constants.map{ |name|
      TestWWWsrv::const_get(name)
    }.select{ |const|
      (const.kind_of? Class) && (const < RUNIT::TestCase)
    }.each do |test_case|
      RUNIT::CUI::TestRunner::run(test_case::suite)
    end
  else
    for class_name in ARGV
      test_case = TestWWWsrv.const_get("Test#{class_name}")
      RUNIT::CUI::TestRunner.run(test_case.suite)
    end
  end
end
