#!/usr/local/bin/ruby
# $Id: wwwsrv.rb,v 1.503.2.3 2003/04/11 14:08:53 toki Exp $
# -
# World Wide Web Server

=begin
Copyright (c) 2001-2002
	TOKI Yoshinori <toki@freedom.ne.jp>. All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  1.  Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.

  2.  Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation
      and/or other materials provided with the distribution.

  3.  The name of the author may not be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
OF SUCH DAMAGE.
=end

require 'socket'
require 'getoptlong'
require 'wwwsrv/server'
require 'wwwsrv/obsolete' unless $WWWSRV_NOT_COMPAT

autoload :Etc, 'etc'

module WWWsrv
  # latest CVS commit date of core libraries for wwwsrv
  update(%q$Date: 2003/04/11 14:08:53 $)

  class Daemon
    def initialize
      @option = {}
      @server = nil
      @srv_sock = nil
      @access_control_list = []
      @alias_list = []
      @document_list = []
      @config_path = nil
      @base_dir = nil
      @prefork = false
      @host_conf = ''
      @serv_conf = ''
      @init_conf = ''
      @final_conf = ''
      @warm_up_conf = ''
      @cool_down_conf = ''
      @host_mod = nil
      @serv_mod = nil
      @init_mod = nil
      @final_mod = nil
      @warm_up_mod = nil
      @cool_down_mod = nil
      @run_sem = Semaphore::new
    end

    attr_accessor :config_path
    attr_accessor :base_dir

    DEF_OPT = {
      :hostname      => 'localhost',
      :port          => 8888,
      :strict        => false,
      :umask         => nil,
      :log_file      => nil,
      :logging_level => Logger::MESSAGE,
      :display_level => Logger::MESSAGE,
      :request_line_logging_priority     => nil,
      :request_header_logging_priority   => nil,
      :request_body_logging_priority     => nil,
      :request_trailer_logging_priority  => nil,
      :status_line_logging_priority      => nil,
      :response_header_logging_priority  => nil,
      :response_body_logging_priority    => nil,
      :response_trailer_logging_priority => nil,
      :effective_user_id  => 'nobody',
      :effective_group_id => 'nobody',
      :timeout       => TIMEOUT,
      :keep_alive    => KEEP_ALIVE,
      :connections   => MAX_CONNECTIONS,
      :max_requests  => MAX_REQUESTS,
      :write_buffer  => BUFFER_SIZE,
      :prefork_processes => 0,
      :quiet         => false,
      :debug         => false
    }

    CONST_OPT = {}
    DEF_OPT.each_key do |opt_id|
      opt_name = opt_id.id2name
      CONST_OPT[opt_id] = opt_name.upcase.intern

      eval <<-EOF
        def #{opt_name}
	  @option[:#{opt_name}]
	end

	def #{opt_name}=(opt_value)
	  @option[:#{opt_name}] = opt_value
	end
      EOF
    end

    def getopt(opt_id, opt_mod=nil)
      if (@option.include? opt_id) then
	return @option[opt_id]
      elsif (opt_mod && (opt_mod.const_defined? CONST_OPT[opt_id])) then
	return opt_mod.const_get(CONST_OPT[opt_id])
      else
	return DEF_OPT[opt_id]
      end
    end
    private :getopt

    def add_access_control(action, filter)
      @access_control_list.push([ action, filter ])
    end

    def add_alias(dst_path, src_path)
      @alias_list.push([ dst_path, src_path ])
    end

    def add_document(&doc_maker)
      @document_list.push(doc_maker)
      self
    end

    def load_conf
      @host_conf = ''
      @serv_conf = ''
      @init_conf = ''
      @final_conf = ''
      @warm_up_conf = ''
      @cool_down_conf = ''
      state = :head
      File::open(@config_path) do |input|
	for line in input
	  case (line)
	  when /^\s*__HOST__\s*$/n
	    state = :host
	  when /^\s*__SERVER__\s*$/n
	    state = :server
	  when /^\s*__INIT__\s*/n
	    state = :init
	  when /^\s*__FINAL__\s*/n
	    state = :final
	  when /^\s*__WARM_UP__\s*$/n
	    state = :warm_up
	  when /^\s*__COOL_DOWN__\s*$/n
	    state = :cool_down
	  else
	    case (state)
	    when :host
	      @host_conf.concat(line)
	    when :server
	      @serv_conf.concat(line)
	    when :init
	      @init_conf.concat(line)
	    when :final
	      @final_conf.concat(line)
	    when :warm_up
	      @warm_up_conf.concat(line)
	    when :cool_down
	      @cool_down_conf.concat(line)
	    end
	  end
	end
      end
    end
    private :load_conf

    def eval_conf(sect, mod, conf)
      begin
	PROCESS_LOCK.synchronize do
	  mod.module_eval(conf)
	end
      rescue StandardError, ScriptError
	unless (STDERR.closed?) then
	  STDERR.print("#{sect} section error: #{$!}\n")
	end
      end
    end
    private :eval_conf

    def copy_const(mod_dst, mod_src)
      mod_dst.module_eval{ include mod_src } if mod_src
    end
    private :copy_const

    def init
      load_conf if @config_path

      unless (@init_mod) then
	@init_mod = Module::new
	@init_mod.const_set(:BASE_DIR, @base_dir)
	eval_conf('__INIT__', @init_mod, @init_conf)
      end

      @warm_up_mod = Module::new
      copy_const(@warm_up_mod, @init_mod)
      eval_conf('__WARM_UP__', @warm_up_mod, @warm_up_conf)

      @host_mod = Module::new
      copy_const(@host_mod, @warm_up_mod)
      eval_conf('__HOST__', @host_mod, @host_conf)

      if (getopt(:umask, @host_mod)) then
	File::umask(getopt(:umask, @host_mod))
      end

      @scheme = HTTP_SCHEME
      if (@host_mod.const_defined? :SERVER_SOCKET) then
	@srv_sock = @host_mod.const_get(:SERVER_SOCKET)
	@scheme = @host_mod.const_get(:SCHEME) if (@host_mod.const_defined? :SCHEME)
      elsif (! @srv_sock) then
	if (getopt(:strict, @host_mod)) then
	  @srv_sock = TCPserver::open(getopt(:hostname, @host_mod), getopt(:port, @host_mod))
	else
	  @srv_sock = TCPserver::open(getopt(:port, @host_mod))
	end
      end

      if (getopt(:debug, @host_mod)) then
	alert = DebugAlertDocument::new
      elsif (@host_mod.const_defined? :ALERT) then
	alert = @host_mod.const_get(:ALERT)
      elsif (@host_mod.const_defined? :ALARM) then
	alert = @host_mod.const_get(:ALARM) # for backward compatibility
      else
	alert = AlertDocument::new
      end

      @server = Server::new(@srv_sock, alert)
      @server.scheme = @scheme
      if (getopt(:hostname, @host_mod) != DEF_OPT[:hostname]) then
	@server.hostname = getopt(:hostname, @host_mod)
      end

      if (Process::euid == 0) then
	begin
	  if (euid = getopt(:effective_user_id, @host_mod)) then
	    unless (euid.kind_of? Numeric) then
	      if (euid =~ /^\d+$/n) then
		euid = euid.to_i
	      else
		euid = Etc::getpwnam(euid).uid
	      end
	    end
	  end

	  if (egid = getopt(:effective_group_id, @host_mod)) then
	    unless (egid.kind_of? Numeric) then
	      if (egid =~ /^\d+$/n) then
		egid = egid.to_i
	      else
		egid = Etc::getgrnam(egid).gid
	      end
	    end
	  end

	  Process.gid = egid if (Process.gid != egid)
	  Process.uid = euid if (Process.uid != euid)
	  Process.egid = egid if (Process.egid != egid)
	  Process.euid = euid if (Process.euid != euid)
	rescue LoadError
	  STDERR.print "warning: #{$!}\n"
	end
      end

      if (@host_mod.const_defined? :LOGGER) then
	@server.add_logger(@host_mod.const_get(:LOGGER))
      end
      if (log_file = getopt(:log_file, @host_mod)) then
	output = File::open(log_file, 'a')
	output.sync = true
	@server.add_logger(Logger::new(output, getopt(:logging_level, @host_mod)))
      end
      if (! getopt(:quiet, @host_mod) && ! STDOUT.closed?) then
	STDOUT.sync = true
	@server.add_logger(PreservedLogger::new(STDOUT, getopt(:display_level, @host_mod)))
      end

      for action, filter in @access_control_list
	@server.add_access_control(action, true, &filter)
      end

      @serv_mod = Module::new
      copy_const(@serv_mod, @host_mod)
      @serv_mod.const_set(:SERVER, @server)
      eval_conf('__SERVER__', @serv_mod, @serv_conf)

      # for backward compatibility
      unless (@host_mod.const_defined? :REQUEST_BODY_LOGGING_PRIORITY) then
	if (@host_mod.const_defined? :REQUEST_CONTENT_LOGGING_PRIORITY) then
	  @server.request_logger.request_body_logging_priority =
	    @host_mod.const_get(:REQUEST_CONTENT_LOGGING_PRIORITY)
	end
      end

      # for backward compatibility
      unless (@host_mod.const_defined? :RESPONSE_BODY_LOGGING_PRIORITY) then
	if (@host_mod.const_defined? :RESPONSE_CONTENT_LOGGING_PRIORITY) then
	  @server.response_logger.response_body_logging_priority =
	    @host_mod.const_get(:RESPONSE_CONTENT_LOGGING_PRIORITY)
	end
      end

      [ [ @server, :timeout=, :timeout ],
	[ @server, :keep_alive=, :keep_alive ],
	[ @server, :connections=, :connections ],
	[ @server, :max_requests=, :max_requests ],
	[ @server, :buffer_size=, :write_buffer ],
	[ @server.request_logger, :request_line_logging_priority=, :request_line_logging_priority ],
	[ @server.request_logger, :request_header_logging_priority=, :request_header_logging_priority ],
	[ @server.request_logger, :request_body_logging_priority=, :request_body_logging_priority ],
	[ @server.request_logger, :request_trailer_logging_priority=, :request_trailer_logging_priority ],
	[ @server.response_logger, :status_line_logging_priority=, :status_line_logging_priority ],
	[ @server.response_logger, :response_header_logging_priority=, :response_header_logging_priority ],
	[ @server.response_logger, :response_body_logging_priority=, :response_body_logging_priority ],
	[ @server.response_logger, :response_trailer_logging_priority=, :response_trailer_logging_priority ]
      ].each do |receiver, opt_set, option|
	if (getopt(option)) then
	  receiver.send(opt_set, getopt(option))
	end
      end

      for dst_path, src_path in @alias_list
	@server.alias(dst_path, src_path)
      end

      for doc_maker in @document_list
	doc_maker.call(@server)
      end

      self
    end
    private :init

    def cool_down
      @cool_down_mod = Module::new
      copy_const(@cool_down_mod, @serv_mod)
      eval_conf('__COOL_DOWN__', @cool_down_mod, @cool_down_conf)

      self
    end
    private :cool_down

    def final
      @final_mod = Module::new
      copy_const(@final_mod, @cool_down_mod)
      eval_conf('__FINAL__', @final_mod, @final_conf)

      self
    end
    private :final

    def run
      @run_sem.synchronize do
	begin
	  loop do
	    begin
	      init
	      unless (@prefork) then
		@prefork = true
		getopt(:prefork_processes).times do
		  fork or break
		end
	      end
	      begin
		@run_sem.post
		@server.accept
	      ensure
		@run_sem.wait
	      end
	    ensure
	      cool_down
	    end
	  end
	ensure
	  final
	end
      end

      self
    end

    def close
      @run_sem.synchronize do
	if (@server) then
	  @server.close
	  return self
	else
	  return nil
	end
      end
    end
  end

  class << Daemon
    def start_daemon_process
      catch(:START_DAEMON) do
	unless (fork) then
	  Process::setsid
	  unless (fork) then
	    Dir::chdir('/')
	    STDIN.reopen('/dev/null', 'r')
	    STDOUT.reopen('/dev/null', 'w')
	    STDERR.reopen('/dev/null', 'w')
	    throw :START_DAEMON
	  end
	end
	exit!
      end
    end
    private :start_daemon_process

    PATH_SEP = File::PATH_SEPARATOR
    PATH_SEP_PATTERN = Regexp::compile(Regexp::quote(PATH_SEP))

    LOGGING_PRIORITY = {
      'internal' => Logger::INTERNAL,
      'debug'    => Logger::DEBUG,
      'system'   => Logger::SYSTEM,
      'message'  => Logger::MESSAGE,
      'warning'  => Logger::WARNING,
      'notice'   => Logger::NOTICE,
      'alert'    => Logger::ALERT,
      'error'    => Logger::ERROR,
      # for backward compatibility
      'raw'      => Logger::INTERNAL,
      'discreet' => Logger::SYSTEM
    }
    LOGGING_PRIORITY_TYPES = '[internal, debug, system, message, warning, notice, alert, error]'

    def parse_logging_priority(prio_type)
      if (LOGGING_PRIORITY.include? prio_type.downcase) then
	return LOGGING_PRIORITY[prio_type.downcase]
      else
	raise "unknown logging priority type: #{prio_type.inspect}"
      end
    end
    private :parse_logging_priority

    def parse_optional_logging_priority(prio_type)
      if (prio_type && ! prio_type.empty?) then
	return parse_logging_priority(prio_type)
      else
	return WWWsrv::Logger::MESSAGE
      end
    end
    private :parse_optional_logging_priority

    def add_access_control(daemon, action, attr_name, pattern)
      if (pattern !~ /^!/) then
	regexp_pattern = Regexp::compile(pattern)
	daemon.add_access_control action, proc{ |request|
	  request.client.send(attr_name) =~ regexp_pattern
	}
      else
	pattern.sub!(/^!/, '')
	regexp_pattern = Regexp::compile(pattern)
	daemon.add_access_control action, proc{ |request|
	  request.client.send(attr_name) !~ regexp_pattern
	}
      end
    end
    private :add_access_control

    def split_page(path)
      path = path.sub(%r"[^/]*$", '')
      page = $&
      if (path != '/') then
	path.sub!(%r"/$", '')
      end
      return path, page
    end
    private :split_page

    def start
      option = {
	:daemon       => false,
	:version      => false,
	:catch_sighup => false,
	:help         => false
      }
      opt_action = []

      opt_action.push [
	[ '-m', '--hostname' ],
	proc{ |daemon, opt_arg|
	  daemon.hostname = opt_arg
	},
	GetoptLong::REQUIRED_ARGUMENT,
	'[HOSTNAME]'
      ]

      opt_action.push [
	[ '-n', '--port' ],
	proc{ |daemon, opt_arg|
	  if (opt_arg =~ /^\d+$/) then
	    daemon.port = opt_arg.to_i
	  else
	    raise "invalid port number: #{opt_arg.inspect}"
	  end
	},
	GetoptLong::REQUIRED_ARGUMENT,
	'[port]'
      ]

      opt_action.push [
	[ '-t', '--strict' ],
	proc{ |daemon, opt_arg|
	  daemon.strict = true
	},
	GetoptLong::NO_ARGUMENT,
      ]

      opt_action.push [
	[ '--not-strict' ],
	proc{ |daemon, opt_arg|
	  daemon.strict = false
	},
	GetoptLong::NO_ARGUMENT
      ]

      opt_action.push [
	[ '-l', '--log-file'],
	proc{ |daemon, opt_arg|
	  daemon.log_file = File::expand_path(opt_arg)
	},
	GetoptLong::REQUIRED_ARGUMENT,
	'[path]'
      ]

      opt_action.push [
	[ '--no-log-file'],
	proc{ |daemon, opt_arg|
	  daemon.log_file = nil
	},
	GetoptLong::NO_ARGUMENT
      ]

      opt_action.push [
	[ '-q', '--quiet' ],
	proc{ |daemon, opt_arg|
	  daemon.quiet = true
	},
	GetoptLong::NO_ARGUMENT
      ]

      opt_action.push [
	[ '--not-quiet' ],
	proc{ |daemon, opt_arg|
	  daemon.quiet = false
	},
	GetoptLong::NO_ARGUMENT
      ]

      opt_action.push [
	[ '-v', '--logging-level' ],
	proc{ |daemon, opt_arg|
	  daemon.logging_level = parse_logging_priority(opt_arg)
	},
	GetoptLong::REQUIRED_ARGUMENT,
	LOGGING_PRIORITY_TYPES
      ]

      opt_action.push [
	[ '-p', '--display-level' ],
	proc{ |daemon, opt_arg|
	  daemon.display_level = parse_logging_priority(opt_arg)
	},
	GetoptLong::REQUIRED_ARGUMENT,
	LOGGING_PRIORITY_TYPES
      ]

      opt_action.push [
	[ '--request-line-logging' ],
	proc{ |daemon, opt_arg|
	  daemon.request_line_logging_priority = parse_optional_logging_priority(opt_arg)
	},
	GetoptLong::OPTIONAL_ARGUMENT,
	LOGGING_PRIORITY_TYPES
      ]

      opt_action.push [
	[ '--request-header-logging' ],
	proc{ |daemon, opt_arg|
	  daemon.request_header_logging_priority = parse_optional_logging_priority(opt_arg)
	},
	GetoptLong::OPTIONAL_ARGUMENT,
	LOGGING_PRIORITY_TYPES
      ]

      opt_action.push [
	[ '--request-body-logging' ],
	proc{ |daemon, opt_arg|
	  daemon.request_body_logging_priority = parse_optional_logging_priority(opt_arg)
	},
	GetoptLong::OPTIONAL_ARGUMENT,
	LOGGING_PRIORITY_TYPES
      ]

      # for backward compatibility
      opt_action.push [
	[ '--request-content-logging' ],
	proc{ |daemon, opt_arg|
	  daemon.request_body_logging_priority = parse_optional_logging_priority(opt_arg)
	},
	GetoptLong::OPTIONAL_ARGUMENT,
	LOGGING_PRIORITY_TYPES,
	:obsolete
      ]

      opt_action.push [
	[ '--request-trailer-logging' ],
	proc{ |daemon, opt_arg|
	  daemon.request_trailer_logging_priority = parse_optional_logging_priority(opt_arg)
	},
	GetoptLong::OPTIONAL_ARGUMENT,
	LOGGING_PRIORITY_TYPES
      ]

      opt_action.push [
	[ '--status-line-logging' ],
	proc{ |daemon, opt_arg|
	  daemon.status_line_logging_priority = parse_optional_logging_priority(opt_arg)
	},
	GetoptLong::OPTIONAL_ARGUMENT,
	LOGGING_PRIORITY_TYPES
      ]

      opt_action.push [
	[ '--response-header-logging' ],
	proc{ |daemon, opt_arg|
	  daemon.response_header_logging_priority = parse_optional_logging_priority(opt_arg)
	},
	GetoptLong::OPTIONAL_ARGUMENT,
	LOGGING_PRIORITY_TYPES
      ]

      opt_action.push [
	[ '--response-body-logging' ],
	proc{ |daemon, opt_arg|
	  daemon.response_body_logging_priority = parse_optional_logging_priority(opt_arg)
	},
	GetoptLong::OPTIONAL_ARGUMENT,
	LOGGING_PRIORITY_TYPES
      ]

      # for backward compatibility
      opt_action.push [
	[ '--response-content-logging' ],
	proc{ |daemon, opt_arg|
	  daemon.response_body_logging_priority = parse_optional_logging_priority(opt_arg)
	},
	GetoptLong::OPTIONAL_ARGUMENT,
	LOGGING_PRIORITY_TYPES,
	:obsolete
      ]

      opt_action.push [
	[ '--response-trailer-logging' ],
	proc{ |daemon, opt_arg|
	  daemon.response_trailer_logging_priority = parse_optional_logging_priority(opt_arg)
	},
	GetoptLong::OPTIONAL_ARGUMENT,
	LOGGING_PRIORITY_TYPES
      ]

      opt_action.push [
	[ '-e', '--effective-user-id' ],
	proc{ |daemon, opt_arg|
	  if (opt_arg =~ /^\d+$/) then
	    daemon.effective_user_id = opt_arg.to_i
	  else
	    daemon.effective_user_id = opt_arg
	  end
	},
	GetoptLong::REQUIRED_ARGUMENT,
	'[euid]'
      ]

      opt_action.push [
	[ '-g', '--effective-group-id' ],
	proc{ |daemon, opt_arg|
	  if (opt_arg =~ /^\d+$/n) then
	    daemon.effective_group_id = opt_arg.to_i
	  else
	    daemon.effective_group_id = opt_arg
	  end
	},
	GetoptLong::REQUIRED_ARGUMENT,
	'[egid]'
      ]

      opt_action.push [
	[ '--timeout' ],
	proc{ |daemon, opt_arg|
	  if (opt_arg =~ /^\d+$/) then
	    daemon.timeout = opt_arg.to_i
	  else
	    raise "invalid timeout: #{timeout.inspect}"
	  end
	},
	GetoptLong::REQUIRED_ARGUMENT,
	'[timeout]'
      ]

      opt_action.push [
	[ '--keep-alive' ],
	proc{ |daemon, opt_arg|
	  if (opt_arg =~ /^\d+$/) then
	    daemon.keep_alive = opt_arg.to_i
	  else
	    raise "invalide keep alive: #{opt_arg.inspect}"
	  end
	},
	GetoptLong::REQUIRED_ARGUMENT,
	'[timeout]'
      ]

      opt_action.push [
	[ '-d', '--debug' ],
	proc{ |daemon, opt_arg|
	  daemon.debug = true
	},
	GetoptLong::NO_ARGUMENT
      ]

      opt_action.push [
	[ '--no-debug' ],
	proc{ |daemon, opt_arg|
	  daemon.debug = false
	},
	GetoptLong::NO_ARGUMENT
      ]

      opt_action.push [
	['-f', '--config'  ],
	proc{ |daemon, opt_arg|
	  daemon.config_path = File::expand_path(opt_arg)
	},
	GetoptLong::REQUIRED_ARGUMENT,
	'[path]'
      ]

      opt_action.push [
	[ '-b', '--base-dir' ],
	proc{ |daemon, opt_arg|
	  daemon.base_dir = opt_arg
	},
	GetoptLong::REQUIRED_ARGUMENT,
	'[directory]'
      ]

      opt_action.push [
	[ '-s', '--daemon' ],
	proc{ |daemon, opt_arg|
	  option[:daemon] = true
	},
	GetoptLong::NO_ARGUMENT
      ]

      opt_action.push [
	[ '--not-daemon'  ],
	proc{ |daemon, opt_arg|
	  option[:daemon] = false
	},
	GetoptLong::NO_ARGUMENT
      ]

      opt_action.push [
	[ '--catch-sighup' ],
	proc{ |daemon, opt_arg|
	  option[:catch_sighup] = true
	},
	GetoptLong::NO_ARGUMENT
      ]

      opt_action.push [
	[ '-u', '--umask' ],
	proc{ |daemon, opt_arg|
	  if (opt_arg =~ /^[0-7]+$/n) then
	    daemon.umask = opt_arg.oct
	  else
	    raise "invalid umask format: #{opt_arg.inspect}"
	  end
	},
	GetoptLong::REQUIRED_ARGUMENT,
	'[umask]'
      ]

      opt_action.push [
	[ '--connections' ],
	proc{ |daemon, opt_arg|
	  if (opt_arg =~ /^\d+$/) then
	    daemon.connections = opt_arg.to_i
	  else
	    raise "invalid connections: #{opt_arg.inspect}"
	  end
	},
	GetoptLong::REQUIRED_ARGUMENT,
	'[max connections]'
      ]

      opt_action.push [
	[ '--max-requests' ],
	proc{ |daemon, opt_arg|
	  if (opt_arg =~ /^\d+$/) then
	    daemon.max_requests = opt_arg.to_i
	  else
	    raise "invalid max requests: #{opt_arg.inspect}"
	  end
	},
	GetoptLong::REQUIRED_ARGUMENT,
	'[max requests]'
      ]

      opt_action.push [
	[ '--write-buffer' ],
	proc{ |daemon, opt_arg|
	  if (opt_arg =~ /^\d+$/) then
	    daemon.write_buffer = opt_arg.to_i
	  else
	    raise "invalid write buffer size: #{opt_arg.inspect}"
	  end
	},
	GetoptLong::REQUIRED_ARGUMENT,
	'[buffer size]',
	:hidden
      ]

      opt_action.push [
	[ '--prefork-processes' ],
	proc{ |daemon, opt_arg|
	  if (opt_arg =~ /^\d+$/) then
	    daemon.prefork_processes = opt_arg.to_i
	  else
	    raise "invalid prefork processes: #{opt_arg.inspect}"
	  end
	},
	GetoptLong::REQUIRED_ARGUMENT,
	'[processes]',
	:hidden
      ]

      opt_action.push [
	[ '--access-allow-name' ],
	proc{ |daemon, opt_arg|
	  add_access_control(daemon, :allow, :name, opt_arg)
	},
	GetoptLong::REQUIRED_ARGUMENT,
	'(!)[regexp]'
      ]

      # for backward compatibility
      opt_action.push [
	[ '--actl-allow-name' ],
	opt_action.last[1],
	opt_action.last[2],
	opt_action.last[3],
	:obsolete
      ]

      opt_action.push [
	[ '--access-deny-name' ],
	proc{ |daemon, opt_arg|
	  add_access_control(daemon, :deny, :name, opt_arg)
	},
	GetoptLong::REQUIRED_ARGUMENT,
	'(!)[regexp]'
      ]

      # for backward compatibility
      opt_action.push [
	[ '--actl-deny-name' ],
	opt_action.last[1],
	opt_action.last[2],
	opt_action.last[3],
	:obsolete
      ]

      opt_action.push [
	[ '--access-allow-address' ],
	proc{ |daemon, opt_arg|
	  add_access_control(daemon, :allow, :address, opt_arg)
	},
	GetoptLong::REQUIRED_ARGUMENT,
	'(!)[regexp]'
      ]

      # for backward compatibility
      opt_action.push [
	[ '--actl-allow-address' ],
	opt_action.last[1],
	opt_action.last[2],
	opt_action.last[3],
	:obsolete
      ]

      opt_action.push [
	[ '--access-deny-address' ],
	proc{ |daemon, opt_arg|
	  add_access_control(daemon, :deny, :address, opt_arg)
	},
	GetoptLong::REQUIRED_ARGUMENT,
	'(!)[regexp]'
      ]

      # for backward compatibility
      opt_action.push [
	[ '--actl-deny-address' ],
	opt_action.last[1],
	opt_action.last[2],
	opt_action.last[3],
	:obsolete
      ]

      opt_action.push [
	[ '-a', '--alias' ],
	proc{ |daemon, opt_arg|
	  dst_path, src_path = opt_arg.split(PATH_SEP_PATTERN, 2)
	  if ((! dst_path || dst_path.empty?) || (! src_path || src_path.empty?)) then
	    raise "invalid alias format: #{opt_arg.inspect}"
	  end
	  daemon.add_alias(dst_path, src_path)
	},
	GetoptLong::REQUIRED_ARGUMENT,
	"[dst. path]#{PATH_SEP}[src. path]"
      ]

      opt_action.push [
	[ '--mount-echo' ],
	proc{ |daemon, opt_arg|
	  require 'wwwsrv/echo'
	  mount_path = opt_arg
	  daemon.add_document{ |server|
	    server.mount(WWWsrv::Echo::new, mount_path)
	  }
	},
	GetoptLong::REQUIRED_ARGUMENT,
	'[HTTP path]'
      ]

      opt_action.push [
	[ '--mount-local' ],
	proc{ |daemon, opt_arg|
	  require 'wwwsrv/local'
	  mount_path, local_path, opt_regexp = opt_arg.split(PATH_SEP_PATTERN, 3)
	  mount_opts = Hash::new
	  if (! mount_path || ! local_path) then
	    raise "invalid local mount format: #{opt_arg.inspect}"
	  end
	  local_path = File::expand_path(local_path)
	  if (File::directory? local_path) then
	    if (opt_regexp && ! opt_regexp.empty?) then
	      mount_opts[:mask] = Regex::compile(opt_regexp, false)
	    end
	  elsif (File::file? local_path) then
	    if (opt_regexp && ! opt_regexp.empty?) then
	      raise "no need regexp option to mount a local file of #{local_path.inspect}: #{opt_regexp.inspect}"
	    end
	    local_path, mount_opts[:page] = File::split(local_path)
	  else
	    raise "not found a local file or directory: #{local_path.inspect}"
	  end
	  daemon.add_document{ |server|
	    local_file_document = WWWsrv::LocalFileDocument::new(local_path)
	    server.mount(local_file_document, mount_path, mount_opts)
	  }
	},
	GetoptLong::REQUIRED_ARGUMENT,
	"[HTTP path]#{PATH_SEP}[local path](#{PATH_SEP}[regexp])"
      ]

      opt_action.push [
	[ '--mount-cgi' ],
	proc{ |daemon, opt_arg|
	  require 'wwwsrv/cgi'
	  mount_path, cgi_path, opt_regexp, cgi_exec = opt_arg.split(PATH_SEP_PATTERN, 4)
	  mount_opts = Hash::new
	  if (! mount_path || ! cgi_path) then
	    raise "invalid cgi mount format: #{opt_arg.inspect}"
	  end
	  cgi_path = File::expand_path(cgi_path)
	  if (cgi_exec && cgi_exec.empty?) then
	    cgi_exec = nil
	  end
	  if (File::directory? cgi_path) then
	    if (opt_regexp && ! opt_regexp.empty?) then
	      mount_opts[:mask] = Regexp::compile(opt_regexp)
	    end
	  elsif (File::file? cgi_path) then
	    if (opt_regexp && ! opt_regexp.empty?) then
	      raise "no need regexp option to mount a local CGI file of #{cgi_path.inspect}: #{opt_regexp.inspect}"
	    end
	    cgi_path, mount_opts[:page] = File::split(cgi_path)
	  else
	    raise "not found a CGI file or directory: #{cgi_path.inspect}"
	  end
	  daemon.add_document{ |server|
	    cgi_document = WWWsrv::CGIdocument::new(cgi_path, false, cgi_exec)
	    server.mount(cgi_document, mount_path, mount_opts)
	  }
	},
	GetoptLong::REQUIRED_ARGUMENT,
	"[HTTP path]#{PATH_SEP}[CGI path](#{PATH_SEP}[regexp](#{PATH_SEP}[interpreter]))"
      ]

      opt_action.push [
	[ '--mount-nph-cgi' ],
	proc{ |daemon, opt_arg|
	  require 'wwwsrv/cgi'
	  mount_path, cgi_path, opt_regexp, cgi_exec = opt_arg.split(PATH_SEP_PATTERN, 4)
	  mount_opts = Hash::new
	  if (! mount_path || ! cgi_path) then
	    raise "invalid cgi mount format: #{opt_arg.inspect}"
	  end
	  cgi_path = File::expand_path(cgi_path)
	  if (cgi_exec && cgi_exec.empty?) then
	    cgi_exec = nil
	  end
	  if (File::directory? cgi_path) then
	    if (opt_regexp && ! opt_regexp.empty?) then
	      mount_opts[:mask] = Regexp::compile(opt_regexp)
	    end
	  elsif (File::file? cgi_path) then
	    if (opt_regexp && ! opt_regexp.empty?) then
	      raise "no need regexp option to mount a local CGI file of #{cgi_path.inspect}: #{opt_regexp.inspect}"
	    end
	    cgi_path, mount_opts[:page] = File::split(cgi_path)
	  else
	    raise "not found a CGI file or directory: #{cgi_path.inspect}"
	  end
	  daemon.add_document{ |server|
	    cgi_document = WWWsrv::CGIdocument::new(cgi_path, true, cgi_exec)
	    server.mount(cgi_document, mount_path, mount_opts)
	  }
	},
	GetoptLong::REQUIRED_ARGUMENT,
	"[HTTP path]#{PATH_SEP}[CGI path](#{PATH_SEP}[regexp](#{PATH_SEP}[interpreter]))"
      ]

      opt_action.push [
	[ '--attach-ssi' ],
	proc{ |daemon, opt_arg|
	  require 'wwwsrv/ssi'
	  mount_path, opt_regexp, ssi_tags = opt_arg.split(PATH_SEP_PATTERN, 3)
	  mount_opts = Hash::new
	  if (opt_regexp && ! opt_regexp.empty?) then
	    mount_opts[:mask] = Regexp::compile(opt_regexp)
	  else
	    mount_path, mount_opts[:page] = split_page(mount_path)
	  end
	  if (ssi_tags && ! ssi_tags.empty?) then
	    ssi_tags = ssi_tags.split(/,/)
	  else
	    ssi_tags = nil
	  end
	  if (ssi_tags) then
	    daemon.add_document{ |server|
	      server.attach(mount_path, mount_opts) { |document|
		WWWsrv::SSIdocument::new(document, ssi_tags)
	      }
	    }
	  else
	    daemon.add_document{ |server|
	      server.attach(mount_path, mount_opts) { |document|
		WWWsrv::SSIdocument::new(document)
	      }
	    }
	  end
	},
	GetoptLong::REQUIRED_ARGUMENT,
	"[HTTP path](#{PATH_SEP}[regexp](#{PATH_SEP}[SSI tags]))"
      ]

      # for backward compatibility
      opt_action.push [
	[ '--mount-ssi' ],
	opt_action.last[1],
	opt_action.last[2],
	opt_action.last[3],
	:obsolete
      ]

      opt_action.push [
	[ '--mount-erb' ],
	proc{ |daemon, opt_arg|
	  require 'wwwsrv/erbdoc'
	  mount_path, local_path, opt_regexp, erb_safe, erb_methods = opt_arg.split(PATH_SEP_PATTERN, 5)
	  mount_opts = Hash::new
	  if (! mount_path || ! local_path) then
	    raise "invalid eRuby mount format: #{opt_arg.inspect}"
	  end
	  local_path = File::expand_path(local_path)
	  if (erb_safe && ! erb_safe.empty?) then
	    if (erb_safe =~ /^\d+$/) then
	      erb_safe = erb_safe.to_i
	    else
	      raise "invalid eRuby safe level format: #{erb_safe.inspect}"
	    end
	  else
	    erb_safe = nil
	  end
	  if (File::directory? local_path) then
	    if (opt_regexp && ! opt_regexp.empty?) then
	      mount_opts[:mask] = Regexp::compile(opt_regexp)
	    end
	  elsif (File::file? local_path) then
	    if (opt_regexp && ! opt_regexp.empty?) then
	      raise "no need regexp option to mount a local eRuby file of #{local_path.inspect}: #{opt_regexp.inspect}"
	    end
	    local_path, mount_opts[:page] = File::split(local_path)
	  else
	    raise "not found a local file or directory: #{local_path.inspect}"
	  end
	  if (erb_methods && ! erb_methods.empty?) then
	    erb_methods = erb_methods.split(/,/n)
	    daemon.add_document{ |server|
	      erb_document = WWWsrv::ERubyDocument::new(local_path, erb_safe, erb_methods)
	      server.mount(erb_document, mount_path, mount_opts)
	    }
	  else
	    daemon.add_document{ |server|
	      erb_document = WWWsrv::ERubyDocument::new(local_path, erb_safe)
	      server.mount(erb_document, mount_path, mount_opts)
	    }
	  end
	},
	GetoptLong::REQUIRED_ARGUMENT,
	"[HTTP path]#{PATH_SEP}[local path](#{PATH_SEP}[regexp](#{PATH_SEP}[sefe level](#{PATH_SEP}[allow methods])))"
      ]

      opt_action.push [
	[ '--attach-erb' ],
	proc{ |daemon, opt_arg|
	  require 'wwwsrv/erbdoc'
	  mount_path, opt_regexp, erb_safe = opt_arg.split(PATH_SEP_PATTERN, 3)
	  mount_opts = Hash::new
	  if (opt_regexp && ! opt_regexp.empty?) then
	    mount_opts[:mask] = Regexp::compile(opt_regexp)
	  else
	    mount_path, mount_opts[:page] = split_page(mount_path)
	  end
	  if (erb_safe && ! erb_safe.empty?) then
	    if (erb_safe =~ /^\d+$/n) then
	      erb_safe = erb_safe.to_i
	    else
	      raise "invalid eRuby safe level format: #{erb_safe.inspect}"
	    end
	  else
	    erb_safe = nil
	  end
	  daemon.add_document{ |server|
	    server.attach(mount_path, mount_opts) { |document|
	      WWWsrv::ERubyFilterDocument::new(document, erb_safe)
	    }
	  }
	},
	GetoptLong::REQUIRED_ARGUMENT,
	"[HTTP path](#{PATH_SEP}[regexp](#{PATH_SEP}[safe level]))"
      ]

      opt_action.push [
	[ '--mount-fcgi' ],
	proc{ |daemon, opt_arg|
	  require 'wwwsrv/fastcgi'
	  mount_path, opt_regexp, sock_type, sock_path = opt_arg.split(/:/, 4)
	  mount_opts = Hash::new
	  case (sock_type.downcase)
	  when 'unix'
	    sock_make = proc{
	      UNIXsocket::open(sock_path)
	    }
	  when 'inet'
	    hostname, port = sock_path.split(/:/, 2)
	    if (port =~ /^\d+$/) then
	      port = port.to_i
	    else
	      raise "invalid port number format: #{sock_path.inspect}"
	    end
	    sock_make = proc{
	      TCPsocket::open(hostname, port)
	    }
	  else
	    raise "unknown FastCGI socket type: #{sock_type.inspect}"
	  end
	  if (opt_regexp && ! opt_regexp.empty?) then
	    mount_opts[:mask] = Regexp::compile(opt_regexp)
	  end
	  daemon.add_document{ |server|
	    fcgi_document = FastCGIDocument::new(&sock_make)
	    server.mount(fcgi_document, mount_path, mount_opts)
	  }
	},
	GetoptLong::REQUIRED_ARGUMENT,
	"[HTTP path]#{PATH_SEP}([regexp])#{PATH_SEP}[unix or inet]#{PATH_SEP}[socket path or hostname:port]"
      ]

      opt_action.push [
	[ '--attach-filter' ],
	proc{ |daemon, opt_arg|
	  raise NotImplementError, 'sorry...'
	},
	GetoptLong::REQUIRED_ARGUMENT,
	"[HTTP path]#{PATH_SEP}/regexp 1/replace 1/[gixm]#{PATH_SEP}/regexp 2/replace 2/[gixm]#{PATH_SEP}...",
	:hidden
      ]

      opt_action.push [
	[ '-h', '--help' ],
	proc{ |daemon, opt_arg|
	  option[:help] = true
	},
	GetoptLong::NO_ARGUMENT
      ]

      opt_action.push [
	[ '--version' ],
	proc{ |daemon, opt_arg|
	  option[:version] = true
	},
	GetoptLong::NO_ARGUMENT
      ]

      opt_proc_map = {}
      opt_obs_map = {}
      for opt_list, opt_proc, opt_type, opt_arg_desc, category in opt_action
	for opt_name in opt_list
	  opt_proc_map[opt_name] = opt_proc
	  opt_obs_map[opt_name] = category == :obsolete
	end
      end

      getoptlong = GetoptLong::new
      getoptlong.set_options *opt_action.map{ |opt_list, opt_proc, opt_type, opt_arg_desc, category|
	opt_list + [ opt_type ]
      }

      daemon = Daemon::new
      for opt_name, opt_arg in getoptlong
	if (opt_proc = opt_proc_map[opt_name]) then
	  if (opt_obs_map[opt_name]) then
	    STDERR.print "warning: option #{opt_name.inspect} is obsolete!\n"
	  end
	  opt_proc.call(daemon, opt_arg)
	end
      end
      daemon.base_dir = Dir::getwd unless daemon.base_dir

      if (! ARGV.empty? || option[:version] || option[:help]) then
	if (! ARGV.empty?) then
	  print 'error: invalid arguments: ', ARGV.map{ |arg|
	    arg.inspect
	  }.join(', '), "\n"
	end
	if (option[:version]) then
	  print WWWsrv::SRVNAME, ' version ', WWWsrv::VERSION
	  print ' (', WWWsrv::CODENAME, '; ', WWWsrv::rcs_date, ")\n"
	end
	if (option[:help]) then
	  print "Usage: #{File::basename($0)} [options]\n"
	  print "Options:\n"
	  for opt_list, opt_proc, opt_type, opt_arg_desc, category in opt_action
	    (category == :obsolete) and next
	    (category == :hidden) and next
	    print '  '
	    print '    ' if (opt_list.first !~ /^-[^-]$/)
	    print opt_list.join(', ')
	    case (opt_type)
	    when GetoptLong::REQUIRED_ARGUMENT
	      print '=', opt_arg_desc
	    when GetoptLong::OPTIONAL_ARGUMENT
	      print '(=', opt_arg_desc, ')'
	    end
	    print "\n"
	  end
	end
	exit
      end

      if (option[:daemon]) then
	daemon.quiet = true
	start_daemon_process
      end

      if (option[:daemon] || option[:catch_sighup]) then
	reload = Queue::new
	trap('HUP') {
	  reload.push(true)
	}
	Thread::new{
	  Thread::current.priority = SYSTEM_PRIORITY
	  loop do
	    reload.pop
	    daemon.close
	  end
	}
      end

      # for handling final hooks of mounted documents
      trap('TERM') {
	raise SignalException, 'SIGTERM'
      }

      if (daemon.debug) then
	Thread.abort_on_exception = true
      end

      daemon.run
    end
  end
end

if ($0 == __FILE__) then
  WWWsrv::Daemon::start
end
