# $Id: cache.rb,v 1.13.2.1 2002/12/27 15:18:09 toki Exp $
# -
# object cache

=begin
Copyright (c) 2001-2002
	TOKI Yoshinori <toki@freedom.ne.jp>. All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  1.  Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.

  2.  Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation
      and/or other materials provided with the distribution.

  3.  The name of the author may not be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
OF SUCH DAMAGE.
=end

require 'thread'
require 'wwwsrv/const'

unless (defined? $WWWSRV_NO_OBJECT_CACHE) then
  $WWWSRV_NO_OBJECT_CACHE = false
end

module WWWsrv
  # latest CVS commit date of core libraries for wwwsrv
  update(%q$Date: 2002/12/27 15:18:09 $)

  # NOTE: cacheable object requirements
  #   * object MUST be immutable.
  class ImmutableObjectCache
    def initialize(cache_limit=1000, &block)
      @builder = block
      @cache_map = Hash::new
      @cache_lock = Mutex::new
      @cache_limit = cache_limit
    end

    attr_reader :cache_limit

    def clear
      @cache_lock.synchronize do
	@cache_map.clear
      end
      nil
    end

    def fetch(*args)
      # Double checked locking algorithm.
      # This algorithm would not work well if ruby thread is native.
      unless (obj = @cache_map[args]) then
	@cache_lock.synchronize do
	  unless (obj = @cache_map[args]) then
	    obj = @builder.call(*args)
	    if (@cache_limit && @cache_map.size >= @cache_limit) then
	      # defence of cache overflow
	      @cache_map.clear
	    end
	    @cache_map[args] = obj
	  end
	end
      end

      obj
    end

    if ($WWWSRV_NO_OBJECT_CACHE) then
      def fetch(*args)
	@builder.call(*args)
      end
    end

    alias new fetch
  end

  # NOTE: cacheable object requirements
  #   * Object#initialize method MUST be able to be called
  #      with no arguemtns.
  #   * Object#clear method MUST be usable and that method
  #      MUST reset object status.
  class RecycleObjectCache
    def initialize(klass, cache_limit=1000)
      @class = klass
      @recycle_queue = Array::new
      @cache_limit = cache_limit
    end

    attr_reader :cache_limit

    def size
      @recycle_queue.size
    end

    alias length size

    def clear
      @recycle_queue.clear
      nil
    end

    def fetch
      new_obj = @recycle_queue.shift || @class.new
      yield(new_obj) if iterator?
      new_obj
    end

    def store(old_obj)
      unless (old_obj.kind_of? @class) then
	raise "mismatch type: #{old_obj.type} vs. #{@class}"
      end
      if (@recycle_queue.length < @cache_limit) then # defence of cache overflow
	old_obj.clear
	@recycle_queue.push(old_obj)
      end
      nil
    end

    if ($WWWSRV_NO_OBJECT_CACHE) then
      def fetch
	new_obj = @class.new
	yield(new_obj) if iterator?
	new_obj
      end

      def store(old_obj)
	nil			# nothing to do.
      end
    end

    alias new fetch
  end
end
