# $Id: cgi.rb,v 1.84.2.3 2003/04/11 14:08:58 toki Exp $
# -
# Common Gateway Interface

=begin
Copyright (c) 2001-2002
	TOKI Yoshinori <toki@freedom.ne.jp>. All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  1.  Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.

  2.  Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation
      and/or other materials provided with the distribution.

  3.  The name of the author may not be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
OF SUCH DAMAGE.
=end

require 'wwwsrv/const'
require 'wwwsrv/ioutil'
require 'wwwsrv/process'
require 'wwwsrv/thread'
require 'wwwsrv/chunk'
require 'wwwsrv/reader'
require 'wwwsrv/header'
require 'wwwsrv/response'
require 'wwwsrv/document'

module WWWsrv
  class CGIdocument < Document
    include Path
    include Timeout
    include IOunit
    include ProtectProcess
    include ChunkedTransferCoding
    include PassAuth

    IO_TIMEOUT = 30

    def initialize(base_dir, nph=false, exec=nil)
      super()
      @base_dir = File::expand_path(base_dir)
      @allow = make_local_path_pattern(@base_dir)
      @path_info_root = nil
      @nph = nph
      @exec = exec
      @env = {}
      @timeout = IO_TIMEOUT
      @query_max_size = 1024
      @arguments_max_size = 1024
    end

    attr_reader   :env
    attr_accessor :timeout
    attr_reader   :path_info_root
    attr_accessor :query_max_size
    attr_accessor :arguments_max_size

    def path_info_root=(new_root)
      @path_info_root = File::expand_path(new_root)
    end

    alias environ env

    def folder?
      true
    end

    def cgi_executable?(cgi_path)
      if (@exec) then
	File::file? cgi_path
      else
	(File::file? cgi_path) && (File::executable? cgi_path)
      end
    end
    private :cgi_executable?

    def split_cgi_path(prefix, request)
      request.scan_path_info(prefix) do |script_name, path_info|
	(script_name =~ %r"/$") and next
	cgi_local_path = path_http2local(script_name, @base_dir)
	cgi_local_path = File::expand_path(cgi_local_path)
	if (cgi_local_path !~ @allow) then
	  http_error(404, "Not found a document: #{request.normalized_path}.")
	end
	if (cgi_executable? cgi_local_path) then
	  path_translated = path_http2local(path_info)
	  if (@path_info_root && ! path_info.empty?) then
	    path_translated = merge_local_path(@path_info_root, path_translated)
	    path_translated = File::expand_path(path_translated)
	  end
	  return cgi_local_path, merge_path(prefix, script_name), path_info, path_translated
	end
      end
      http_error(404, "Not found a executable document: #{request.normalized_path}.")
    end
    private :split_cgi_path

    def cgi_params(prefix, request, response)
      cgi_local_path, script_name, path_info, path_translated = split_cgi_path(prefix, request)
      response[:local_path] = cgi_local_path

      cgi_var = fetch_cgi_var(request, @env.dup, script_name)
      unless (path_info.empty?) then
	cgi_var['PATH_INFO'] = path_info
	cgi_var['PATH_TRANSLATED'] = path_translated
      end

      request.each_symbol do |param_name, param_value|
	if (param_name.to_s =~ /^F?CGI_/i) then
	  cgi_param_name = param_name.sub(/^F?CGI_/i, '')
	  cgi_param_name.upcase!
	  cgi_var[cgi_param_name] = param_value
	end
      end

      return cgi_local_path, cgi_var
    end
    private :cgi_params

    def cgi_open(request, cgi_local_path, cgi_var)
      cgi_exec = ''
      cgi_exec.concat(@exec + ' ') if @exec
      cgi_exec.concat(cgi_local_path)
      if (request.query) then
	if (request.query.length > @query_max_size) then
	  http_error(414, 'Too long CGI query.')
	end
	if (! request.query.empty? && request.query !~ /=/) then
	  cgi_args = ' ' + request.query.split(/\+/).map{ |argument|
	    quoteSHELL(decodeURL(argument))
	  }.join(' ')
	  if (cgi_args.length > @arguments_max_size) then
	    http_error(414, 'Too long CGI arguments.')
	  end
	  cgi_exec.concat(cgi_args)
	end
      end

      flags = 'r'
      flags.concat('+') if (request.method == HTTP_POST || request.header.content_length)
      cgi_io = PROCESS_LOCK.synchronize{
	protect_environ(cgi_var) {
	  protect_chdir(File::dirname(cgi_local_path)) {
	    IO::popen(cgi_exec, flags)
	  }
	}
      }
      cgi_io.binmode

      return cgi_io, cgi_exec
    end
    private :cgi_open

    def receive_request_body(request, cgi_io, cgi_exec)
      if (request.method == HTTP_POST || request.header.content_length) then
	begin
	  if (request.chunked?) then
	    # NOTE: Request trailer is ignored because CGI/1.1 has no
	    # interface to pass a trailer to a CGI application.
	    while (body_fragment = request.body.read_chunk)
	      wait(@timeout) do
		cgi_io.write(body_fragment)
	      end
	    end
	  else
	    while (body_fragment = request.body.read(IO_UNIT_SIZE))
	      wait(@timeout) do
		cgi_io.write(body_fragment)
	      end
	    end
	  end
	  wait(@timeout) do
	    cgi_io.flush
	  end
	ensure
	  cgi_io.close_write
	end
      end
    end
    private :receive_request_body

    def cgi_response_head(srv_opt, request, response, cgi_io, cgi_exec)
      if (request.version >= HTTP_1_0) then
	begin
	  parse_nph_cgi_status_line(srv_opt, request, response, cgi_io, cgi_exec) if @nph
	  wait(@timeout) { response.header.parse(cgi_io) }
	rescue ParseError
	  srv_opt.doc_log_warning(self, "failed to parse response header: #{cgi_exec}")
	  http_error(500, "Failed to parse CGI header: #{$!.message} (#{$!.class}).")
	end
	convert_cgi_resonse_header(srv_opt, request, response) unless @nph
      end
      response_header = response.header.dup
      yield(response) # response object may be modified by server.

      response_header
    end
    private :cgi_response_head

    def parse_nph_cgi_status_line(srv_opt, request, response, cgi_io, cgi_exec)
      status_line = wait(@timeout) { Response::parse_status_line(cgi_io) }
      unless (status_line) then
	srv_opt.doc_log_warning(self, "empty status line: #{cgi_exec}")
	http_error(500, "Empty status line: #{request.normalized_path}.")
      end
      response.version, response.status, response.reason = status_line
    end
    private :parse_nph_cgi_status_line

    def convert_cgi_resonse_header(srv_opt, request, response)
      if (response.header.location) then
	response.header.location = srv_opt.expand_location(request, response.header.location)
      end

      if (response.header.status) then
	response.status, response.reason = Response::parse_status(response.header.status)
	response.header.status = nil
      elsif (response.header.location) then
	response.status = 302
      end
    end
    private :convert_cgi_resonse_header

    def cgi_response_body(srv_opt, response_header, response_trailer, cgi_io, cgi_exec)
      if (Response::chunked? response_header) then
	cgi_reader = ChunkedTransferCodingReader::new(cgi_io)
	cgi_reader.each_chunk_with_ext do |chunk_data, chunk_extension|
	  yield(chunk_data, chunk_extension)
	end
	response_trailer.update(cgi_reader.trailer)
      else
	if (response_header.content_length) then
	  content_length = response_header.content_length.to_i
	  cgi_reader = LengthLimitedReader::new(cgi_io, content_length)
	else
	  cgi_reader = cgi_io
	end
	block_size = io_unit(cgi_io.stat)
	while (body_fragment = wait(@timeout) { cgi_reader.read(block_size) })
	  yield(body_fragment)
	end
      end
    end
    private :cgi_response_body

    def respond(srv_opt, prefix, request, response)
      srv_opt.doc_log_debug(self, 'enter')

      cgi_local_path, cgi_var = cgi_params(prefix, request, response)
      cgi_io, cgi_exec = cgi_open(request, cgi_local_path, cgi_var)
      begin
	srv_opt.doc_log_debug(self, "execution: #{cgi_exec}")
	receive_request_body(request, cgi_io, cgi_exec)
	# HTTP header of response may be modified by outer documents,
	# and current state of response header is saved at this point.
	response_header = \
	cgi_response_head(srv_opt, request, response, cgi_io, cgi_exec) { |arg, opts|
	  yield(arg, opts)
	}
	# HTTP trailer of response will not be modified by outer
	# documents until this CGIdocument#respond() method ended.
	cgi_response_body(srv_opt, response_header, response.trailer, cgi_io, cgi_exec) { |arg, opts|
	  yield(arg, opts)
	}
      ensure
	cgi_io.close
      end

      nil
    end
  end

  # alas
  CGIDocument = CGIdocument
end

require 'wwwsrv/cgi_obs' unless $WWWSRV_NOT_COMPAT
