# $Id: document.rb,v 1.123.2.2 2003/04/11 14:09:00 toki Exp $
# -
# server document

=begin
Copyright (c) 2001-2002
	TOKI Yoshinori <toki@freedom.ne.jp>. All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  1.  Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.

  2.  Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation
      and/or other materials provided with the distribution.

  3.  The name of the author may not be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
OF SUCH DAMAGE.
=end

require 'time'
require 'thread'
require 'digest/md5'
require 'wwwsrv/const'
require 'wwwsrv/delegate'
require 'wwwsrv/path'
require 'wwwsrv/header'
require 'wwwsrv/version'
require 'wwwsrv/htmlconv'

module WWWsrv
  # latest CVS commit date of core libraries for wwwsrv
  update(%q$Date: 2003/04/11 14:09:00 $)

  module HTTPerror
    X_HTTP_VERSION = Header::normalize('X-HTTP-Version').freeze

    # NOTE: if you need to change default HTTP version of HTTP/1.1 on
    # error response, you should set your HTTP version to a value of
    # special field of "X-HTTP-Version".
    def http_error(status, message=nil, header={})
      if (header.include? X_HTTP_VERSION) then
	http_version_field_value = header.delete(X_HTTP_VERSION)
	case (http_version_field_value)
	when HTTPversion
	  version = http_version_field_value
	when String
	  version = HTTPversion::parse(http_version_field_value)
	else
	  raise 'failed to convert HTTP version: ' + http_version_field_value.inspect
	end
      else
	version = HTTP_1_1
      end

      throw :WWWSRV_HTTP_ERROR, [ :error, [ status, message, header, version ] ]
    end
    module_function :http_error
  end

  # alias
  HTTPError = HTTPerror

  class Document
    include HTMLconv
    include HTTPerror

    def self::catch_status
      doc_status, err_args = catch(:WWWSRV_HTTP_ERROR) {
	yield
	:success
      }
      return doc_status, err_args
    end

    # patch for ruby-1.6.8
    unless (defined? object_id) then
      alias object_id id
    end

    def identity
      "#{self.class}(#{format('0x%x', object_id)})"
    end

    def warm_up(log)
      log.debug(identity + ': warm up...')
      nil
    end

    def cool_down(log)
      log.debug(identity + ': cool down...')
      nil
    end

    def subdocument(path)
      DocumentPart::new(self, path)
    end

    def respond(srv_opt, prefix, request, response)
      srv_opt.doc_log_debug(self, 'enter')
      http_error(404, "Empty document: #{request.normalized_path}.")
      nil
    end

    # If document is directory of contents, this method shuld return true.
    def folder?
      false
    end

    # If a document has other documents, this method must return all
    # internal documents as enumerable list.
    def internal_documents
      []
    end

    def to_s
      identity
    end
  end

  module ExportResponse
    def export_response(document, srv_opt, prefix, request, response, export_method=HTTP_GET)
      export_request = request.dup
      export_request.method = export_method
      export_request.erase_cond_header_fields
      export_request.erase_message_body
      document.respond(srv_opt, prefix, request, response) do |arg, opts|
	yield(arg, opts)
      end

      nil
    end
    module_function :export_response
  end

  class DocumentPart < Document
    include Path

    def initialize(document, path)
      super()
      @document = document
      @path = expand_http_path(path)
      @path.sub!(%r"/+$", '')
      @path_pattern = Regexp::compile(Regexp::quote(@path) + '$')
    end

    attr_reader :document

    def folder?
      @document.folder?
    end

    def warm_up(log)
      super
      log.debug("#{self}: part of document: #{@document} -> #{@path}")
      nil
    end

    def respond(srv_opt, prefix, request, response)
      srv_opt.doc_log_debug(self, 'enter')
      if (prefix =~ @path_pattern) then
	modified_prefix = prefix.sub(@path_pattern, '')
	srv_opt.doc_log_debug(self, "modify request path: #{request.subpath(prefix)} -> #{request.subpath(modified_prefix)}")
	prefix = modified_prefix
      else
	unless (subpath = request.subpath(prefix)) then
	  http_error(404, "Invalid path: #{request.normalized_path}.")
	end
	srv_opt.doc_log_debug(self, "modify request path: #{subpath}  -> #{@path}#{subpath}")
	modified_path = prefix + @path + subpath
	modified_path += '?' + request.query if request.query
	request.path = modified_path
      end

      @document.respond(srv_opt, prefix, request, response) do |arg, opts|
	yield(arg, opts)
      end

      nil
    end

    def internal_documents
      super + [ @document ]
    end
  end

  class DocumentNode
    def initialize
      @page = {}
      @filter = []
      @document = nil
      @subnode = {}
    end

    def mount(document, path_fragments, options={})
      if (path_fragments.empty?) then
	if (options.include? :page) then
	  page = options[:page]
	  if (@subnode.include? page) then
	    raise "already mounted as directory: #{page}"
	  end
	  @page[page] = document
	elsif (options.include? :mask) then
	  mask = options[:mask]
	  if (filter_pair = @filter.assoc(mask)) then
	    filter_pair[1] = document
	  else
	    @filter.push([ mask, document ])
	  end
	else
	  @document = document
	end
      else
	path_segment = path_fragments.shift
	unless (child_node = @subnode[path_segment]) then
	  if (@page.include? path_segment) then
	    raise "already mounted as page: #{path_segment.inspect}"
	  end
	  @subnode[path_segment] = child_node = DocumentNode::new
	end
	child_node.mount(document, path_fragments, options)
      end

      document
    end

    def umount(path_fragments, options={})
      if (path_fragments.empty?) then
	if (options.include? :page) then
	  page = options[:page]
	  return @page.delete(page)
	elsif (options.include? :mask) then
	  document = nil
	  mask = options[:mask]
	  @filter.delete_if{ |mask2, document2|
	    if (mask2 == mask) then
	      document = document2
	      true
	    end
	  }
	  return document
	else
	  document = @document
	  @document = nil
	  return document
	end
      else
	path_segment = path_fragments.shift
	if (child_node = @subnode[path_segment]) then
	  return child_node.umount(path_fragments, options)
	end
      end

      nil
    end

    def find(path_fragments, mount_path)
      if (path_fragments.empty?) then
	if (document = @page['']) then
	  # page of directory
	  return document
	end
      else
	path_segment = path_fragments.shift
	if (path_fragments.empty?) then
	  if (document = @page[path_segment]) then
	    return document
	  end
	end

	if (child_node = @subnode[path_segment]) then
	  mount_path.push(path_segment)
	  if (document = child_node.find(path_fragments, mount_path)) then
	    return document
	  end
	  mount_path.pop
	end

	path_fragments.unshift(path_segment)
	page_path = path_fragments.join('/')
	page_path = '/' + page_path unless mount_path.empty?
	for regexp_mask, document in @filter
	  if (page_path =~ regexp_mask) then
	    return document
	  end
	end
      end

      @document
    end

    def each(mount_path)
      if (@document) then
	yield(@document, {})
      end

      for mask, document in @filter
	yield(document, { :mask => mask })
      end

      for page, document in @page
	yield(document, { :page => page })
      end

      for path_segment, child_node in @subnode
	mount_path.push(path_segment)
	begin
	  child_node.each(mount_path) do |document, mount_opts|
	    yield(document, mount_opts)
	  end
	ensure
	  mount_path.pop
	end
      end

      self
    end
  end

  class DocumentTree
    include Path
    include Enumerable

    def initialize
      @root = DocumentNode::new
    end

    def split(path)
      path = path.sub(%r"/+$", '')
      path.empty? ? [ '' ] : path.split(%r"/", -1)
    end
    private :split

    def join(path_fragments)
      if (path_fragments.length == 1 && path_fragments.first.empty?) then
	return '/'
      else
	return path_fragments.join('/')
      end
    end
    private :join

    def mount(document, path, options={})
      path_fragments = split(path)
      @root.mount(document, path_fragments, options)
    end

    def umount(path, options={})
      path_fragments = split(path)
      @root.umount(path_fragments, options)
    end

    def find(path)
      mount_path = []
      path_fragments = split(path)
      if (document = @root.find(path_fragments, mount_path)) then
	return document, join(mount_path)
      end

      nil
    end

    def each
      mount_path = []
      @root.each(mount_path) do |document, mount_opts|
	yield(document, join(mount_path), mount_opts)
      end

      self
    end
  end

  class DocumentFolder < Document
    include Path

    def initialize
      super()
      @folder = {}
    end

    def folder?
      true
    end

    def mount(document, path, options={})
      if ((options.include? :page) && (options.include? :mask)) then
	raise "conflict parameters: page(#{options[:page].inspect}) vs. mask(#{options[:mask].inspect})"
      end

      mount_opts = {}
      for opt_name in [ :page, :mask ]
	if (options.include? opt_name) then
	  mount_opts[opt_name] = options[opt_name]
	end
      end

      host = :default
      host = options[:host] if (options.include? :host)

      @folder[host] = DocumentTree::new unless (@folder.include? host)
      @folder[host].mount(document, path, mount_opts)

      document
    end

    def umount(path, options={})
      if ((options.include? :page) && (options.include? :mask)) then
	raise "conflict parameters: page(#{options[:page].inspect}) vs. mask(#{options[:mask].inspect})"
      end

      mount_opts = {}
      for opt_name in [ :page, :mask ]
	if (options.include? opt_name) then
	  mount_opts[opt_name] = options[opt_name]
	end
      end

      host = :default
      host = options[:host] if (options.include? :host)

      if (@folder.include? host) then
	return @folder[host].umount(path, mount_opts)
      end

      nil
    end

    def attach(path, options={})
      if ((options.include? :page) && (options.include? :mask)) then
	raise "conflict parameters: page(#{options[:page].inspect}) vs. mask(#{options[:mask].inspect})"
      end

      host = :default
      host = options[:host] if (options.include? :host)

      target = nil
      target = options[:target] if (options.include? :target)

      mount_opts = options.dup
      mount_opts.delete(:target)

      target_path = path
      if (target) then
	target_path += '/' if (path !~ %r"/$")
	target_path += target
      end
      mount_pair = find(target_path, host) or raise "not found a mounted document: #{target_path.inspect}"

      document, mount_path = mount_pair
      if (path == mount_path) then
	modified_document = yield(document)
      else
	subpath = WWWsrv::Path::subpath(mount_path, path)
	subdocument = document.subdocument(subpath)
	modified_document = yield(subdocument)
      end

      mount(modified_document, path, mount_opts)
    end

    def find(path, host=:default)
      if (@folder.include? host) then
	return @folder[host].find(path)
      end

      nil
    end

    def each
      if (@folder.include? :default) then
	for document, mount_path, mount_opts in @folder[:default]
	  yield(document, mount_path, mount_opts)
	end
      end

      hosts = @folder.keys
      hosts -= [ :default ]
      hosts.sort!

      for name in hosts
	for document, mount_path, mount_opts in @folder[name]
	  mount_opts[:host] = name
	  yield(document, mount_path, mount_opts)
	end
      end

      self
    end

    def virtual_host
      @folder.keys.reject{ |hostname| hostname != :default }
    end

    def virtual_host_defined?(hostname)
      @folder.include? hostname
    end

    def warm_up(log)
      super

      for document, mount_path, mount_opts in self
	msg = "#{self}: mount #{document} -> #{mount_path.inspect}"
	mount_opts.keys.map{ |opt_sym|
	  [ opt_sym.to_s, opt_sym ]
	}.sort{ |a, b|
	  a[0] <=> b[0]
	}.map{ |opt_name, opt_sym|
	  opt_sym
	}.each do |opt_name|
	  msg += " + #{opt_name}=#{mount_opts[opt_name].inspect}"
	end
	log.debug(msg)
      end

      nil
    end

    def respond(srv_opt, prefix, request, response)
      srv_opt.doc_log_debug(self, 'enter')
      if (request) then
	path = request.subpath(prefix)
	if (virtual_host_defined? request.host.to_s) then
	  mount_pair = find(path, request.host.to_s)
	elsif (virtual_host_defined? request.host.name) then
	  mount_pair = find(path, request.host.name)
	else
	  mount_pair = find(path)
	end
      else
	# for fatal alert document
	path = '/'
	mount_pair = find(path)
      end

      if (mount_pair) then
	document, mount_path = mount_pair
	if (document.folder? && request.normalized_path !~ %r"/$" && prefix + mount_path == request.normalized_path) then
	  srv_opt.doc_log_debug(self, "redirect from #{request.normalized_path} to #{request.normalized_path}/")
	  response.status = 301
	  response.header.location =
	    srv_opt.scheme + '://' + request.host.name + ':' + request.host.port.to_s +
	    request.normalized_path + '/'
	  response.header.content_type = 'text/plain'
	  location = response.header.location
	  yield(response)
	  yield("Moved to #{location}\n")
	else
	  srv_opt.doc_log_debug(self, "request to subdocument: #{prefix}#{mount_path}")
	  prefix += mount_path.sub(%r"/+$", '')
	  document.respond(srv_opt, prefix, request, response) do |arg, opts|
	    yield(arg, opts)
	  end
	end
      else
	http_error(404, "Not found a document of #{request.path}.")
      end

      nil
    end

    def internal_documents
      documents = []
      for hostname, doc_tree in @folder
	for document, path, page in @folder[hostname]
	  documents.push(document)
	end
      end

      super + documents
    end
  end

  class DocumentDispatcher < Document
    def initialize(default_document=nil)
      super()
      @default_document = default_document
      @document_alist = []
      @vary_field_names = {}
    end

    attr_accessor :default_document

    def folder?
      @default_document.folder?
    end

    def add_cond(document, *vary_field_names, &cond)
      @document_alist.push([ cond, document ])
      for field_name in vary_field_names
	@vary_field_names[Header::normalize(field_name)] = true
      end

      document
    end

    def warm_up(log)
      super
      log.debug("#{self}: default -> #{@default_document}") if @default_document
      for cond, document in @document_alist
	log.debug("#{self}: condition #{cond.inspect} -> #{document}")
      end

      nil
    end

    def cond_call(cond, srv_opt, prefix, request)
      if ($WWWSRV_NOT_COMPAT || cond.arity.abs == 3) then
	return cond.call(srv_opt, prefix, request)
      elsif (cond.arity.abs >= 4) then
	# for backward compatibility
	return cond.call(srv_opt.sid, srv_opt, prefix, request)
      elsif (cond.arity.abs >= 2) then
	# for backward compatibility
	return cond.call(srv_opt, request)
      else
	# for backward compatibility
	return cond.call(request)
      end
    end
    private :cond_call

    def respond(srv_opt, prefix, request, response)
      srv_opt.doc_log_debug(self, 'enter')

      @vary_field_names.each_key do |field_name|
	if (! response.header.vary || ! response.header.vary.index(field_name)) then
	  response.header.push('Vary', field_name)
	end
      end

      for cond, document in @document_alist
	if (cond_call(cond, srv_opt, prefix, request)) then
	  srv_opt.doc_log_debug(self, "dispatch document: #{document.class}")
	  document.respond(srv_opt, prefix, request, response) do |arg, opts|
	    yield(arg, opts)
	  end
	  return
	end
      end

      if (@default_document) then
	srv_opt.doc_log_debug(self, "default document: #{@default_document.class}")
	@default_document.respond(srv_opt, prefix, request, response) do |arg, opts|
	  yield(arg, opts)
	end
	return
      end

      http_error(404, "Not found a document at #{request.normalized_path}.\n")
    end

    def internal_documents
      super + [ @default_document ] + @document_alist.map{ |cond, document| document }
    end
  end

  class Page < Document
    def initialize(message, type='text/html', mtime=nil, add_content_md5=false)
      super()
      @message = message.dup
      @message.freeze
      @type = type
      @mtime = mtime
      etag = [
	@message.length,
	@message.hash,
	@type.hash,
	@mtime.hash
      ].map{ |n| format('%x', n) }.join('-')
      @entity_tag = '"' + etag + '"'
      self.content_md5 = add_content_md5
    end

    def content_md5
      @content_md5 ? true : false
    end

    def content_md5=(boolean)
      if (boolean) then
	@content_md5 = [ Digest::MD5::digest(@message) ].pack('m').strip
      else
	@content_md5 = nil
      end

      boolean
    end

    def respond(srv_opt, prefix, request, response)
      srv_opt.doc_log_debug(self, 'enter')
      case (request.method)
      when HTTP_GET, HTTP_HEAD
	response.status = request.cond_if(@mtime, @entity_tag)
	if (response.status >= 400) then
	  http_error(response.status)
	end
	response.header.etag = @entity_tag
	response.header.content_md5 = @content_md5 if @content_md5
	response.header.last_modified = http_date(@mtime) if @mtime
	if (response.status == 200) then
	  response.header.content_type = @type
	  response.header.content_length = @message.length.to_s
	end
	status = response.status
	yield(response)
	yield(@message.dup) if (request.method == HTTP_GET && status == 200)
      else
	http_error(405, "Not supported method: #{request.method}.", 'Allow' => 'GET, HEAD')
      end

      nil
    end
  end

  class Redirect < Document
    include ExportResponse

    REDIRECT_PAGE = Document::new

    def REDIRECT_PAGE.respond(srv_opt, prefix, request, response)
      location = response.header.location
      yield(response)
      yield(<<-EOF)
<html>
<head>
<title>Location: #{HTMLconv::quoteHTML(response.header.location)}</title>
</head>
<body>
<h1>Location: #{HTMLconv::quoteHTML(response.header.location)}</h1>
<p>Jump to Location of <kbd>#{HTMLconv::quoteHTML(response.header.location)}</kbd>...</p>
</body>
</html>
      EOF

      nil
    end

    def initialize(status=301, page=REDIRECT_PAGE, &redirect)
      super()
      @page = ResponseFilter::new(page) { |srv_opt, prefix, request, response|
	case (response.status)
	when 200...300
	  if ($WWWSRV_NOT_COMPAT || redirect.arity.abs == 3) then
	    location = redirect.call(srv_opt, prefix, request)
	  elsif (redirect.arity.abs >= 4) then
	    # for backward compatibility
	    location = redirect.call(srv_opt.sid, srv_opt, prefix, request)
	  elsif (redirect.arity.abs >= 2) then
	    # for backward compatibility
	    location = redirect.call(srv_opt, request)
	  else
	    # for backward compatibility
	    location = redirect.call(request)
	  end

	  if (location !~ %r"^[A-Za-z][A-Za-z0-9.+-]*://") then
	    if (location !~ %r"^/") then
	      location = request.normalized_path.sub(%r"[^/]*$", '') + location
	    end
	    location = srv_opt.scheme + '://' + request.host.name + ':' + request.host.port.to_s + location
	  end

	  response.status = status
	  response.header.location = location
	  response.header.content_type = 'text/plain'

	  srv_opt.doc_log_debug(self, "redirect from #{request.normalized_path} to #{response.header.location}")
	else
	  raise "failed to get a redirect page from #{page.inspect}"
	end
      }
    end

    def warm_up(log)
      super
      log.debug("#{self}: redirect page -> #{@page}")
      nil
    end

    def respond(srv_opt, prefix, request, response)
      srv_opt.doc_log_debug(self, 'enter')
      export_response(@page, srv_opt, prefix, request, response) do |arg, opts|
	yield(arg, opts)
      end

      nil
    end

    def internal_documents
      super + [ @page ]
    end
  end

  class DocumentModifier < Document
    include Delegatable

    def initialize(document)
      super()
      @document = document
    end

    attr_reader :document

    def folder?
      document.folder?
    end

    def terminate
      throw(:wwwsrv_document_modifier_terminal)
    end
    private :terminate

    alias prune terminate; private :prune

    def warm_up(log)
      super
      log.debug("#{self}: modified document -> #{document}")
      nil
    end

    def respond(srv_opt, prefix, request, response, &block)
      srv_opt.doc_log_debug(self, 'enter')
      context = {}
      begin
	catch(:wwwsrv_document_modifier_terminal) do
	  front_guard(srv_opt, context)
	  modify_request(srv_opt, context, prefix, request, response, &block)
	  doc_response = nil
	  @document.respond(srv_opt, prefix, request, response) do |arg, opts|
	    unless (doc_response) then
	      doc_response = arg
	      modify_response(srv_opt, context, prefix, request, doc_response, &block)
	    else
	      modify_message_body(srv_opt, context, arg, opts, &block)
	    end
	  end
	end
	modify_terminal(srv_opt, context, &block)
      ensure
	rear_guard(srv_opt, context)
      end

      nil
    end

    def front_guard(srv_opt, context)
      nil
    end

    def modify_request(srv_opt, context, prefix, request, response)
      nil
    end

    def modify_response(srv_opt, context, prefix, request, response)
      yield(response)
      nil
    end

    def modify_message_body(srv_opt, context, body_fragment, opts=nil)
      yield(body_fragment, opts)
      nil
    end

    def modify_terminal(srv_opt, context)
      nil
    end

    def rear_guard(srv_opt, context)
      nil
    end

    def internal_documents
      super + [ @document ]
    end

    delegate(Document, :document)
  end

  class AliasDocument < DocumentModifier
    include Path

    def initialize(document)
      super(document)
      @map = {}
    end

    def alias(dst_path, src_path)
      @map[dst_path] = src_path
      self
    end

    def modify_request(srv_opt, context, prefix, request, response)
      if (alias_path = @map[request.subpath(prefix)]) then
	count = 1
	expanded_path = alias_path
	while (alias_path = @map[expanded_path])
	  if (count > 10) then
	    http_error(404, "Failed to expand alias path.")
	  end
	  count += 1
	  expanded_path = alias_path
	end

	srv_opt.doc_log_debug(self, "expand alias path: #{request.normalized_path} -> #{expanded_path}")

	expanded_path += '?' + request.query if request.query
	request.path = expanded_path
      else
	srv_opt.doc_log_debug(self, "not modified path: #{request.normalized_path}")
      end

      nil
    end
  end

  class RequestFilter < DocumentModifier
    def initialize(document, &block)
      super(document)
      @filter = block
    end

    def modify_request(srv_opt, context, prefix, request, response)
      if ($WWWSRV_NOT_COMPAT || @filter.arity.abs <= 3) then
	@filter.call(srv_opt, prefix, request)
      else
	# for backward compatibility
	@filter.call(srv_opt.sid, srv_opt, prefix, request)
      end

      nil
    end
  end

  class ResponseFilter < DocumentModifier
    def initialize(document, &block)
      super(document)
      @filter = block
    end

    def modify_response(srv_opt, context, prefix, request, response)
      if ($WWWSRV_NOT_COMPAT || @filter.arity.abs <= 4) then
	@filter.call(srv_opt, prefix, request, response)
      else
	# for backward compatibility
	@filter.call(srv_opt.sid, srv_opt, prefix, request, response)
      end
      yield(response)

      nil
    end
  end

  class AccessControl < DocumentModifier
    ALLOW = :allow
    DENY  = :deny

    def initialize(document)
      super(document)
      @access_list = []
    end

    def add_access_control(action, logging=false, &filter)
      @access_list.push([ action, filter, logging ])
      self
    end

    def filter_call(filter, srv_opt, prefix, request)
      if ($WWWSRV_NOT_COMPAT || filter.arity.abs == 3) then
	return filter.call(srv_opt, prefix, request)
      elsif (filter.arity.abs >= 4) then
	# for backward compatibility
	return filter.call(srv_opt.sid, srv_opt, prefix, request)
      elsif (filter.arity.abs >= 2) then
	# for backward compatibility
	return filter.call(srv_opt, request)
      else
	# for backward compatibility
	return filter.call(request)
      end
    end
    private :filter_call

    def modify_request(srv_opt, context, prefix, request, response)
      catch(:ACCESS_ALLOW) do
	allow_filter_used = false
	for action, filter, logging in @access_list
	  case (action)
	  when ALLOW
	    allow_filter_used = true
	    if (filter_call(filter, srv_opt, prefix, request)) then
	      if (logging) then
		srv_opt.srv_log_system("access allow: #{request.client}")
	      end
	      throw(:ACCESS_ALLOW)
	    end
	  when DENY
	    if (filter_call(filter, srv_opt, prefix, request)) then
	      if (logging) then
		srv_opt.srv_log_system("access deny: #{request.client}")
	      end
	      http_error(403, 'Access denied.')
	    end
	  else
	    raise "unknown action: #{action.inspect}"
	  end
	end

	if (allow_filter_used) then
	  if (logging) then
	    srv_opt.srv_log_system("access deny: #{request.client}")
	  end
	  http_error(403, 'Access denied.')
	end
      end

      nil
    end
  end

  class PartialContent < DocumentModifier
    def self::parse_range(field_value)
      case (field_value)
      when /^bytes=(\d+)-(\d+)?$/
	return $1.to_i, $2 && $2.to_i
      when /^bytes=-(\d+)$/
	return nil, $1.to_i
      else
	# not support multipart/byteranges.
	return nil
      end
    end

    def self::normalize_range(first_pos, last_pos, content_length)
      if (content_length == 0) then
	HTTPerror::http_error(416, 'no content.', 'Content-Range' => '*/0')
      end
      if (first_pos) then
	if (first_pos >= content_length) then
	  HTTPerror::http_error(416, 'overflow of first position of range.', 'Content-Range' => "*/#{content_length}")
	end
	if (! last_pos || last_pos >= content_length) then
	  last_pos = content_length - 1
	end
      else
	if (last_pos < content_length) then
	  first_pos = content_length - last_pos
	  last_pos = content_length - 1
	else
	  first_pos = 0
	  last_pos = content_length - 1
	end
      end
      if (first_pos >= last_pos) then
	HTTPerror::http_error(416, 'left position is greater than right position.', 'Content-Range' => "*/#{content_length}")
      end

      return first_pos, last_pos
    end

    def self::partial_length(first_pos, last_pos)
      last_pos - first_pos + 1
    end

    def self::set_content_length(response_header, first_pos, last_pos)
      response_header.content_length = partial_length(first_pos, last_pos).to_s
    end

    def self::set_content_range(response_header, first_pos, last_pos, content_length, bytes_unit='bytes')
      response_header.content_range = "#{bytes_unit} #{first_pos}-#{last_pos}/#{content_length}"
    end

    def self::setup_partial_content(response_header, first_pos, last_pos, content_length, bytes_unit='bytes')
      set_content_length(response_header, first_pos, last_pos)
      set_content_range(response_header, first_pos, last_pos, content_length, bytes_unit)
      nil
    end

    def self::set_accept_ranges(response_header, bytes_unit='bytes')
      if (bytes_unit == 'bytes') then
	if (response_header.accept_ranges !~ /bytes/i) then
	  response_header.push('Accept-Ranges', 'bytes')
	end
      else
	if (response_header.accept_ranges !~ /#{bytes_unit}/i) then
	  response_header.push('Accept-Ranges', bytes_unit)
	end
      end
    end

    def cut_range(body_fragment, curr_pos, first_pos, last_pos)
      end_pos = curr_pos + body_fragment.length - 1
      if (first_pos <= curr_pos && curr_pos <= last_pos) then
	if (end_pos <= last_pos) then
	  yield(body_fragment)
	else
	  yield(body_fragment[0, last_pos - curr_pos + 1])
	end
      elsif (first_pos <= end_pos && end_pos <= last_pos) then
	yield(body_fragment[first_pos - curr_pos, end_pos - first_pos + 1])
      end
    end
    private :cut_range

    def modify_response(srv_opt, context, prefix, request, response)
      context[:sender] = catch(:wwwsrv_partial_content_sender) {
	if (response.header.content_length) then
	  PartialContent::set_accept_ranges(response.header)
	  if (response.status == 200 && request.header.range) then
	    if (req_range = PartialContent::parse_range(request.header.range)) then
	      if (request.header.if_range) then
		if (response.header.last_modified) then
		  begin
		    mtime = Time::parse(response.header.last_modified)
		  rescue ArgumentError
		    mtime = nil
		  end
		end
		response.status = request.cond_if_range(mtime, response.header.etag)
	      else
		response.status = 206
	      end
	      if (response.status == 206) then
		srv_opt.doc_log_debug(self, "partial content: range -> #{req_range[0]}..#{req_range[1]}")
		response.header.content_md5 = nil
		content_length = response.header.content_length.to_i
		partial_first, partial_last = PartialContent::normalize_range(req_range[0], req_range[1], content_length)
		PartialContent::setup_partial_content(response.header, partial_first, partial_last, content_length)
		context[:range] = [ partial_first, partial_last ]
		context[:position] = 0
		throw(:wwwsrv_partial_content_sender, method(:send_partial))
	      end
	    end
	  end
	end
	method(:send_through)
      }
      yield(response)

      nil
    end

    def send_partial(srv_opt, context, body_fragment, opts=nil)
      cut_range(body_fragment, context[:position], *context[:range]) do |range_fragment|
	yield(range_fragment)
      end
      context[:position] += body_fragment.length
    end
    private :send_partial

    def send_through(srv_opt, context, body_fragment, opts=nil)
      yield(body_fragment, opts)
    end
    private :send_through

    def modify_message_body(srv_opt, context, body_fragment, opts=nil)
      context[:sender].call(srv_opt, context, body_fragment, opts) do |body_fragment2, opts2|
	yield(body_fragment2, opts2)
      end

      nil
    end
  end
end
