# $Id: fastcgi.rb,v 1.126.2.2 2003/01/05 13:05:52 toki Exp $
# -
# Fast CGI interface

=begin
Copyright (c) 2001-2002
	TOKI Yoshinori <toki@freedom.ne.jp>. All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  1.  Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.

  2.  Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation
      and/or other materials provided with the distribution.

  3.  The name of the author may not be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
OF SUCH DAMAGE.
=end

require 'time'
require 'socket'
require 'thread'
require 'wwwsrv/const'
require 'wwwsrv/cache'
require 'wwwsrv/delegate'
require 'wwwsrv/thread'
require 'wwwsrv/ioutil'
require 'wwwsrv/chunk'
require 'wwwsrv/reader'
require 'wwwsrv/path'
require 'wwwsrv/header'
require 'wwwsrv/request'
require 'wwwsrv/response'
require 'wwwsrv/document'

module WWWsrv
  module FastCGIconst
    FCGI_VERSION = 1

    FCGI_BEGIN_REQUEST      = 1
    FCGI_ABORT_REQUEST      = 2
    FCGI_END_REQUEST        = 3
    FCGI_PARAMS             = 4
    FCGI_STDIN              = 5
    FCGI_STDOUT             = 6
    FCGI_STDERR             = 7
    FCGI_DATA               = 8
    FCGI_GET_VALUES         = 9
    FCGI_GET_VALUES_RESULT  = 10
    FCGI_UNKNOWN_TYPE       = 11
    FCGI_MAXTYPE            = FCGI_UNKNOWN_TYPE

    FCGI_NULL_REQUEST_ID = 0

    FCGI_KEEP_CONN = 1

    FCGI_RESPONDER  = 1
    FCGI_AUTHORIZER = 2
    FCGI_FILTER     = 3

    FCGI_REQUEST_COMPLETE = 0
    FCGI_CANT_MPX_CONN    = 1
    FCGI_OVERLOADED       = 2
    FCGI_UNKNOWN_ROLE     = 3
  end

  class FastCGIError < StandardError; end
  class FastCGIParseError < FastCGIError; end
  class FastCGIProtocolError < FastCGIError; end
  class FastCGIConnectionError < FastCGIError; end

  class FastCGIRecord
    include FastCGIconst
    include ConfirmedRead
    include ConfirmedWrite

    def initialize(record_type=nil, request_id=nil, content=nil, padding_length=0, reserved=0, version=FCGI_VERSION)
      @record_type = record_type
      @request_id = request_id
      @content = content
      @padding_length = padding_length
      @reserved = reserved
      @version = version
    end

    attr_accessor :record_type
    attr_accessor :request_id
    attr_reader   :content
    attr_accessor :padding_length
    attr_accessor :reserved
    attr_accessor :version

    def content=(new_content)
      if (new_content) then
	@content = new_content
	@padding_length = new_content.length % 8
      else
	@content = nil
	@padding_length = 0
      end

      new_content
    end

    def clear
      @record_type = nil
      @request_id = nil
      @content = nil
      @padding_length = 0
      @reserved = 0
      @version = FCGI_VERSION

      self
    end

    def receive_from(input)
      record_header = confirmed_read(input, 8)
      if (! record_header || record_header.length != 8) then
	raise FastCGIParseError, 'terminated input stream'
      end
      record_header = record_header.unpack('CCnnCC')

      @version        = record_header[0]
      @record_type    = record_header[1]
      @request_id     = record_header[2]
      content_length  = record_header[3]
      @padding_length = record_header[4]
      @reserved       = record_header[5]

      body_length = content_length + @padding_length
      if (body_length > 0) then
	body = confirmed_read(input, body_length)
	if (! body || body.length != body_length) then
	  raise FastCGIParseError, 'terminated input stream'
	end
	if (content_length > 0) then
	  @content = body[0, content_length]
	else
	  @content = nil
	end
      else
	@content = nil
      end

      self
    end

    def send_to(output)
      message = [
	@version,
	@record_type,
	@request_id,
	@content ? @content.length : 0,
	@padding_length,
	@reserved
      ].pack('CCnnCC')
      message.concat(@content) if @content
      message.concat("\0" * @padding_length) if (@padding_length > 0)
      confirmed_write(output, message)

      self
    end

    short_limit_shift = 7
    long_limit_shift = 31
    long_limit_mask = 0
    long_limit_shift.times do
      long_limit_mask = (long_limit_mask << 1) | 1
    end

    NAME_VALUE_PAIR_SHORT_LIMIT = 1 << short_limit_shift
    NAME_VALUE_PAIR_LONG_LIMIT = 1 << long_limit_shift
    NAME_VALUE_PAIR_LONG_MASK = long_limit_mask

    def self::encode_name_value_pairs(name_value_pairs)
      encoded_name_value_pairs = ''
      for name, value in name_value_pairs
	value = '' unless value
	for len in [ name.length, value.length ]
	  if (len < NAME_VALUE_PAIR_SHORT_LIMIT) then
	    encoded_name_value_pairs.concat [
	      len
	    ].pack('C')
	  elsif (len < NAME_VALUE_PAIR_LONG_LIMIT) then
	    encoded_name_value_pairs.concat [
	      len | NAME_VALUE_PAIR_LONG_LIMIT
	    ].pack('N')
	  else
	    raise FastCGIParseError, 'too long data in name-value pair'
	  end
	end
	encoded_name_value_pairs.concat(name)
	encoded_name_value_pairs.concat(value)
      end

      encoded_name_value_pairs
    end

    def self::decode_name_value_pairs(string, fcgi_var=((! iterator?) ? {} : nil))
      i = 0
      while (i < string.length)
	name_len = string[i, 1].unpack('C').first
	if (name_len & NAME_VALUE_PAIR_SHORT_LIMIT != 0) then
	  unless (i + 3 < string.length) then
	    raise FastCGIParseError, "too short name length"
	  end
	  name_len = string[i, 4].unpack('N').first & NAME_VALUE_PAIR_LONG_MASK
	  i += 4
	else
	  i += 1
	end

	unless (i < string.length) then
	  raise FastCGIParseError, "too short value length"
	end
	value_len = string[i, 1].unpack('C').first
	if (value_len & NAME_VALUE_PAIR_SHORT_LIMIT != 0) then
	  unless (i + 3 < string.length) then
	    raise FastCGIParseError, "too short value length"
	  end
	  value_len = string[i, 4].unpack('N').first & NAME_VALUE_PAIR_LONG_MASK
	  i += 4
	else
	  i += 1
	end

	unless (i + name_len + value_len <= string.length) then
	  raise FastCGIParseError, "too short name value data"
	end
	name = string[i, name_len]; i += name_len
	value = string[i, value_len]; i += value_len

	if (iterator?) then
	  yield(name, value)
	else
	  fcgi_var[name] = value
	end
      end

      (! iterator?) ? fcgi_var : nil
    end

    def self::encode_begin_request(role, flags=0, reserved='')
      if (reserved.length > 5) then
	raise FastCGIParseError, "too long reserved data: #{reserved.inspect}"
      end
      [ role, flags, reserved ].pack('nCa5')
    end

    def self::decode_begin_request(string)
      if (string.length != 8) then
	raise FastCGIParseError, "mismatch record length: #{string.inspect}"
      end
      string.unpack('nCa5')
    end

    def self::encode_end_request(app_status, protocol_status, reserved='')
      if (reserved.length > 3) then
	raise FastCGIParseError, "too long reserved data: #{reserved.inspect}"
      end
      [ app_status, protocol_status, reserved ].pack('NCa3')
    end

    def self::decode_end_request(string)
      if (string.length != 8) then
	raise FastCGIParseError, "mismatch record length: #{string.inspect}"
      end
      string.unpack('NCa3')
    end

    def self::encode_unknown_type(unknown_type, reserved='')
      if (reserved.length > 7) then
	raise FastCGIParseError, "too long reserved data: #{reserved.inspect}"
      end
      [ unknown_type, reserved ].pack('Ca7')
    end

    def self::decode_unknown_type(string)
      if (string.length != 8) then
	raise FastCGIParseError, "mismatch record length: #{string.inspect}"
      end
      string.unpack('Ca7')
    end

    CACHE = RecycleObjectCache::new(self)

    def self::fetch_cache(&block)
      CACHE.fetch(&block)
    end

    def self::store_cache(old_obj)
      CACHE.store(old_obj)
    end
  end

  class FastCGIRequestIDPool
    def initialize
      @max_id = 16
      @pool_list = (1..@max_id).to_a
      @pool_lock = Mutex::new
    end

    def fetch
      @pool_list.shift || @pool_lock.synchronize{
	@max_id += 1
      }
    end
    private :fetch

    def store(req_id)
      @pool_list.push(req_id)
    end
    private :store

    def use_id
      req_id = fetch
      begin
	return yield(req_id)
      ensure
	store(req_id)
      end
    end
  end

  class FastCGIConnection
    include Timeout
    include FastCGIconst
    include SocketManipulation

    def initialize(options={}, &block)
      @conn_timeout = (options.include? :conn_timeout) ? options[:conn_timeout] : FastCGIConnectable::CONN_TIMEOUT
      @max_conns    = (options.include? :max_conns)    ? options[:max_conns]    : FastCGIConnectable::MAX_CONNS
      @retries      = (options.include? :retries)      ? options[:retries]      : FastCGIConnectable::RETRIES
      @retry_wait   = (options.include? :retry_wait)   ? options[:retry_wait]   : FastCGIConnectable::RETRY_WAIT
      @keep_conn    = (options.include? :keep_conn)    ? options[:keep_conn]    : FastCGIConnectable::KEEP_CONN
      @sock_lock = Mutex::new
      @sock_cond = ConditionVariable::new
      @sock_count = 0
      @sock_queue = Array::new
      @sock_make = block
    end

    attr_accessor :conn_timeout
    attr_accessor :max_conns
    attr_accessor :retries
    attr_accessor :retry_wait
    attr_accessor :keep_conn

    def make_socket(srv_opt)
      socket = nil
      begin
	wait(@conn_timeout) do
	  socket = @sock_make.call
	end
      rescue
	if (socket) then
	  begin
	    begin
	      socket.shutdown
	    ensure
	      socket.close
	    end
	  rescue
	    # nothing to do.
	  end
	end
	raise FastCGIConnectionError, $!.inspect, $!.backtrace
      end
      socket = ProtectedSocket::new(socket, FastCGIConnectionError)
      srv_opt.srv_log_debug("open FastCGI connection: #{connect_log(socket)}")
      socket
    end
    private :make_socket

    def make_id_pool
      FastCGIRequestIDPool::new
    end
    private :make_id_pool

    def each_sock
      @sock_lock.synchronize do
	for socket, id_pool in @sock_queue
	  yield(socket)
	end
      end

      nil
    end

    def close(document, log)
      self.each_sock do |socket|
	begin
	  sock_info = connect_log(socket)
	  begin
	    socket.shutdown
	  ensure
	    socket.close
	  end
	  log.debug("#{document}: close FastCGI connection: #{sock_info}")
	rescue FastCGIConnectionError
	  log.alert("#{document}: failed to close FastCGI connection: #{socket.inspect}: #{$!}")
	  Logger::backtrace(log, Logger::WARNING, srv_opt.sid, $!)
	end
      end

      nil
    end

    def fetch(srv_opt)
      @sock_lock.synchronize do
	loop do
	  if (! @sock_queue.empty?) then
	    socket, id_pool = @sock_queue.shift
	    return socket, id_pool
	  elsif (@sock_count < @max_conns) then
	    socket, id_pool = make_socket(srv_opt), make_id_pool # an exception may occur.
	    @sock_count += 1	# count up after building socket up completely.
	    return socket, id_pool
	  end
	  @sock_cond.wait(@sock_lock)
	end
      end
    end
    private :fetch

    def sock_cond_signal
      if (@sock_queue.empty? || @sock_count >= @max_conns) then
	@sock_cond.signal
      end
    end
    private :sock_cond_signal

    def sock_close(srv_opt, socket)
      begin
	sock_info = connect_log(socket)
	begin
	  socket.shutdown
	ensure
	  socket.close
	end
	srv_opt.srv_log_debug("close FastCGI connection: #{sock_info}")
      rescue FastCGIConnectionError
	srv_opt.srv_log_alert("failed to close FastCGI connection: #{socket.inspect}: #{$!}")
	Logger::backtrace(srv_opt.log, Logger::WARNING, srv_opt.sid, $!)
      ensure
	if (iterator?) then
	  yield
	end
      end
    end
    private :sock_close

    def store(srv_opt, socket, id_pool)
      @sock_lock.synchronize do
	sock_cond_signal
	if (@keep_conn) then
	  @sock_queue.push([ socket, id_pool ])
	else
	  sock_close(srv_opt, socket) {
	    @sock_count -= 1
	  }
	end
      end
      nil
    end
    private :store

    def session(srv_opt)
      begin
	count = 0
	begin
	  socket, id_pool = fetch(srv_opt)
	  return yield(socket, id_pool)
	rescue FastCGIError
	  srv_opt.srv_log_alert("FastCGI session error: #{$!}")
	  Logger::backtrace(srv_opt.log, Logger::WARNING, srv_opt.sid, $!)

	  if (socket) then
	    sock_close(srv_opt, socket) {
	      @sock_lock.synchronize do
		sock_cond_signal
		@sock_count -= 1
	      end
	      socket, id_pool = nil, nil
	    }
	  end

	  count += 1
	  if (srv_opt[:fastcgi_unrecoverable_session]) then
	    srv_opt.srv_log_alert("overrun recoverable point and FastCGI session aborted")
	    raise
	  elsif (count <= @retries) then
	    srv_opt.srv_log_warning("retry FastCGI session: sleep -> #{@retry_wait}")
	    sleep(@retry_wait) if @retry_wait
	    srv_opt.srv_log_warning("retry FastCGI session: count -> #{count}")
	    retry
	  else
	    srv_opt.srv_log_alert("excess FastCGI session retrial limit: count -> #{count}")
	    raise
	  end
	rescue
	  # handling errors unrelated to FastCGI (ex. HTTP connection error, etc.)
	  if (socket) then
	    sock_close(srv_opt, socket) {
	      @sock_lock.synchronize do
		sock_cond_signal
		@sock_count -= 1
	      end
	      socket, id_pool = nil, nil
	    }
	  end
	  raise
	end
      ensure
	if (socket && id_pool) then
	  store(srv_opt, socket, id_pool)
	end
      end
    end

    def self::query_max_connections(socket, log=nil)
      log.debug('debug: query FastCGI max connections') if log

      record = FastCGIRecord::fetch_cache

      record.record_type = FCGI_GET_VALUES
      record.request_id = 0
      record.content = FastCGIRecord::encode_name_value_pairs(%w[ FCGI_MAX_CONNS ])
      log.internal("debug: send FastCGI record: FCGI_GET_VALUES: #{record.inspect}") if log
      record.send_to(socket)
      socket.flush

      record.clear
      record.receive_from(socket)
      if (record.record_type != FCGI_GET_VALUES_RESULT) then
	raise FastCGIProtocolError, "unexpected record type: #{record.inspect}"
      end
      log.internal("debug: receive FastCGI record: FCGI_GET_VALUES_RESULT: #{record.inspect}") if log
      fcgi_values = FastCGIRecord::decode_name_value_pairs(encoded_name_value_pairs)

      FastCGIRecord::store_cache(record)

      if (fcgi_values.include? 'FCGI_MAX_CONNS') then
	log.debug("debug: FastCGI max connecitons -> #{fcgi_values['FCGI_MAX_CONNS']}")
	return fcgi_values['FCGI_MAX_CONNS'].to_i
      else
	log.debug('debug: failed to get FastCGI max connecitons')
	return nil
      end
    end
  end

  module FastCGIConnectable
    include SocketManipulation

    # default connection parameters
    CONN_TIMEOUT = 4
    MAX_CONNS    = 4
    RETRIES      = 1
    RETRY_WAIT   = 1
    KEEP_CONN    = false

    def initialize(options={}, &block)
      super_args = (options.include? :super_args) ? options.delete(:super_args) : []
      super_block = (options.include? :super_block) ? options.delete(:super_block) : nil
      super(*super_args, &super_block)
      @connect = FastCGIConnection::new(options, &block)
    end

    def conn_timeout
      @connect.conn_timeout
    end

    def conn_timeout=(new_conn_timeout)
      @connect.conn_timeout = new_conn_timeout
    end

    def max_conns
      @connect.max_conns
    end

    def max_conns=(new_max_conns)
      @connect.max_conns = new_max_conns
    end

    def retries
      @connect.retries
    end

    def retries=(new_retries)
      @connect.retries = new_retries
    end

    def retry_wait
      @connect.retry_wait
    end

    def retry_wait=(new_retry_wait)
      @connect.retry_wait = new_retry_wait
    end

    def keep_conn
      @connect.keep_conn
    end

    def keep_conn=(new_keep_conn)
      @connect.keep_conn = new_keep_conn
    end

    def warm_up(log)
      super
      log.debug("#{self}: FastCGI connection parameter: conn_timeout -> #{self.conn_timeout}")
      log.debug("#{self}: FastCGI connection parameter: max_conns -> #{self.max_conns}")
      log.debug("#{self}: FastCGI connection parameter: retries -> #{self.retries}")
      log.debug("#{self}: FastCGI connection parameter: retry_wait -> #{self.retry_wait}")
      log.debug("#{self}: FastCGI connection parameter: keep_conn -> #{self.keep_conn}")
      @connect.each_sock do |socket|
	log.debug("#{self}: open FastCGI connection: #{connect_log(socket)}")
      end

      nil
    end

    def cool_down(log)
      super
      @connect.close(self, log)

      nil
    end
  end

  class FastCGIRecordReader < Reader
    include FastCGIconst
    include ChunkedReadable

    def initialize(socket=nil, document=nil, srv_opt=nil)
      super()
      @socket = socket
      @document = document
      @srv_opt = srv_opt
      @record = FastCGIRecord::new # reserved persistently
      @stdout_state = :unused	   # triplet: unused, opened, closed
      @stderr_state = :unused	   # triplet: unused, opened, closed
      @end_of_request = false	   # doublet: true, false
    end

    attr_accessor :socket
    attr_accessor :document
    attr_accessor :srv_opt

    def clear
      flush_read_buffer
      @socket = nil
      @document = nil
      @srv_opt = nil
      @record.clear		# reserved persistently
      @stdout_state = :unused	# triplet: unused, opened, closed
      @stderr_state = :unused	# triplet: unused, opened, closed
      @end_of_request = false	# doublet: true, false

      nil
    end

    # Protocol Handshake Summary
    # (1.1) [Web Server] <---{FCGI_STDOUT}---------- [FastCGI Appliation]
    # (1.2) [Web Server] <---{FCGI_STDERR}---------- [FastCGI Appliation]
    # (2)   [Web Server] <---{FCGI_END_REQUEST}----- [FastCGI Appliation]
    def read_chunk
      until (@end_of_request)
	begin
	  @record.receive_from(@socket)
	  case (@record.record_type)
	  when FCGI_STDOUT
	    if (@stdout_state == :closed) then
	      raise FastCGIProtocolError, "closed FCGI_STDOUT stream: #{@record.inspect}"
	    elsif (@record.content) then
	      @stdout_state = :opened
	      @srv_opt.doc_log_internal(@document) {
		"receive FastCGI record: FCGI_STDOUT: #{@record.inspect}"
	      }
	      return @record.content
	    else
	      @stdout_state = :closed
	      @srv_opt.doc_log_internal(@document) {
		"receive FastCGI record: End of FCGI_STDOUT: #{@record.inspect}"
	      }
	    end
	  when FCGI_STDERR
	    if (@stderr_state == :closed) then
	      raise FastCGIProtocolError, "closed FCGI_STDERR stream: #{@record.inspect}"
	    elsif (@record.content) then
	      @stderr_state = :opened
	      @srv_opt.doc_log_internal(@document) {
		"receive FastCGI record: FCGI_STDERR: #{@record.inspect}"
	      }
	      @srv_opt.doc_log_warning(@document, "stderr message: #{@record.content.inspect}")
	    else
	      @stderr_state = :closed
	      @srv_opt.doc_log_internal(@document) {
		"receive FastCGI record: End of FCGI_STDERR: #{@record.inspect}"
	      }
	    end
	  when FCGI_END_REQUEST
	    @srv_opt.doc_log_internal(@document) {
	      "receive FastCGI record: FCGI_END_REQUEST: #{@record.inspect}"
	    }
	    case (@stdout_state)
	    when :unused, :closed
	      # nothing to do.
	    when :opened
	      @srv_opt.srv_log_warning('not closed FastCGI stdout stream')
	    else
	      raise "internal error: #{@stdout_state.inspect}"
	    end
	    case (@stderr_state)
	    when :unused, :closed
	      # nothing to do.
	    when :opened
	      @srv_opt.srv_log_warning('not closed FastCGI stderr stream')
	    else
	      raise "internal error: #{@stderr_state.inspect}"
	    end
	    app_status, protocol_status, reserved = FastCGIRecord::decode_end_request(@record.content)
	    if (app_status == 0) then
	      @srv_opt.srv_log_system("FastCGI successful status -> #{app_status}")
	    else
	      @srv_opt.srv_log_warning("FastCGI abnormal status -> #{app_status}")
	    end
	    case (protocol_status)
	    when FCGI_REQUEST_COMPLETE
	      @srv_opt.srv_log_system('FastCGI request completed')
	    when FCGI_CANT_MPX_CONN
	      @srv_opt.srv_log_warning('FastCGI concurrent request rejected')
	    when FCGI_OVERLOADED
	      @srv_opt.srv_log_alert('FastCGI new request rejected for overloaded responder')
	    when FCGI_UNKNOWN_ROLE
	      @srv_opt.srv_log_error('unknown FastCGI role')
	    else
	      raise FastCGIProtocolError, "unknown FastCGI protocol status: #{protocol_status}"
	    end
	    @end_of_request = true
	  else
	    @srv_opt.doc_log_internal(@document) {
	      "receive unexpected FastCGI record: #{@record.inspect}"
	    }
	    raise FastCGIProtocolError, "unexpected FastCGI record type: #{@record.inspect}"
	  end
	ensure
	  @record.clear
	end
      end

      nil
    end

    CACHE = RecycleObjectCache::new(self)

    def self::fetch_cache(&block)
      CACHE.fetch(&block)
    end

    def self::store_cache(old_obj)
      CACHE.store(old_obj)
    end
  end

  module FastCGIRecordSendable
    include FastCGIconst
    include ChunkedTransferCoding

    # Protocol Handshake Summary
    # (1) [Web Server] ----{FCGI_BEGIN_REQUEST}--> [FastCGI Appliation]
    # (2) [Web Server] ----{FCGI_PARAMS}---------> [FastCGI Appliation]
    # (3) [Web Server] ----{FCGI_STDIN}----------> [FastCGI Appliation] (optional)
    def fcgi_send_request(srv_opt, socket, req_id, role, flags, fcgi_var, fcgi_stdin=nil)
      record = FastCGIRecord::fetch_cache

      record.record_type = FCGI_BEGIN_REQUEST
      record.request_id = req_id
      record.content = FastCGIRecord::encode_begin_request(role, flags)
      srv_opt.doc_log_internal(self) {
	"send FastCGI record: FCGI_BEGIN_REQUEST: #{record.inspect}"
      }
      record.send_to(socket)

      unless (fcgi_var.empty?) then
	record.clear
	record.record_type = FCGI_PARAMS
	record.request_id = req_id
	record.content = FastCGIRecord::encode_name_value_pairs(fcgi_var)
	srv_opt.doc_log_internal(self) {
	  "send FastCGI record: FCGI_PARAMS: #{record.inspect}"
	}
	record.send_to(socket)
      end

      record.clear
      record.record_type = FCGI_PARAMS
      record.request_id = req_id
      record.content = nil
      srv_opt.doc_log_internal(self) {
	"send FastCGI record: End of FCGI_PARAMS: #{record.inspect}"
      }
      record.send_to(socket)

      if (role == FCGI_FILTER) then
	if (fcgi_stdin) then
	  record.clear
	  record.recored_type = FCGI_STDIN
	  record.request_id = req_id
	  record.content = fcgi_stdin
	  srv_opt.doc_log_internal(self) {
	    "send FastCGI record: FCGI_STDIN: #{record.inspect}"
	  }
	  record.send_to(socket)
	end

	record.clear
	record.record_type = FCGI_STDIN
	record.request_id = req_id
	record.content = nil
	srv_opt.doc_log_internal(self) {
	  "send FastCGI record: End of FCGI_STDIN: #{record.inspect}"
	}
	record.send_to(socket)
      end

      FastCGIRecord::store_cache(record)
      socket.flush

      nil
    end
    module_function :fcgi_send_request

    # Protocol Handshake Summary
    # (1) [Web Server] ----{FCGI_STDIN}----------> [FastCGI Appliation]
    def fcgi_send_stdin(srv_opt, socket, req_id, request)
      record = FastCGIRecord::fetch_cache
      record.record_type = FCGI_STDIN
      record.request_id = req_id
      begin
	if (request.chunked?) then
	  # NOTE: If FastCGI I/O format is same as CGI/1.1, request
	  # trailer is ignored because CGI/1.1 has no interface to
	  # pass a trailer to a CGI application.
	  while (record.content = request.body.read_chunk)
	    srv_opt.doc_log_internal(self) {
	      "send FastCGI record: FCGI_STDIN: #{record.inspect}"
	    }
	    record.send_to(socket)
	  end
	else
	  while (record.content = request.body.read(IO_UNIT_SIZE))
	    srv_opt.doc_log_internal(self) {
	      "send FastCGI record: FCGI_STDIN: #{record.inspect}"
	    }
	    record.send_to(socket)
	  end
	end
      ensure
	record.content = nil
	srv_opt.doc_log_internal(self) {
	  "send FastCGI record: End of FCGI_STDIN: #{record.inspect}"
	}
	record.send_to(socket)
	socket.flush
      end
      FastCGIRecord::store_cache(record)

      nil
    end
    module_function :fcgi_send_stdin

    # Protocol Handshake Summary
    # (1) [Web Server] ----{FCGI_STDIN}----------> [FastCGI Appliation]
    def fcgi_send_data(srv_opt, socket, req_id, content)
      record = FastCGIRecord::fetch_cache

      record.record_type = FCGI_DATA
      record.request_id = req_id
      record.content = content
      srv_opt.doc_log_internal(self) {
	if (content) then
	  "send FastCGI record: FCGI_DATA: #{record.inspect}"
	else
	  "send FastCGI record: End of FCGI_DATA: #{record.inspect}"
	end
      }
      record.send_to(socket)

      FastCGIRecord::store_cache(record)
      socket.flush

      nil
    end
    private :fcgi_send_data
  end

  class FastCGIResponder < Document
    # Protocol Handshake Summary
    # (1)   [Web Server] ----{FCGI_BEGIN_REQUEST}--> [FastCGI Appliation]
    # (2)   [Web Server] ----{FCGI_PARAMS}---------> [FastCGI Appliation]
    # (3.1) [Web Server] ----{FCGI_STDIN}----------> [FastCGI Appliation]
    # (3.2) [Web Server] <---{FCGI_STDOUT}---------- [FastCGI Appliation]
    # (3.3) [Web Server] <---{FCGI_STDERR}---------- [FastCGI Appliation]
    # (4)   [Web Server] <---{FCGI_END_REQUEST}----- [FastCGI Appliation]

    include Path
    include PassAuth
    include FastCGIconst
    include FastCGIConnectable
    include FastCGIRecordSendable

    def respond(srv_opt, prefix, request, response)
      srv_opt.doc_log_debug(self, 'enter')
      @connect.session(srv_opt) do |socket, id_pool|
	id_pool.use_id do |req_id|
	  srv_opt.doc_log_debug(self) {
	    "establish session: #{connect_log(socket)}: request id -> #{req_id}"
	  }
	  fcgi_var = fetch_cgi_var(request, Hash::new, prefix)
	  unless (fcgi_var.include? 'CONTENT_LENGTH') then
	    # streaming message body length is unknown.
	    # specification of responder role of FastCGI protocol dose not consider streaming message body.
	    fcgi_var['CONTENT_LENGTH'] = '2147483647'
	  end
	  request.each_symbol do |param_name, param_value|
	    if (param_name.to_s =~ /^F?CGI_/i) then
	      fcgi_param_name = param_name.sub(/^F?CGI_/i, '')
	      fcgi_param_name.upcase!
	      fcgi_var[fcgi_param_name] = param_value
	    end
	  end

	  # request options
	  fcgi_flags = 0
	  fcgi_flags |= FCGI_KEEP_CONN if self.keep_conn

	  # (1)   [Web Server] ----{FCGI_BEGIN_REQUEST}--> [FastCGI Appliation]
	  # (2)   [Web Server] ----{FCGI_PARAMS}---------> [FastCGI Appliation]
	  fcgi_send_request(srv_opt, socket, req_id, FCGI_RESPONDER, fcgi_flags, fcgi_var)

	  # (3.2) [Web Server] <---{FCGI_STDOUT}---------- [FastCGI Appliation]
	  # (3.3) [Web Server] <---{FCGI_STDERR}---------- [FastCGI Appliation]
	  # (4)   [Web Server] <---{FCGI_END_REQUEST}----- [FastCGI Appliation]
	  fcgi_record_reader = FastCGIRecordReader::fetch_cache{ |new_reader|
	    new_reader.socket = socket
	    new_reader.document = self
	    new_reader.srv_opt = srv_opt
	  }

	  begin
	    # request body
	    fcgi_stdin_thread = Thread::new{
	      # (3.1) [Web Server] ----{FCGI_STDIN}----------> [FastCGI Appliation]
	      fcgi_send_stdin(srv_opt, socket, req_id, request)
	    }

	    # response head
	    if (request.version >= HTTP_1_0) then
	      response.header.parse(fcgi_record_reader)
	      if (response.header.location) then
		response.header.location = srv_opt.expand_location(request, response.header.location)
		response.status = 302
	      end
	      if (response.header.status) then
		response.status, response.reason = Response::parse_status(response.header.status)
		response.header.status = nil
	      end
	    end
	    response_header = response.header.dup
	    yield(response)
	    srv_opt[:fastcgi_unrecoverable_session] = true

	    # response body
	    if (Response::chunked? response_header) then
	      chunked_reader = ChunkedTransferCodingReader::new(fcgi_record_reader)
	      chunked_reader.each_chunk_with_ext do |chunk_data, chunk_extension|
		yield(chunk_data, chunk_extension)
	      end
	      response.trailer.update(chunked_reader.trailer)
	    else
	      if (rest_record_content = fcgi_record_reader.flush_read_buffer) then
		yield(rest_record_content)
	      end
	      fcgi_record_reader.each_chunk do |record_content|
		yield(record_content)
	      end
	    end
	  rescue ConnectionError, TimeoutError
	    # To complete this response of FastCGI application
	    fcgi_record_reader.flush_read_buffer
	    fcgi_record_reader.each_chunk do |record_content|
	      # nothing to do.
	    end
	    raise # through exception
	  ensure
	    fcgi_stdin_thread.join if fcgi_stdin_thread
	  end

	  FastCGIRecordReader::store_cache(fcgi_record_reader)
	end
      end

      nil
    end
  end

  class FastCGIAuthorizer < DocumentModifier
    # Protocol Handshake Summary
    # (1)   [Web Server] ----{FCGI_BEGIN_REQUEST}--> [FastCGI Appliation]
    # (2)   [Web Server] ----{FCGI_PARAMS}---------> [FastCGI Appliation]
    # (3.1) [Web Server] <---{FCGI_STDOUT}---------- [FastCGI Appliation]
    # (3.2) [Web Server] <---{FCGI_STDERR}---------- [FastCGI Appliation]
    # (4)   [Web Server] <---{FCGI_END_REQUEST}----- [FastCGI Appliation]

    include Path
    include FastCGIconst
    include FastCGIConnectable
    include FastCGIRecordSendable

    def initialize(document, options={}, &block)
      options[:super_args] = [ document ]
      super(options, &block)
    end

    def modify_request(srv_opt, context, prefix, request, response)
      @connect.session(srv_opt) do |socket, id_pool|
	id_pool.use_id do |req_id|
	  srv_opt.doc_log_debug(self) {
	    "establish session: #{connect_log(socket)}: request id -> #{req_id}"
	  }
	  fcgi_var = request.cgi_var(Hash::new, nil, true)
	  fcgi_var.delete('SCRIPT_NAME')
	  fcgi_var.delete('CONTENT_LENGTH')

	  # request options
	  fcgi_flags = 0
	  fcgi_flags |= FCGI_KEEP_CONN if self.keep_conn

	  # (1)   [Web Server] ----{FCGI_BEGIN_REQUEST}--> [FastCGI Appliation]
	  # (2)   [Web Server] ----{FCGI_PARAMS}---------> [FastCGI Appliation]
	  fcgi_send_request(srv_opt, socket, req_id, FCGI_AUTHORIZER, fcgi_flags, fcgi_var)

	  # (3.1) [Web Server] <---{FCGI_STDOUT}---------- [FastCGI Appliation]
	  # (3.2) [Web Server] <---{FCGI_STDERR}---------- [FastCGI Appliation]
	  # (4)   [Web Server] <---{FCGI_END_REQUEST}----- [FastCGI Appliation]
	  fcgi_record_reader = FastCGIRecordReader::fetch_cache{ |new_reader|
	    new_reader.socket = socket
	    new_reader.document = self
	    new_reader.srv_opt = srv_opt
	  }

	  begin
	    fcgi_auth_header = Header::parse(fcgi_record_reader)
	    if (fcgi_auth_header.status) then
	      status, reason = Response::parse_status(fcgi_auth_header.status)
	      if (status != 200) then
		# response head
		response.status = status
		response.reason = reason
		response.header.update(fcgi_auth_header)
		response.header.delete_if{ |field_name, field_value| field_name =~ /^Variable-/i }
		if (fcgi_auth_header.location) then
		  response.header.location = srv_opt.expand_location(request, fcgi_auth_header.location)
		end
		yield(response)
		srv_opt[:fastcgi_unrecoverable_session] = true

		# response body
		if (Response::chunked? fcgi_auth_header) then
		  chunked_reader = ChunkedTransferCodingReader::new(fcgi_record_reader)
		  chunked_reader.each_chunk_with_ext do |chunk_data, chunk_extension|
		    yield(chunk_data, chunk_extension)
		  end
		  response.trailer.update(chunked_reader.trailer)
		else
		  if (rest_record_content = fcgi_record_reader.flush_read_buffer) then
		    yield(rest_record_content)
		  end
		  fcgi_record_reader.each_chunk do |record_content|
		    yield(record_content)
		  end
		end

		# deny request
		terminate
	      end
	    end
	    # allow request

	    # skip response body
	    fcgi_record_reader.flush_read_buffer
	    fcgi_record_reader.each_chunk do |record_content|
	      # nothing to do.
	    end

	    # FastCGI variables
	    for field_name, field_value in fcgi_auth_header
	      if (field_name =~ /^Variable-/i) then
		fcgi_param_preifx, fcgi_param_name = field_name.split(/-/, 2)
		fcgi_param_name.upcase!
		request[('FCGI_' + fcgi_param_name).intern] = field_value
	      end
	    end
	  rescue ConnectionError, TimeoutError
	    # To complete this response of FastCGI application
	    fcgi_record_reader.flush_read_buffer
	    fcgi_record_reader.each_chunk do |record_content|
	      # nothing to do.
	    end
	    raise # through exception
	  end

	  FastCGIRecordReader::store_cache(fcgi_record_reader)
	end
      end

      nil
    end
  end

  class FastCGIFilter < Document
    # Protocol Handshake Summary
    # (1)   [Web Server] ----{FCGI_BEGIN_REQUEST}--> [FastCGI Appliation]
    # (2)   [Web Server] ----{FCGI_PARAMS}---------> [FastCGI Appliation]
    # (3)   [Web Server] ----{FCGI_STDIN}----------> [FastCGI Appliation]
    # (4.1) [Web Server] ----{FCGI_DATA}-----------> [FastCGI Appliation]
    # (4.2) [Web Server] <---{FCGI_STDOUT}---------- [FastCGI Appliation]
    # (4.3) [Web Server] <---{FCGI_STDERR}---------- [FastCGI Appliation]
    # (5)   [Web Server] <---{FCGI_END_REQUEST}----- [FastCGI Appliation]

    include Path
    include PassAuth
    include FastCGIconst
    include FastCGIConnectable
    include FastCGIRecordSendable
    include Delegatable

    REQUEST_BODY_LIMIT = 1024 * 32

    def initialize(document, options={}, &block)
      super(options, &block)
      @document = document
      @request_body_limit = REQUEST_BODY_LIMIT
    end

    attr_reader :document
    attr_accessor :request_body_limit

    def folder?
      @document.folder?
    end

    def respond(srv_opt, prefix, request, response)
      srv_opt.doc_log_debug(self, 'enter')

      fcgi_var = fetch_cgi_var(request, Hash::new, prefix)
      unless (fcgi_var.include? 'CONTENT_LENGTH') then
	# streaming message body length is unknown.
	# specification of filter role of FastCGI protocol dose not consider streaming message body.
	fcgi_var['CONTENT_LENGTH'] = '2147483647'
      end
      request.each_symbol do |param_name, param_value|
	if (param_name.to_s =~ /^F?CGI_/i) then
	  fcgi_param_name = param_name.sub(/^F?CGI_/i, '')
	  fcgi_param_name.upcase!
	  fcgi_var[fcgi_param_name] = param_value
	end
      end

      fcgi_stdin = request.body.read(@request_body_limit)
      read_count = 0
      while (read_data = request.body.read(IO_UNIT_SIZE))
	read_count += 1
      end
      if (read_count > 0) then
	# Request Entity Too Large
	http_error(413)
      end

      request.body = StringReader::new(fcgi_stdin && fcgi_stdin.dup)
      request.header.transfer_encoding = nil if request.chunked?

      fcgi_response = Response::new(response.version, response.status, response.header.dup, response.trailer.dup)
      for opt_key, opt_value in response
	fcgi_response[opt_key] = opt_value
      end

      @connect.session(srv_opt) do |socket, id_pool|
	id_pool.use_id do |req_id|
	  srv_opt.doc_log_debug(self) {
	    "establish session: #{connect_log(socket)}: request id -> #{req_id}"
	  }
	  fcgi_ipc_queue = Queue::new

	  fcgi_stdin_thread = Thread::new{
	    begin
	      doc_status, err_args = Document::catch_status{
		# request options
		fcgi_flags = 0
		fcgi_flags |= FCGI_KEEP_CONN if self.keep_conn

		fcgi_state = :head
		begin
		  @document.respond(srv_opt, prefix, request, response) do |arg, opts|
		    case (fcgi_state)
		    when :head
		      doc_response = arg
		      fcgi_state = :body
		      if (doc_response.header.last_modified) then
			begin
			  mtime = Time::parse(doc_response.header.last_modified)
			rescue ArgumentError
			  # nothing to do.
			else
			  fcgi_var['FCGI_DATA_LAST_MOD'] = mtime.to_i.to_s
			end
		      end
		      if (doc_response.header.content_length) then
			fcgi_var['FCGI_DATA_LENGTH'] = doc_response.header.content_length
		      else
			# streaming message body length is unknown.
			# specification of filter role of FastCGI protocol dose not consider streaming message body.
			fcgi_var['FCGI_DATA_LENGTH'] = '2147483647'
		      end
		      # (1)   [Web Server] ----{FCGI_BEGIN_REQUEST}--> [FastCGI Appliation]
		      # (2)   [Web Server] ----{FCGI_PARAMS}---------> [FastCGI Appliation]
		      # (3)   [Web Server] ----{FCGI_STDIN}----------> [FastCGI Appliation]
		      fcgi_send_request(srv_opt, socket, req_id, FCGI_FILTER, fcgi_flags, fcgi_var, fcgi_stdin)
		    when :body
		      # (4.1) [Web Server] ----{FCGI_DATA}-----------> [FastCGI Appliation]
		      fcgi_send_data(srv_opt, socket, req_id, arg) if arg
		    else
		      raise "internal error: #{fcgi_state.inspect}"
		    end
		  end
		  fcgi_state = :end
		  fcgi_send_data(srv_opt, socket, req_id, nil)
		ensure
		  case (fcgi_state)
		  when :head
		    # (1)   [Web Server] ----{FCGI_BEGIN_REQUEST}--> [FastCGI Appliation]
		    # (2)   [Web Server] ----{FCGI_PARAMS}---------> [FastCGI Appliation]
		    # (3)   [Web Server] ----{FCGI_STDIN}----------> [FastCGI Appliation]
		    fcgi_send_request(srv_opt, socket, req_id, FCGI_FILTER, fcgi_flags, fcgi_var, nil)
		    fcgi_send_data(srv_opt, socket, req_id, nil)
		  when :body
		    # (4.1) [Web Server] ----{FCGI_DATA}-----------> [FastCGI Appliation]
		    fcgi_send_data(srv_opt, socket, req_id, nil)
		  when :end
		    # nothing to do.
		  else
		    raise "internal error: #{fcgi_state.inspect}"
		  end
		end
	      }
	      if (doc_status == :error) then
		fcgi_ipc_queue.push([ :fcgi_stdin_error, err_args ])
	      end
	    ensure
	      fcgi_ipc_queue.push(:end_of_fcgi_stdin)
	    end
	  }

	  fcgi_stdout_thread = Thread::new{
	    begin
	      # (4.2) [Web Server] <---{FCGI_STDOUT}---------- [FastCGI Appliation]
	      # (4.3) [Web Server] <---{FCGI_STDERR}---------- [FastCGI Appliation]
	      # (5)   [Web Server] <---{FCGI_END_REQUEST}----- [FastCGI Appliation]
	      fcgi_record_reader = FastCGIRecordReader::fetch_cache{ |new_reader|
		new_reader.socket = socket
		new_reader.document = self
		new_reader.srv_opt = srv_opt
	      }

	      # response head
	      if (request.version >= HTTP_1_0) then
		fcgi_response.header.parse(fcgi_record_reader)
		if (fcgi_response.header.location) then
		  fcgi_response.header.location = srv_opt.expand_location(request, fcgi_response.header.location)
		  fcgi_response.status = 302
		end
		if (fcgi_response.header.status) then
		  fcgi_response.status, fcgi_response.reason = Response::parse_status(fcgi_response.header.status)
		  fcgi_response.header.status = nil
		end
	      end
	      fcgi_response_header = fcgi_response.header.dup
	      fcgi_ipc_queue.push([ :fcgi_stdout_head, fcgi_response ])

	      # response body
	      if (Response::chunked? fcgi_response_header) then
		chunked_reader = ChunkedTransferCodingReader::new(fcgi_record_reader)
		chunked_reader.each_chunk_with_ext do |chunk_data, chunk_extension|
		  fcgi_ipc_queue.push([ :fcgi_stdout_body, chunk_data, chunk_extension ])
		end
	      else
		if (rest_record_content = fcgi_record_reader.flush_read_buffer) then
		  fcgi_ipc_queue.push([ :fcgi_stdout_body, rest_record_content ])
		end
		fcgi_record_reader.each_chunk do |record_content|
		  fcgi_ipc_queue.push([ :fcgi_stdout_body, rest_record_content ])
		end
	      end

	      FastCGIRecordReader::store_cache(fcgi_record_reader)
	    ensure
	      fcgi_ipc_queue.push(:end_of_fcgi_stdout)
	    end
	  }

	  doc_error = nil
	  fcgi_stdin_closed = false
	  fcgi_stdout_closed = false
	  begin
	    fcgi_read_state, arg, opts = fcgi_ipc_queue.pop
	    case (fcgi_read_state)
	    when :fcgi_stdin_error
	      doc_error = arg
	    when :end_of_fcgi_stdin
	      fcgi_stdin_closed = true
	    when :fcgi_stdout_head
	      yield(arg) # send message head
	      srv_opt[:fastcgi_unrecoverable_session] = true
	    when :fcgi_stdout_body
	      yield(arg, opts)	# send message body
	    when :end_of_fcgi_stdout
	      fcgi_stdout_closed = true
	    else
	      raise "internal error: #{fcgi_read_state.inspect}"
	    end
	  end until (fcgi_stdin_closed && fcgi_stdout_closed)
	  fcgi_stdin_thread.value
	  fcgi_stdout_thread.value

	  if (doc_error) then
	    http_error(*doc_error)
	  end
	end
      end
    end

    delegate(Document, :document)
  end

  # alias
  FastCGIDocument = FastCGIResponder
end
