# $Id: header.rb,v 1.46.2.2 2003/01/05 13:05:54 toki Exp $
# -
# HTTP header

=begin
Copyright (c) 2001-2002
	TOKI Yoshinori <toki@freedom.ne.jp>. All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  1.  Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.

  2.  Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation
      and/or other materials provided with the distribution.

  3.  The name of the author may not be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
OF SUCH DAMAGE.
=end

require 'thread'
require 'wwwsrv/const'
require 'wwwsrv/cache'

module WWWsrv
  # latest CVS commit date of core libraries for wwwsrv
  update(%q$Date: 2003/01/05 13:05:54 $)

  class HeaderParseError < ParseError; end

  class Header
    include Enumerable

    IRREG_WORD = {
      'WWW'  => 'WWW',
      'ETAG' => 'ETag',
      'MD5'  => 'MD5',
      'TE'   => 'TE',
      'HTTP' => 'HTTP',
      'URL'  => 'URL',
      'URI'  => 'URI'
    }

    FIELD_NAME_CACHE = ImmutableObjectCache::new{ |field_name|
      field_name_segments = field_name.split(/-/, -1)
      field_name_segments.map!{ |word|
	IRREG_WORD[word.upcase] || word.capitalize
      }
      field_name_segments.join('-') # <- normalized field name
    }

    def self::normalize(field_name)
      FIELD_NAME_CACHE.fetch(field_name)
    end

    MULTI_FIELD = {}

    [ 'WWW-Authenticate',
      'Proxy-Authenticate',
      'Set-Cookie',
      'Set-Cookie2'
    ].each do |field_name|
      MULTI_FIELD[normalize(field_name)] = true
    end


    def initialize
      # NOTE: Data Structure
      # -
      # @field_map ------> { "name A",  name B", ......, "name C" }
      #                         |         |                 |
      #                         V         V                 V
      # @field_alist ---> [ ["name A",  ["name B", ...., ["name C",
      #                      "value 1",  "value 1",       "value 1",
      #                      "value 2",  "value 2",       "value 2",
      #                       ...... ],   ...... ],        ...... ] ]
      #

      @field_map = {}
      @field_alist = []
      @multi_field = {}
      @multi_field.update(MULTI_FIELD)
    end

    def clear
      @field_map.clear
      @field_alist.clear
      @multi_field.clear
      @multi_field.update(MULTI_FIELD)

      self
    end

    def clone
      new_header = super
      new_header.instance_eval do
	@field_map = @field_map.dup
	@field_alist = @field_alist.dup
	@field_alist.map!{ |field_list| field_list.dup }
	for field_list in @field_alist
	  @field_map[field_list.first] = field_list
	end
	@multi_field = @multi_field.dup
      end

      new_header
    end

    def dup
      clone
    end

    def length
      len = 0
      for field_name, *field_value_list in @field_alist
	if (@multi_field.include? field_name) then
	  len += field_value_list.length
	else
	  len += 1
	end
      end

      len
    end

    alias size length

    def empty?
      @field_alist.empty?
    end

    def set_multi_field(field_name, state=true)
      field_name = Header::normalize(field_name)
      if (state) then
	@multi_field[field_name] = true
      else
	@multi_field.delete(field_name)
      end
      self
    end

    def unset_multi_field(field_name)
      set_multi_field(field_name, false)
      self
    end

    def push(field_name, field_value)
      field_name = Header::normalize(field_name)
      if (field_value) then
	if (field_list = @field_map[field_name]) then
	  field_list.push(field_value)
	else
	  field_list = [ field_name, field_value ]
	  @field_map[field_name] = field_list
	  @field_alist.push(field_list)
	end
      end

      field_value
    end

    def pop(field_name)
      field_name = Header::normalize(field_name)
      field_list = @field_map[field_name] or return nil
      field_value = field_list.pop
      if (field_list.length == 1) then
	@field_map.delete(field_name)
	@field_alist.delete_if{ |del_field_name, *del_field_value_list|
	  del_field_name == field_name
	}
      end

      field_value
    end


    def set_first(field_name)
      field_name = Header::normalize(field_name)
      pos = nil
      @field_alist.each_with_index do |(first_field_name, *field_value_list), i|
	if (first_field_name == field_name) then
	  pos = i
	  break
	end
      end
      return nil unless pos
      first_field_list = @field_alist.delete_at(pos)
      @field_alist.unshift(first_field_list)

      self
    end

    def set_last(field_name)
      field_name = Header::normalize(field_name)
      pos = nil
      @field_alist.each_with_index do |(last_field_name, *field_value_list), i|
	if (last_field_name == field_name) then
	  pos = i
	  break
	end
      end
      return nil unless pos
      last_field_list = @field_alist.delete_at(pos)
      @field_alist.push(last_field_list)

      self
    end

    def fields(field_name)
      field_name = Header::normalize(field_name)
      field_list = @field_map[field_name] or return []
      field_list[1, field_list.length - 1]
    end

    def include?(field_name)
      field_name = Header::normalize(field_name)
      @field_map.include? field_name
    end

    alias key? include?
    alias has_key? include?

    def _delete(field_name)
      field_list = @field_map.delete(field_name) or return nil
      @field_alist.delete_if{ |del_field_name, *field_value_list|
	del_field_name == field_name
      }
      field_list.shift
      if (@multi_field.include? field_name) then
	return field_list
      else
	return field_list.join(', ')
      end
    end
    private :_delete

    def delete(field_name)
      _delete(Header::normalize(field_name))
    end

    def delete_if
      @field_alist.delete_if{ |field_list|
	field_name = field_list.first
	if (@multi_field.include? field_name) then
	  i = 0
	  field_list.delete_if{ |field_value|
	    i += 1
	    (i > 1) && yield(field_name, field_value)
	  }
	  if (field_list.length == 1) then
	    @field_map.delete(field_name)
	    true
	  end
	else
	  if (yield(field_name, field_list[1, field_list.length - 1].join(', '))) then
	    @field_map.delete(field_name)
	    true
	  end
	end
      }

      self
    end

    def _at(field_name)
      field_list = @field_map[field_name] or return nil
      if (field_list.length == 2) then
	return field_list[1]
      else
	return field_list[1, field_list.length - 1].join(', ')
      end
    end
    private :_at

    def _set(field_name, field_value)
      field_value or return _delete(field_name)
      if (field_list = @field_map[field_name]) then
	if (field_list.length == 2) then
	  field_list[1] = field_value
	else
	  field_list[1, field_list.length - 1] = [ field_value ]
	end
      else
	field_list = [ field_name, field_value ]
	@field_map[field_name] = field_list
	@field_alist.push(field_list)
      end

      field_value
    end
    private :_set

    def [](field_name)
      _at(Header::normalize(field_name))
    end

    def []=(field_name, field_value)
      _set(Header::normalize(field_name), field_value)
    end

    def update(other_header)
      for field_name, field_value in other_header
	push(field_name, field_value)
      end

      self
    end

    def self::define_field(field_name)
      field_name = normalize(field_name)
      field_attr = field_name.downcase
      field_attr.gsub!(/-/, '_')

      class_eval %Q{
	def #{field_attr}
	  _at('#{field_name}')
	end

	def #{field_attr}=(field_value)
	  _set('#{field_name}', field_value)
	end
      }

      nil
    end

    # definition of accessors of header fields
    [ # for HTTP/1.1 (RFC2616)
      'Accept',
      'Accept-Charset',
      'Accept-Encoding',
      'Accept-Language',
      'Accept-Ranges',
      'Age',
      'Allow',
      'Authorization',
      'Cache-Control',
      'Connection',
      'Content-Encoding',
      'Content-Language',
      'Content-Length',
      'Content-Location',
      'Content-MD5',
      'Content-Range',
      'Content-Type',
      'Date',
      'ETag',
      'Expect',
      'Expires',
      'From',
      'Host',
      'If-Match',
      'If-Modified-Since',
      'If-None-Match',
      'If-Range',
      'If-Unmodified-Since',
      'Keep-Alive',		# for HTTP/1.0
      'Last-Modified',
      'Location',
      'Max-Forwards',
      'Pragma',
      'Proxy-Authenticate',
      'Proxy-Authorization',
      'Range',
      'Referer',
      'Retry-After',
      'Server',
      'TE',
      'Trailer',
      'Transfer-Encoding',
      'Upgrade',
      'User-Agent',
      'Vary',
      'Via',
      'Warning',
      'WWW-Authenticate',
      # for CGI/1.1
      'Status',
      # for E-Mail (RFC822)
      'Return-Path',
      'Received',
      'Reply',
      'From',
      'Sender',
      'From',
      'Resent-Reply-To',
      'Resent-From',
      'Resent-Sender',
      'Resent-From',
      'Date',
      'Resent-Date',
      'To',
      'Resent-To',
      'Cc',
      'Resent-Cc',
      'Bcc',
      'Resent-Bcc',
      'Message-ID',
      'Resent-Message-ID',
      'In-Reply-To',
      'References',
      'Keywords',
      'Subject',
      'Comments',
      'Encrypted',
      # for RFC1806
      'Content-Disposition',
      # for MIME (RFC2045)
      'MIME-Version',
      'Content-Type',
      'Content-Transfer-Encoding',
      'Content-ID',
      'Content-Description'
    ].map{ |field_name|
      normalize(field_name)
    }.uniq.each do |field_name|
      define_field(field_name)
    end

    def each
      for field_name, *field_value_list in @field_alist
	if (@multi_field.include? field_name) then
	  for field_value in field_value_list
	    yield(field_name, field_value)
	  end
	else
	  yield(field_name, field_value_list.join(', '))
	end
      end

      self
    end

    def each_key
      for field_name, *field_value_list in @field_alist
	yield(field_name)
      end

      self
    end

    def keys
      @field_alist.map{ |field_name, *field_value_list| field_name }
    end

    def to_s
      header_message = ''
      for field_name, field_value in self
	header_message.concat(field_name)
	header_message.concat(': ')
	header_message.concat(field_value)
	header_message.concat(CRLF)
      end

      header_message
    end

    def ==(other)
      left_header_alist = self.to_a
      left_header_alist.sort!

      right_header_alist = other.to_a
      right_header_alist.sort!

      left_header_alist == right_header_alist
    end

    def parse(input)
      field_name = nil
      field_value = nil
      loop do
	line = input.gets(LF) or raise HeaderParseError, 'closed input stream'
	line.chomp!(LF)
	line.chomp!(CR)
	line.empty? and break
	if (line !~ /^\s/) then
	  field_value.strip! if field_value # strip previous field value
	  field_name, field_value = line.split(/:/, 2)
	  unless ($&) then
	    raise HeaderParseError, "failed to parse HTTP header field: #{line.inspect}"
	  end
	  field_name = Header::normalize(field_name)
	  if (field_list = @field_map[field_name]) then
	    field_list.push(field_value)
	  else
	    field_list = [ field_name, field_value ]
	    @field_map[field_name] = field_list
	    @field_alist.push(field_list)
	  end
	else
	  unless (field_name) then
	    raise HeaderParseError, "failed to parse HTTP header field: #{line.inspect}"
	  end
	  field_value.concat(line)
	end
      end
      field_value.strip! if field_value # strip last field value

      self
    end

    CACHE = RecycleObjectCache::new(self)

    def self::fetch_cache(&block)
      CACHE.fetch(&block)
    end

    def self::store_cache(old_obj)
      CACHE.store(old_obj)
    end

    def self::parse(input, header=self::new)
      header.parse(input)
    end

    def self::build(head_map={}, header=self::new)
      for field_name, field_value in head_map
	header.push(field_name, field_value)
      end

      header
    end
  end
end
