# $Id: ioutil.rb,v 1.17.2.1 2003/01/05 13:05:57 toki Exp $
# -
# I/O utilities

=begin
Copyright (c) 2001-2002
	TOKI Yoshinori <toki@freedom.ne.jp>. All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  1.  Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.

  2.  Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation
      and/or other materials provided with the distribution.

  3.  The name of the author may not be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
OF SUCH DAMAGE.
=end

require 'socket'
require 'wwwsrv/const'
require 'wwwsrv/delegate'

module WWWsrv
  # latest CVS commit date of core libraries for wwwsrv
  update(%q$Date: 2003/01/05 13:05:57 $)

  IO_UNIT_SIZE = 1024 * 4

  module IOunit
    def io_unit(stat, unit_size=IO_UNIT_SIZE)
      chunk_size = (stat.blksize > 0) ? stat.blksize : unit_size
      chunk_size *= 2 while (chunk_size < unit_size)
      chunk_size
    end
    module_function :io_unit
  end

  module ConfirmedRead
    def confirmed_read(input, length)
      if (string = input.read(length)) then
	while (string.length < length)
	  remaining = input.read(length - string.length) or break
	  string.concat(remaining)
	end
	return string
      else
	return nil
      end
    end
    module_function :confirmed_read
  end

  module ConfirmedWrite
    def confirmed_write(output, string)
      loop do
	write_bytes = output.write(string)
	if (write_bytes < string.length) then
	  string = string[write_bytes..-1]
	else
	  break
	end
      end

      nil
    end
    module_function  :confirmed_write
  end

  class ProtectedSocket
    include Delegatable

    def initialize(socket, except=ConnectionError)
      @socket = socket
      @except = except
    end

    def to_io
      @socket
    end

    def __wrap_io__(id, *args)
      begin
	return yield(args)
      rescue TimeoutError
	# through timeout error
	raise
      rescue StandardError
	raise @except, $!.inspect, $!.backtrace
      end
    end
    private :__wrap_io__

    delegate(TCPsocket, :to_io, :__wrap_io__)
  end

  class ProtectedServerSocket
    include Delegatable

    def initialize(socket, except=ConnectionError)
      @socket = socket
      @except = except
    end

    def to_io
      @socket
    end

    def __wrap_io__(id, *args)
      begin
	return yield(args)
      rescue TimeoutError
	# through timeout error
	raise
      rescue StandardError
	raise @except, $!.inspect, $!.backtrace
      end
    end
    private :__wrap_io__

    delegate(TCPserver, :to_io, :__wrap_io__)
  end

  module SocketManipulation
    def set_tcp_no_delay(socket)
      # socket may be wrapped.
      if (raw_sock = socket.to_io) then
	if (raw_sock.kind_of? TCPsocket) then
	  socket.setsockopt(Socket::IPPROTO_TCP, Socket::TCP_NODELAY, 1)
	end
      end

      socket
    end
    module_function :set_tcp_no_delay

    def socket_address(socket, peer=false)
      accessor = peer ? :peeraddr : :addr
      # socket may be wrapped.
      if ((raw_sock = socket.to_io) && (raw_sock.respond_to? accessor)) then
	begin
	  case (socket)
	  when IPsocket
	    family, port, name, addr = socket.__send__(accessor)
	    return [ family, name + '(' + addr + ')', port ].join(':')
	  else
	    addr_list = socket.__send__(accessor)
	    family = addr_list.shift
	    addr_list.map!{ |addr_elem| addr_elem.inspect }
	    addr_list.unshift(family)
	    return addr_list.join(':')
	  end
	rescue
	  return "[failed to get an address from #{socket.inspect}]"
	end
      else
	return socket.inspect
      end
    end
    module_function :socket_address

    def connect_log(socket)
      # socket may be wrapped.
      if ((raw_sock = socket.to_io) && (raw_sock.respond_to? :addr) && (raw_sock.respond_to? :peeraddr)) then
	begin
	  case (raw_sock)
	  when IPsocket
	    here_family, here_port, here_name, here_addr = socket.addr
	    there_family, there_port, there_name, there_addr = socket.peeraddr
	    return [
	      here_family,
	      here_name + '(' + here_addr + ')',
	      here_port
	    ].join(':') + ' <-> ' + [
	      there_family,
	      there_name + '(' + there_addr + ')',
	      there_port
	    ].join(':')
	  else
	    here_addr_list = socket.addr
	    here_family = here_addr_list.shift
	    here_addr_list.map!{ |addr_elem| addr_elem.inspect }
	    here_addr_list.unshift(here_family)

	    there_addr_list = socket.peeraddr
	    there_family = there_addr_list.shift
	    there_addr_list.map!{ |addr_elem| addr_elem.inspect }
	    there_addr_list.unshift(there_family)

	    return here_addr_list.join(':') + ' <-> ' + there_addr_list.join(':')
	  end
	rescue
	  return "[failed to get addresses from #{socket.inspect}]"
	end
      else
	return socket.inspect
      end
    end
    module_function :connect_log
  end
end
