# $Id: logger.rb,v 1.20 2002/11/22 04:38:13 toki Exp $
# -
# logging utilities

=begin
Copyright (c) 2001-2002
	TOKI Yoshinori <toki@freedom.ne.jp>. All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  1.  Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.

  2.  Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation
      and/or other materials provided with the distribution.

  3.  The name of the author may not be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
OF SUCH DAMAGE.
=end

require 'wwwsrv/const'
require 'wwwsrv/ioutil'

module WWWsrv
  # latest CVS commit date of core libraries for wwwsrv
  update(%q$Date: 2002/11/22 04:38:13 $)

  class Logger
    # logging level
    INTERNAL = 0
    DEBUG    = 1
    SYSTEM   = 2
    MESSAGE  = 3
    WARNING  = 4
    NOTICE   = 5
    ALERT    = 6
    ERROR    = 7

    def initialize(output=STDOUT, level=MESSAGE)
      @output = output
      @level = level
    end

    def print(priority, *messages)
      if (priority >= @level) then
	message = "#{Time::now}: pid #{$$}: "
	if (messages.empty?) then
	  message.concat(yield)
	else
	  for message_fragment in messages
	    message.concat(message_fragment)
	  end
	end
	message.concat("\n")
	@output.write(message)
      end

      nil
    end

    def internal(*msgs, &block)
      print(INTERNAL, *msgs, &block)
    end

    def debug(*msgs, &block)
      print(DEBUG, *msgs, &block)
    end

    def system(*msgs, &block)
      print(SYSTEM, *msgs, &block)
    end

    def message(*msgs, &block)
      print(MESSAGE, *msgs, &block)
    end

    def warning(*msgs, &block)
      print(WARNING, *msgs, &block)
    end

    def notice(*msgs, &block)
      print(NOTICE, *msgs, &block)
    end

    def alert(*msgs, &block)
      print(ALERT, *msgs, &block)
    end

    def error(*msgs, &block)
      print(ERROR, *msgs, &block)
    end

    def close
      @output.close
      self
    end

    def self::backtrace(logger, logging_level, sid, except)
      except.backtrace.each_with_index do |frame, i|
	if (i < 10) then
	  logger.print(logging_level, "sid #{sid}: backtrace[#{i}]: #{frame}")
	else
	  logger.print(logging_level, "sid #{sid}: backtrace[#{i}]: ...")
	  break
	end
      end

      nil
    end

    def self::srv_format(sid, msg)
      "sid #{sid}: #{msg}"
    end

    def self::doc_format(sid, doc, msg)
      "sid #{sid}: #{doc}: #{msg}"
    end
  end

  class PreservedLogger < Logger
    def close
      self
    end
  end

  class MultiLogger < Logger
    def initialize
      @log_vec = []
    end

    def add_logger(log)
      @log_vec.push(log)
    end

    def print(priority, *messages, &block)
      for log in @log_vec
	log.print(priority, *messages, &block)
      end

      nil
    end

    def close
      for log in @log_vec
	log.close
      end

      nil
    end
  end

  class ConnectionWatcher
    def watch_accept_start(sid, srv_sock)
      nil
    end

    def watch_accept_error(sid, srv_sock, except)
      nil
    end

    def watch_connect_open(sid, socket)
      nil
    end

    def watch_connect_keep_alive(sid, socket)
      nil
    end

    def watch_connect_close(sid, socket)
      nil
    end

    def watch_connect_count(sid, socket, count)
      nil
    end

    def watch_connect_error(sid, socket, except)
      nil
    end
  end

  class RequestWatcher
    def watch_request_head(sid, request)
      nil
    end

    def watch_request_body(sid, body_fragment, body_opts=nil)
      nil
    end

    def watch_request_trailer(sid, request_trailer)
      nil
    end
  end

  class ResponseWatcher
    def watch_response_head(sid, response)
      nil
    end

    def watch_response_body(sid, body_fragment, body_opts=nil)
      nil
    end

    def watch_response_trailer(sid, response_trailer)
      nil
    end
  end

  class MultiConnectionWatcher < ConnectionWatcher
    def initialize
      super()
      @connect_watcher_list = []
    end

    def add_connect_watcher(watcher)
      @connect_watcher_list.push(watcher)
    end

    def watch_accept_start(sid, srv_sock)
      for connect_watcher in @connect_watcher_list
	connect_watcher.watch_accept_start(sid, srv_sock)
      end

      nil
    end

    def watch_accept_error(sid, srv_sock, except)
      for connect_watcher in @connect_watcher_list
	connect_watcher.watch_accept_error(sid, srv_sock, except)
      end

      nil
    end

    def watch_connect_open(sid, socket)
      for connect_watcher in @connect_watcher_list
	connect_watcher.watch_connect_open(sid, socket)
      end

      nil
    end

    def watch_connect_keep_alive(sid, socket)
      for connect_watcher in @connect_watcher_list
	connect_watcher.watch_connect_keep_alive(sid, socket)
      end

      nil
    end

    def watch_connect_close(sid, socket)
      for connect_watcher in @connect_watcher_list
	connect_watcher.watch_connect_close(sid, socket)
      end

      nil
    end

    def watch_connect_count(sid, socket, count)
      for connect_watcher in @connect_watcher_list
	connect_watcher.watch_connect_count(sid, socket, count)
      end

      nil
    end

    def watch_connect_error(sid, socket, except)
      for connect_watcher in @connect_watcher_list
	connect_watcher.watch_connect_error(sid, socket, except)
      end

      nil
    end
  end

  class MultiRequestWatcher < RequestWatcher
    def initialize
      super()
      @request_watcher_list = []
    end

    def add_request_watcher(watcher)
      @request_watcher_list.push(watcher)
    end

    def watch_request_head(sid, request)
      for request_watcher in @request_watcher_list
	request_watcher.watch_request_head(sid, request)
      end

      nil
    end

    def watch_request_body(sid, body_fragment, body_opts=nil)
      for request_watcher in @request_watcher_list
	request_watcher.watch_request_body(sid, body_fragment, body_opts)
      end

      nil
    end

    def watch_request_trailer(sid, request_trailer)
      for request_watcher in @request_watcher_list
	request_watcher.watch_request_trailer(sid, request_trailer)
      end

      nil
    end
  end

  class MultiResponseWatcher < ResponseWatcher
    def initialize
      super()
      @response_watcher_list = []
    end

    def add_response_watcher(watcher)
      @response_watcher_list.push(watcher)
    end

    def watch_response_head(sid, response)
      for response_watcher in @response_watcher_list
	response_watcher.watch_response_head(sid, response)
      end

      nil
    end

    def watch_response_body(sid, body_fragment, body_opts=nil)
      for response_watcher in @response_watcher_list
	response_watcher.watch_response_body(sid, body_fragment, body_opts)
      end

      nil
    end

    def watch_response_trailer(sid, response_trailer)
      for response_watcher in @response_watcher_list
	response_watcher.watch_response_trailer(sid, response_trailer)
      end

      nil
    end
  end

  class ConnectionLogger < ConnectionWatcher
    include SocketManipulation

    def initialize(logger,
		   accept_logging_priority=Logger::NOTICE,
		   open_logging_priority=Logger::MESSAGE,
		   keep_alive_logging_priority=Logger::SYSTEM,
		   close_logging_priority=Logger::SYSTEM,
		   count_logging_priority=Logger::SYSTEM,
		   error_logging_priority=Logger::WARNING,
		   backtrace_logging_priority=Logger::DEBUG)
      @log = logger
      @accept_logging_priority = accept_logging_priority
      @open_logging_priority = open_logging_priority 
      @keep_alive_logging_priority = keep_alive_logging_priority
      @close_logging_priority = close_logging_priority
      @count_logging_priority = count_logging_priority
      @error_logging_priority = error_logging_priority
      @backtrace_logging_priority = backtrace_logging_priority
    end

    attr_accessor :accept_logging_priority
    attr_accessor :open_logging_priority
    attr_accessor :keep_alive_logging_priority
    attr_accessor :close_logging_priority
    attr_accessor :count_logging_priority
    attr_accessor :error_logging_priority
    attr_accessor :backtrace_logging_priority

    def watch_accept_start(sid, srv_sock)
      @log.print(@accept_logging_priority) {
	Logger::srv_format(sid, "start server: #{socket_address(srv_sock)}")
      }
      nil
    end

    def watch_accept_error(sid, srv_sock, except)
      @log.print(@error_logging_priority, Logger::srv_format(sid, "accept error: #{$!}"))
      Logger::backtrace(@log, @backtrace_logging_priority, sid, except)
      nil
    end

    def watch_connect_open(sid, socket)
      @log.print(@open_logging_priority) {
	Logger::srv_format(sid, "open connection: #{connect_log(socket)}")
      }
      nil
    end

    def watch_connect_keep_alive(sid, socket)
      @log.print(@keep_alive_logging_priority, Logger::srv_format(sid, 'keep alive'))
      nil
    end

    def watch_connect_close(sid, socket)
      @log.print(@close_logging_priority, Logger::srv_format(sid, 'close connection'))
      nil
    end

    def watch_connect_count(sid, socket, count)
      @log.print(@count_logging_priority, Logger::srv_format(sid, "connections: #{count}"))
      nil
    end

    def watch_connect_error(sid, socket, except)
      @log.print(@error_logging_priority, Logger::srv_format(sid, "connection error: #{except} (#{except.type})"))
      Logger::backtrace(@log, @backtrace_logging_priority, sid, except)
      nil
    end
  end

  class RequestLogger < RequestWatcher
    def initialize(logger,
		   request_line_logging_priority=Logger::MESSAGE,
		   request_header_logging_priority=Logger::DEBUG,
		   request_body_logging_priority=Logger::INTERNAL,
		   request_trailer_logging_priority=Logger::DEBUG)
      super()
      @log = logger
      @request_line_logging_priority = request_line_logging_priority
      @request_header_logging_priority = request_header_logging_priority
      @request_body_logging_priority = request_body_logging_priority
      @request_trailer_logging_priority = request_trailer_logging_priority
    end

    attr_accessor :request_line_logging_priority
    attr_accessor :request_header_logging_priority
    attr_accessor :request_body_logging_priority
    attr_accessor :request_trailer_logging_priority

    def watch_request_head(sid, request)
      @log.print(@request_line_logging_priority,
		 Logger::srv_format(sid, "request line: #{request.line}"))

      for field_name, field_value in request.header
	@log.print(@request_header_logging_priority,
		   Logger::srv_format(sid, "request header: #{field_name}: #{field_value}"))
      end

      nil
    end

    def watch_request_body(sid, body_fragment, body_opts=nil)
      @log.print(@request_body_logging_priority) {
	msg = 'receiving ' + (body_fragment ? body_fragment.length.to_s : '0') + ' octets'
	msg.concat(': options -> ' + body_opts.join(', ')) if (body_opts && ! body_opts.empty?)
	msg.concat(': message -> ' + body_fragment.inspect)
	Logger::srv_format(sid, msg)
      }

      nil
    end

    def watch_request_trailer(sid, request_trailer)
      for field_name, field_value in request_trailer
	@log.print(@request_trailer_logging_priority,
		   Logger::srv_format(sid, "request trailer: #{field_name}: #{field_value}"))
      end

      nil
    end
  end

  class ResponseLogger < ResponseWatcher
    def initialize(logger,
		   status_line_logging_priority=Logger::SYSTEM,
		   response_header_logging_priority=Logger::DEBUG,
		   response_body_logging_priority=Logger::INTERNAL,
		   response_trailer_logging_priority=Logger::DEBUG)
      super()
      @log = logger
      @status_line_logging_priority = status_line_logging_priority
      @response_header_logging_priority = response_header_logging_priority
      @response_body_logging_priority = response_body_logging_priority
      @response_trailer_logging_priority = response_trailer_logging_priority
    end

    attr_accessor :status_line_logging_priority
    attr_accessor :response_header_logging_priority
    attr_accessor :response_body_logging_priority
    attr_accessor :response_trailer_logging_priority

    def watch_response_head(sid, response)
      @log.print(@status_line_logging_priority,
		 Logger::srv_format(sid, "status line: #{response.line}"))

      for field_name, field_value in response.header
	@log.print(@response_header_logging_priority,
		   Logger::srv_format(sid, "response header: #{field_name}: #{field_value}"))
      end

      nil
    end

    def watch_response_body(sid, body_fragment, body_opts=nil)
      @log.print(@response_body_logging_priority) {
	msg = 'sending ' + (body_fragment ? body_fragment.length.to_s : '0') + ' octets'
	msg.concat(': options -> ' + body_opts.join(', ')) if (body_opts && ! body_opts.empty?)
	msg.concat(': message -> ' + body_fragment.inspect)
	Logger::srv_format(sid, msg)
      }

      nil
    end

    def watch_response_trailer(sid, response_trailer)
      for field_name, field_value in response_trailer
	@log.print(@response_trailer_logging_priority,
		   Logger::srv_format(sid, "response trailer: #{field_name}: #{field_value}"))
      end

      nil
    end
  end
end
