# $Id: multipart.rb,v 1.14.2.2 2003/04/11 14:09:04 toki Exp $
# -
# methods of parsing multipart

=begin
Copyright (c) 2001-2002
	TOKI Yoshinori <toki@freedom.ne.jp>. All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  1.  Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.

  2.  Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation
      and/or other materials provided with the distribution.

  3.  The name of the author may not be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
OF SUCH DAMAGE.
=end

require 'wwwsrv/const'
require 'wwwsrv/token'
require 'wwwsrv/header'
require 'wwwsrv/reader'

module WWWsrv
  # latest CVS commit date of core libraries for wwwsrv
  update(%q$Date: 2003/04/11 14:09:04 $)

  class MultipartParseError < ParseError; end

  class MultipartParser
    # NOTE: stack state chart
    # -
    # [push] := push stack
    # [pop] := pop stack
    # -
    # 1. :multipart_preamble
    # 2. :part_head
    # 3. if multipart/* -> :multipart_body
    #                        :multipart_preamble [push]
    #                        ...(nested multipart)...
    #                        :multipart_epilogue [pop]
    #    else           -> :part_body
    # 4. :part_head(2)
    # 5. if multipart/* -> :multipart_body
    #                        :multipart_preamble [push]
    #                        ...(nested multipart)...
    #                        :multipart_epilogue [pop]
    #    else           -> :part_body
    # ...
    # 6. :multipart_epilogue

    # NOTE: examples
    # -
    # [1]
    # 1. :multipart_preamble
    # 2. :multipart_epilogue
    # [2]
    # 1. :multipart_preamble
    # 2. :part_head
    # 3. :part_body
    # 4. ...
    # 5. :multipart_epilogue
    # [3]
    # 1. :multipart_preamble
    # 2. :part_head
    # 3  :multipart_body
    # 3.1  :multipart_preamble
    # 3.2  ...
    # 3.3  :multipart_epilogue
    # 4. ...
    # 5. :multipart_epilogue

    def chomp_CRLF(line)
      line.chomp(LF).chomp(CR)
    end
    private :chomp_CRLF

    def multipart?(content_type)
      content_type && content_type =~ %r"^multipart/"i
    end
    private :multipart?

    def get_boundary(content_type)
      case (content_type)
      when /boundary="([^"]+)"/i
	boundary = $1
      when /boundary=([^\s]+)/i
	boundary = $1
      else
	raise MultipartParseError, "not found a multipart boundary: #{content_type.inspect}"
      end

      boundary
    end
    private :get_boundary

    def initialize(content_type)
      unless (multipart? content_type) then
	raise MultipartParseError, "not multipart Content-Type: #{content_type.inspect}"
      end
      @state = :multipart_preamble
      @closed = false
      @stopped = false
      @boundary = get_boundary(content_type)
      @boundary_begin = '--' + @boundary
      @boundary_end = '--' + @boundary + '--'
      @head = nil
      @head_reader = nil
      @last_field_name = nil
      @last_line = nil
      @parser_list = Array.new
    end

    def init_part_head
      @head = Header.new
      @head_reader = StringReader.new
      @last_field_name = nil
      nil
    end
    private :init_part_head

    def clear_part_head
      @head = nil
      @head_reader = nil
      @last_field_name = nil
      nil
    end
    private :clear_part_head

    def scan_head(line)
      header_field = chomp_CRLF(line)
      case (header_field)
      when @boundary_begin
	yield(:part_head, @head)
	yield(:part_end, nil)
	init_part_head
	@state = :part_head
      when @boundary_end
	yield(:part_head, @head)
	yield(:part_end, nil)
	yield(:end_multipart, nil)
	clear_part_head
	@state = :multipart_epilogue
      else
	@head_reader << line
	if (header_field.empty?) then
	  @head.parse(@head_reader)
	  yield(:part_head, @head)
	  if (multipart? @head.content_type) then
	    @state = :multipart_body
	    @parser_list.push(MultipartParser.new(@head.content_type))
	  else
	    @state = :part_body
	    @last_line = nil
	  end
	end
      end

      nil
    end
    private :scan_head

    def flush_part_body
      if (@last_line) then
	part_bottom = chomp_CRLF(@last_line)
	unless (part_bottom.empty?) then
	  yield(:part_body, part_bottom)
	end
	@last_line = nil
      end

      nil
    end
    private :flush_part_body

    def scan_body(line)
      case (chomp_CRLF(line))
      when @boundary_begin
	flush_part_body do |tag, value|
	  yield(tag, value)
	end
	yield(:part_end, nil)
	init_part_head
	@state = :part_head
      when @boundary_end
	flush_part_body do |tag, value|
	  yield(tag, value)
	end
	yield(:part_end, nil)
	yield(:end_multipart, nil)
	@state = :multipart_epilogue
      else
	if (@last_line) then
	  unless (@last_line.empty?) then
	    yield(:part_body, @last_line)
	  end
	end
	@last_line = line
      end

      nil
    end
    private :scan_body

    def scan_multipart_preamble(line)
      case (chomp_CRLF(line))
      when @boundary_begin
	yield(:begin_multipart, nil)
	init_part_head
	@state = :part_head
      when @boundary_end
	yield(:begin_multipart, nil)
	yield(:end_multipart, nil)
	@state = :multipart_epilogue
      end

      nil
    end
    private :scan_multipart_preamble

    def scan_multipart_epilogue(line)
      # nothing to do.
      nil
    end
    private :scan_multipart_epilogue

    def scan_multipart_body(line)
      multipart_parser = @parser_list.last
      case (chomp_CRLF(line))
      when @boundary_begin
	multipart_parser.stop do |tag, value|
	  yield(tag, value)
	end
	init_part_head
	@state = :part_head
      when @boundary_end
	multipart_parser.stop do |tag, value|
	  yield(tag, value)
	end
	yield(:end_multipart, nil)
	@state = :multipart_epilogue
      else
	multipart_parser.scan(line) do |tag, value|
	  yield(tag, value)
	end
      end

      nil
    end
    private :scan_multipart_body

    def scan(line)
      if (@closed) then
	raise 'closed multipart parser'
      end
      if (@stopped) then
	raise 'stopped multipart parser'
      end

      case (@state)
      when :part_body
	scan_body(line) do |tag, value|
	  yield(tag, value)
	end
      when :part_head
	scan_head(line) do |tag, value|
	  yield(tag, value)
	end
      when :multipart_body
	scan_multipart_body(line) do |tag, value|
	  yield(tag, value)
	end
      when :multipart_preamble
	scan_multipart_preamble(line) do |tag, value|
	  yield(tag, value)
	end
      when :multipart_epilogue
	scan_multipart_epilogue(line) do |tag, value|
	  yield(tag, value)
	end
      else
	raise "unknown state: #{@state.inspect}"
      end

      # return true for parsing multipart section.
      @state != :multipart_epilogue
    end

    def stopped?
      @stopped
    end

    def stop
      if (@closed) then
	raise 'closed multipart parser'
      end
      if (@stopped) then
	raise 'stopped multipart parser'
      end
      @stopped = true

      case (@state)
      when :part_head
	yield(:part_head, @head)
	yield(:part_end, nil)
	yield(:end_multipart, nil)
	clear_part_head
      when :part_body
	if (@last_line) then
	  unless (@last_line.empty?) then
	    yield(:part_body, @last_line)
	  end
	  @last_line = nil
	end
	yield(:part_end, nil)
	yield(:end_multipart, nil)
      when :multipart_preamble
	yield(:begin_multipart, nil)
	yield(:end_multipart, nil)
      when :multipart_epilogue
	# nothing to do.
      when :multipart_body
	multipart_parser = @parser_list.last
	multipart_parser.stop do |tag, value|
	  yield(tag, value)
	end
	yield(:end_multipart, nil)
      else
	raise "unknown state: #{@state.inspect}"
      end

      nil
    end

    def closed?
      @closed
    end

    def close
      if (@closed) then
	raise 'closed multipart parser'
      end
      @closed = true

      for multipart_parser in @parser_list
	multipart_parser.close
      end
      @parser_list = nil

      if (@state != :multipart_epilogue) then
	raise MultipartParseError, "not terminated multipart body with boudary of #{@boundary.inspect}"
      end

      nil
    end

    def parse(body_input)
      for line in body_input
	scan(line) { |tag, value|
	  yield(tag, value)
	} or break
      end
      close

      nil
    end

    def self.parse(content_type, body_input)
      multipart_parser = MultipartParser.new(content_type)
      multipart_parser.parse(body_input) do |tag, value|
	yield(tag, value)
      end

      nil
    end
  end

  # for backward compatibility
  module Multipart
    def scan_multipart(input, content_type)
      MultipartParser.parse(content_type, input) do |tag, value|
	yield(tag, value)
      end

      nil
    end
    module_function :scan_multipart
  end
end
