# $Id: path.rb,v 1.18.2.1 2003/04/11 14:09:06 toki Exp $
# -
# HTTP path utilities

=begin
Copyright (c) 2001-2002
	TOKI Yoshinori <toki@freedom.ne.jp>. All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  1.  Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.

  2.  Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation
      and/or other materials provided with the distribution.

  3.  The name of the author may not be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
OF SUCH DAMAGE.
=end

require 'wwwsrv/const'

module WWWsrv
  # latest CVS commit date of core libraries for wwwsrv
  update(%q$Date: 2003/04/11 14:09:06 $)

  module Path
    # if the last character of prefix is `/', then it is ignored.
    # ex)
    # 1.  prefix => "",      path => "/foo",      subpath => "/foo"
    # 2.  prefix => "",      path => "/foo/",     subpath => "/foo/"
    # 3.  prefix => "/",     path => "/foo",      subpath => "/foo"
    # 4.  prefix => "/foo",  path => "/foo",      subpath => ""
    # 5.  prefix => "/foo",  path => "/foo/",     subpath => "/"
    # 6.  prefix => "/foo/", path => "/foo/",     subpath => "/"
    # 7.  prefix => "/foo",  path => "/foo/bar",  subpath => "/bar"
    # 8.  prefix => "/foo",  path => "/foo/bar/", subpath => "/bar/"
    # 9.  prefix => "/foo/", path => "/foo/bar",  subpath => "/bar"
    # 10. prefix => "",      path => "foo",       subpath => "foo"
    # 11. prefix => "",      path => "foo/",      subpath => "foo/"
    # 12. prefix => "foo",   path => "foo/bar",   subpath => "/bar"
    # 13. prefix => "foo",   path => "foo/bar/",  subpath => "/bar/"
    # 14. prefix => "foo/",  path => "foo/bar",   subpath => "/bar"
    def subpath(prefix, path)
      prefix = prefix.sub(%r"/$", '')
      prefix.empty? and return path

      prefix_fragments = prefix.split(%r"/", -1)
      path_fragments = path.split(%r"/", -1)

      while (! prefix_fragments.empty? && ! path_fragments.empty? && prefix_fragments.first == path_fragments.first)
	prefix_fragments.shift
	path_fragments.shift
      end

      unless (prefix_fragments.empty?) then
	raise "failed to match prefix to path: #{prefix.inspect} vs #{path.inspect}"
      end

      if (path_fragments.empty?) then
	return ''
      else
	path_fragments.unshift('')
	return path_fragments.join('/')
      end
    end
    module_function :subpath

    def scan_path_info(path)
      yield(path, '')
      script_name_fragments = path.split(%r"/", -1)
      path_info_fragments = []
      while (script_name_fragments.length > 1)
	path_info_fragments.unshift(script_name_fragments.pop)
	path_info_fragments.unshift('')
	yield(script_name_fragments.join('/'), path_info_fragments.join('/'))
	path_info_fragments.shift
      end
      unless (script_name_fragments.first.empty?) then
	path_info_fragments.unshift(script_name_fragments.pop)
	yield(script_name_fragments.join('/'), path_info_fragments.join('/'))
      end

      nil
    end
    module_function :scan_path_info

    FIRST_FSEP = Regexp::compile('^' + Regexp::quote(File::SEPARATOR))
    LAST_FSEP = Regexp::compile(Regexp::quote(File::SEPARATOR) + '$')

    # if the last character of prefix is `/', then it is ignored.
    def merge_path(prefix, subpath)
      prefix = prefix.sub(%r"/$", '')
      if (subpath =~ %r"^/") then
	return prefix + subpath
      else
	return prefix + '/' + subpath
      end
    end
    module_function :merge_path

    # if the last character of prefix is `File::SEPARATOR', then it is ignored.
    def merge_local_path(prefix, subpath)
      prefix = prefix.sub(LAST_FSEP, '')
      if (subpath =~ FIRST_FSEP) then
	return prefix + subpath
      else
	return prefix + File::SEPARATOR + subpath
      end
    end
    module_function :merge_local_path

    def path_http2local(http_path, prefix=nil)
      if (prefix) then
	return merge_local_path(prefix, http_path.gsub(%r"/", File::SEPARATOR))
      else
	return http_path.gsub(%r"/", File::SEPARATOR)
      end
    end
    module_function :path_http2local

    def expand_http_path(path)
      if (path.empty?) then
	normalized_path = '/'
      else
	normalized_path = path.dup

	# expand URL encoding
	normalized_path.gsub!(/\+/, ' ')
	normalized_path.gsub!(/%([0-9A-Fa-f][0-9A-Fa-f])/) { |fragment| $1.hex.chr }

	# expand path (from shttpsrv)
	while (normalized_path.gsub!(%r"/\./", '/'))
	  # foo/./bar => foo/bar
	end
	while (normalized_path.gsub!(%r"[^/]+/\.\./", ''))
	  # foo/../bar => bar
	end
	while (normalized_path.gsub!(%r"[^/]+/\.\.$", ''))
	  # foo/bar/.. => foo/
	end
	while (normalized_path.gsub!(%r"^/\.\./", '/'))
	  # /../foo => /foo
	end
	normalized_path.sub!(%r"/\.$", '/')  # foo/. => foo/
	normalized_path.sub!(%r"^/..$", '/') # /.. => /
      end

      normalized_path
    end
    module_function :expand_http_path

    def normalize_path(path)
      path_part, query_part = path.split(/\?/, 2)
      return expand_http_path(path_part), query_part
    end
    module_function :normalize_path

    def make_local_path_pattern(base_path)
      if (base_path =~ LAST_FSEP) then
	return Regexp::compile('^' + Regexp::quote(base_path))
      else
	return Regexp::compile('^' + Regexp::quote(base_path) + '(' + Regexp::quote(File::SEPARATOR) + '|$)')
      end
    end
    module_function :make_local_path_pattern

    def expand_location(host, request_path, location_path, scheme=HTTP_SCHEME)
      if (location_path !~ %r"^[A-Za-z][A-Za-z0-9.+-]*://") then
	location_path = request_path.sub(%r"[^/]*$", '') + location_path if (location_path !~ %r"^/")
	location_path = scheme + '://' + host.name + ':' + host.port.to_s + location_path
      end

      location_path
    end
    module_function :expand_location
  end
end
