# $Id: reader.rb,v 1.32.2.3 2003/04/11 14:09:07 toki Exp $
# -
# input stream reader

=begin
Copyright (c) 2001-2002
	TOKI Yoshinori <toki@freedom.ne.jp>. All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  1.  Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.

  2.  Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation
      and/or other materials provided with the distribution.

  3.  The name of the author may not be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
OF SUCH DAMAGE.
=end

require 'singleton'
require 'wwwsrv/const'
require 'wwwsrv/thread'
require 'wwwsrv/ioutil'
require 'wwwsrv/chunk'
require 'wwwsrv/header'

module WWWsrv
  # latest CVS commit date of core libraries for wwwsrv
  update(%q$Date: 2003/04/11 14:09:07 $)

  # NOTE: hierarchy of classes and modules
  #
  # * WWWsrv::BufferedReadable module
  #    |
  #    +- WWWsrv::ArbitraryReadable module
  #    |
  #    +- WWWsrv::ChunkedReadable module
  #
  # * WWWsrv::NullTrailerReadable module
  #
  # * WWWsrv::Reader class <--{include}--- Enumerable module
  #    |
  #    +- WWWsrv::EmptyReader class <--{include}--- Singleton module
  #    |   |
  #    |   +- WWWsrv::EmptyContentReader class <--{include}--- WWWsrv::NullTrailerReadable module
  #    |
  #    +- WWWsrv::StringReader class
  #    |
  #    +- WWWsrv::LengthLimitedReader class <--{include}--- WWWsrv::ArbitraryReadable module
  #    |   |
  #    |   +- WWWsrv::LengthLimitedContentReader <--{include}-+- WWWsrv::Timeout module
  #    |                                                      |
  #    |                                                      +- WWWsrv::NullTrailerReadable module
  #    |
  #    +- WWWsrv::ChunkedTransferCodingReader class <--{include}-+- WWWsrv::ChunkedReadable module
  #    |   |                                                     |
  #    |   |                                                     +- WWWsrv::ChunkedTransferCoding module
  #    |   |
  #    |   +- WWWsrv::ChunkedContentReader class <--{include}--- WWWsrv::Timeout module
  #    |
  #    +- WWWsrv::ContentReader class <--{include}-+- WWWsrv::Timeout module
  #                                                |
  #                                                +- WWWsrv::NullTrailerReadable module

  class Reader
    # NOTE: abstract methods
    #  * public read
    #  * public gets
    # NOTE: dependency of methods
    #  * read is called from ...
    #      - getc
    #      - each_byte
    #      - readchar
    #  * gets is called from ...
    #      - each_line
    #      - readline
    #      - readlines

    include Enumerable

    def getc
      (ch = read(1)) and return ch[0]
    end

    def each_byte
      while (ch = getc)
	yield(ch)
      end

      self
    end

    def each_line(eol=$/)
      while (line = gets(eol))
	yield(line)
      end

      self
    end

    alias each each_line

    def readchar
      getc or raise EOFError, "end of #{self.class} stream"
    end

    def readline(eol=$/)
      gets(eol) or raise EOFError, "end of #{self.class} stream"
    end

    def readlines(eol=$/)
      lines = []
      self.each_line(eol) do |line|
	lines.push(line)
      end

      lines
    end

    def to_io
      nil
    end
  end

  class EmptyReader < Reader
    include Singleton

    def read(length=nil)
      nil
    end

    def gets(eol=$/)
      nil
    end
  end

  class StringReader < Reader
    def initialize(string=nil)
      @string = string
    end

    def length
      @string ? @string.length : 0
    end

    alias size length

    def concat(string)
      if (@string) then
	@string.concat(string)
      else
	@string = string.dup
      end

      self
    end

    def <<(obj)
      concat(obj.to_s)
      self
    end

    def read(length=nil)
      if (length && length < 0) then
	raise ArgumentError, "negative length #{length} given"
      end

      if (@string) then
	if (! length || length >= @string.length) then
	  read_string = @string
	  @string = nil
	  return read_string
	else
	  read_string = @string[0, length]
	  @string[0, length] = ''
	  return read_string
	end
      else
	return nil
      end
    end

    def gets(eol=$/)
      if (@string) then
	if (eol_pos = @string.index(eol)) then
	  return read(eol_pos + eol.length)
	else
	  return read
	end
      else
	return nil
      end
    end

    def unget(string)
      if (@string) then
	@string = string + @string
      else
	@string = string.dup
      end

      self
    end

    def to_io
      nil
    end
  end

  module BufferedReadable
    # NOTE: abstract methods
    #  * public read
    #  * public read_chunk
    # NOTE: warning
    #  * Because read_chunk method may ignore buffer data, therefore
    #    flush_read_buffer method should be called before read_chunk
    #    method call.

    def initialize(*args, &block)
      super(*args, &block)
      @buffer = ''
    end

    def cut_buf(length)
      read_string = @buffer[0, length]
      @buffer[0, length] = ''
      read_string
    end
    private :cut_buf

    def flush_read_buf
      read_string = @buffer
      @buffer = ''
      read_string
    end
    private :flush_read_buf

    def flush_read_buffer
      unless (@buffer.empty?) then
	return flush_read_buf
      else
	return nil
      end
    end

    def gets(eol=$/)
      until (eol_pos = @buffer.index(eol))
	unless (read_string = read_chunk) then
	  unless (@buffer.empty?) then
	    return flush_read_buf
	  else
	    return nil
	  end
	end
	@buffer.concat(read_string)
      end

      cut_buf(eol_pos + eol.length)
    end

    def each_chunk
      while (read_string = read_chunk)
	yield(read_string)
      end

      self
    end
  end

  module ArbitraryReadable
    # NOTE: abstract methods
    #  * public _read

    include BufferedReadable

    def read(length=nil)
      if (length && length < 0) then
	raise ArgumentError, "negative length #{length} given"
      end

      if (length) then
	if (@buffer.empty?) then
	  return _read(length)
	end
	if (length <= @buffer.length) then
	  return cut_buf(length)
	end
	if (read_string = _read(length - @buffer.length)) then
	  @buffer.concat(read_string)
	end
      else
	if (@buffer.empty?) then
	  return _read
	end
	if (read_string = _read) then
	  @buffer.concat(read_string)
	end
      end

      flush_read_buf
    end

    def read_chunk
      _read(IO_UNIT_SIZE)
    end
  end

  module ChunkedReadable
    # NOTE: abstract methods
    #  * public read_chunk

    include BufferedReadable

    def read(length=nil)
      if (length && length < 0) then
	raise ArgumentError, "negative length #{length} given"
      end

      if (@buffer.empty?) then
	read_string = read_chunk or return nil
	@buffer.concat(read_string)
      end

      if (length) then
	while (@buffer.length < length)
	  unless (read_string = read_chunk) then
	    return flush_read_buf
	  end
	  @buffer.concat(read_string)
	end
	return cut_buf(length)
      else
	while (read_string = read_chunk)
	  @buffer.concat(read_string)
	end
	return flush_read_buf
      end
    end
  end

  class LengthLimitedReader < Reader
    include ArbitraryReadable

    def initialize(reader, read_length, padding=0)
      super()
      @reader = reader
      @read_length = read_length
      @padding = padding
    end

    def _read(length=nil)
      if (@read_length > 0) then
	length = @read_length if (! length || length > @read_length)
	if (read_string = ConfirmedRead::confirmed_read(@reader, length)) then
	  @read_length -= read_string.length
	  if (@read_length == 0 && @padding > 0) then
	    ConfirmedRead::confirmed_read(@reader, @padding)
	  end
	else
	  @read_length = 0
	end
	return read_string
      else
	return nil
      end
    end

    def to_io
      @reader && @reader.to_io
    end
  end

  class ChunkedTransferCodingReader < Reader
    include ChunkedReadable
    include ChunkedTransferCoding

    def initialize(reader)
      super()
      @reader = reader
      @chunk_extension = nil
      @trailer = nil
      @last_chunk = false
    end

    attr_reader :chunk_extension
    attr_reader :trailer

    def read_chunk
      unless (@last_chunk) then
	chunk_data, @chunk_extension = parse_chunk(@reader)
	unless (chunk_data) then
	  @trailer = Header::parse(@reader)
	  @last_chunk = true
	end
	return chunk_data
      else
	return nil
      end
    end

    def read_chunk_with_ext
      unless (@last_chunk) then
	chunk_data = read_chunk
	return chunk_data, @chunk_extension
      else
	return nil
      end
    end

    def each_chunk_with_ext
      until (@last_chunk)
	yield(*read_chunk_with_ext)
      end

      self
    end

    def to_io
      @reader && @reader.to_io
    end
  end

  module NullTrailerReadable
    def trailer
      nil
    end
  end

  class EmptyContentReader < EmptyReader
    include NullTrailerReadable
  end

  class ContentReader < Reader
    include Timeout
    include NullTrailerReadable

    def initialize(reader, sid, timeout, request_watcher)
      super()
      @reader = reader
      @sid = sid
      @timeout = timeout
      @request_watcher = request_watcher
    end

    def read(*args)
      read_string = wait(@timeout) { @reader.read(*args) }
      @request_watcher.watch_request_body(@sid, read_string) if read_string
      read_string
    end

    def gets(*args)
      read_string = wait(@timeout) { @reader.read(*args) }
      @request_watcher.watch_request_body(@sid, read_string) if read_string
      read_string
    end

    def to_io
      @reader && @reader.to_io
    end
  end

  class LengthLimitedContentReader < LengthLimitedReader
    include Timeout
    include NullTrailerReadable

    def initialize(reader, length, sid, timeout, request_watcher, padding=0)
      super(reader, length, padding)
      @sid = sid
      @timeout = timeout
      @request_watcher = request_watcher
    end

    def _read(*args)
      read_string = wait(@timeout) { super }
      @request_watcher.watch_request_body(@sid, read_string) if read_string
      read_string
    end
  end

  class ChunkedContentReader < ChunkedTransferCodingReader
    include Timeout

    def initialize(reader, sid, timeout, request_watcher)
      super(reader)
      @sid = sid
      @timeout = timeout
      @request_watcher = request_watcher
    end

    def read_chunk
      chunk_data = wait(@timeout) { super }
      @request_watcher.watch_request_body(@sid, chunk_data, @chunk_extension)
      @request_watcher.watch_request_trailer(@sid, @trailer) unless chunk_data
      chunk_data
    end
  end
end
