# $Id: request.rb,v 1.56.2.3 2003/04/11 14:09:08 toki Exp $
# -
# HTTP request

=begin
Copyright (c) 2001-2002
	TOKI Yoshinori <toki@freedom.ne.jp>. All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  1.  Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.

  2.  Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation
      and/or other materials provided with the distribution.

  3.  The name of the author may not be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
OF SUCH DAMAGE.
=end

require 'time'
require 'thread'
require 'wwwsrv/const'
require 'wwwsrv/cache'
require 'wwwsrv/path'
require 'wwwsrv/token'
require 'wwwsrv/htmlconv'
require 'wwwsrv/form'
require 'wwwsrv/reader'
require 'wwwsrv/version'
require 'wwwsrv/header'
require 'wwwsrv/srvopt'

module WWWsrv
  # latest CVS commit date of core libraries for wwwsrv
  update(%q$Date: 2003/04/11 14:09:08 $)

  class Host
    HTTP_PORT = 80

    def initialize(name=nil, port=HTTP_PORT)
      @name = name
      @port = port
    end

    attr_reader :name
    attr_reader :port

    def to_s
      "#{@name}:#{@port}"
    end

    def ==(host)
      (host.kind_of? self.class) && @name == host.name && @port == host.port
    end

    CACHE = ImmutableObjectCache::new{ |*args| self::new(*args) }

    def self::fetch_cache(*args)
      CACHE.fetch(*args)
    end
  end

  class Client
    def initialize(name=nil, address=nil)
      @name = name
      @address = address
    end

    attr_reader :name
    attr_reader :address

    def to_s
      if (@name == @address) then
	@address
      else
	"#{@name}(#{@address})"
      end
    end

    def to_i
      adrs_vec = @address.split(/\./, 4)
      n = 0
      for i in adrs_vec
	n <<= 8
	n |= i.to_i
      end

      n
    end

    def ==(client)
      (client.kind_of? self.class) && @name == client.name && @address == client.address
    end

    CACHE = ImmutableObjectCache::new{ |*args| self::new(*args) }

    def self::fetch_cache(*args)
      CACHE.fetch(*args)
    end
  end

  class RequestLineParseError < ParseError; end
  class NotSupportedVersionError < ParseError; end

  class Request
    include Path
    include HTMLconv
    include Attributes

    DEFAULT_QUERY_SEPARATOR = /&/

    def initialize(method=nil, path=nil, version=nil, header=Header::new, body=nil)
      @method = method
      self.path = path
      @version = version
      @header = header
      @body = body
      @host = nil
      @client = nil
      @option = {}
      @form_data = nil
      @params = nil
      @query_separator = DEFAULT_QUERY_SEPARATOR
    end

    # *NOTICE*
    # It is impossible to efficiently clone a pseudo I/O of
    # WWWsrv::Request#body attribute. Therefore WWWsrv::Request#clone
    # method dose not clone a WWWsrv::Request#body attribute and it
    # set `nil' to WWWsrv::Request#body attribute of a new clone
    # WWWsrv::Request object.
    def clone
      new_request = super
      new_request.instance_eval do
	@header = @header.dup
	@body = nil
	@option = @option.dup
	@form_data = @form_data.dup if @form_data
	@params = @params.dup  if @params
      end

      new_request
    end

    def dup
      clone
    end

    def clear
      @method = nil
      self.path = nil
      @version = nil
      if (@header) then
	@header.clear
      else
	@header = Header::new
      end
      @body = nil
      @host = nil
      @client = nil
      @option.clear
      @form_data = nil

      self
    end

    alias object_method method

    def method(*args, &block)
      (args.empty? && ! block) ? @method : object_method(*args, &block)
    end

    attr_writer   :method
    attr_reader   :path
    attr_accessor :version
    attr_accessor :header
    attr_accessor :body
    attr_accessor :host
    attr_accessor :client

    # for hook of WWWsrv::Attributes module.
    attr_reader :option; private :option

    def path=(new_path)
      if (@path = new_path) then
	@normalized_path, @query = normalize_path(@path)
      else
	@normalized_path = nil
	@query = nil
      end

      @path
    end

    attr_reader :normalized_path
    attr_reader :query

    def subpath(prefix)
      Path::subpath(prefix, @normalized_path)
    end

    alias path_info subpath

    def scan_path_info(prefix='')
      Path::scan_path_info(subpath(prefix)) do |script_name, path_info|
	yield(script_name, path_info)
      end

      self
    end

    def line
      if (@version >= HTTP_1_0) then
	return "#{@method} #{@path} #{@version}"
      else
	return "#{@method} #{@path}"
      end
    end

    def trailer
      @body.trailer
    end

    def self::chunked?(request_header)
      if (request_header.transfer_encoding) then
	if (request_header.transfer_encoding =~ /chunked/i) then
	  return true
	else
	  raise "unknown Transfer-Encoding type of request body: #{request_header.transfer_encoding.inspect}"
	end
      else
	return false
      end
    end

    def chunked?
      @header && (Request::chunked? @header)
    end

    def closed?
      if (@header && @header.connection && @header.connection =~ /close/i) then
	return true
      elsif (@version) then
	if (@version > HTTP_1_0) then
	  return false
	elsif (@version == HTTP_1_0) then
	  if (@header && @header.connection && @header.connection =~ /Keep-Alive/i) then
	    return false
	  else
	    return true
	  end
	else
	  return true
	end
      else
	return true
      end
    end

    def cond_if(mtime, etag=nil)
      unless (@header.if_range) then
	unless (@header.if_match || @header.if_unmodified_since) then
	  if (etag && @header.if_none_match) then
	    unless (@header.if_none_match != '*' && ! (@header.if_none_match.split(/\s*,\s*/).include? etag)) then
	      return 304
	    end
	  elsif (mtime && @header.if_modified_since) then
	    begin
	      timestamp = Time::parse(@header.if_modified_since)
	    rescue ArgumentError
	      # nothing to do.
	    else
	      unless (timestamp != mtime) then
		return 304
	      end
	    end
	  end
	end

	unless (@header.if_none_match || @header.if_modified_since) then
	  if (etag && @header.if_match) then
	    unless (@header.if_match == '*' || (@header.if_match.split(/\s*,\s*/).include? etag)) then
	      return 412
	    end
	  elsif (mtime && @header.if_unmodified_since) then
	    begin
	      timestamp = Time::parse(@header.if_unmodified_since)
	    rescue ArgumentError
	      # nothing to do.
	    else
	      unless (timestamp == mtime) then
		return 412
	      end
	    end
	  end
	end
      end

      200
    end

    def cond_if_range(mtime, etag=nil)
      unless (@header.if_match || @header.if_none_match) then
	unless (@header.if_modified_since || @header.if_unmodified_since) then
	  if (@header.if_range) then
	    begin
	      timestamp = Time::parse(@header.if_range)
	    rescue ArgumentError
	      # Entity Tag
	      if (etag && @header.if_range == etag) then
		return 206
	      end
	    else
	      # HTTP date
	      if (mtime && timestamp == mtime) then
		return 206
	      end
	    end
	  end
	end
      end

      200
    end

    RANGE = Header::normalize('Range')

    def erase_cond_header_fields
      if (@header) then
	@header.delete_if{ |field_name, field_value|
	  case (field_name)
	  when /^If-/i, RANGE
	    true
	  else
	    false
	  end
	}
      end

      nil
    end

    LAST_MODIFIED = Header::normalize('Last-Modified')

    def erase_message_body
      if (@body) then
	@body = EmptyContentReader::instance
      end
      if (@header) then
	@header.delete_if{ |field_name, field_value|
	  case (field_name)
	  when /^Content-/i, LAST_MODIFIED
	    true
	  else
	    false
	  end
	}
      end

      nil
    end

    attr_accessor :query_separator

    def scan_form_data
      case (@method)
      when HTTP_POST
	case (@header.content_type)
	when %r"^application/x-www-form-urlencoded"i
	  FormData.scan_query(@body) do |tag, value|
	    yield(tag, value)
	  end
	when %r"^multipart/form-data"i
	  FormData.scan_form_data(@body, @header.content_type) do |tag, value|
	    yield(tag, value)
	  end
	end
      else
	FormData.parse_query(@query || '', @query_separator) do |tag, value|
	  yield(tag, value)
	end
      end

      nil
    end
    private :scan_form_data

    def get_form_data
      @form_data = FormData.new
      name = nil; head = nil; body = nil
      filename = nil
      multipart_mixed = false
      scan_form_data do |tag, value|
	case (tag)
	when :form_data_name
	  name = value
	when :form_data_file
	  filename = value
	when :form_data_head
	  head = value
	when :form_data_body
	  body = '' unless body
	  body.concat(value)
	when :form_data_end
	  unless (multipart_mixed) then
	    @form_data.add_form_data(name, head)
	    @form_data.add_form_data_body(name, head, body, filename)
	    name = nil; head = nil; body = nil
	    filename = nil
	  else
	    @form_data.add_form_data_body(name, head, body, filename)
	    head = nil; body = nil
	    filename = nil
	  end
	when :begin_mixed
	  multipart_mixed = true
	  @form_data.add_form_data(name, head)
	when :end_mixed
	  multipart_mixed = false
	end
      end

      @form_data
    end
    private :get_form_data

    def form_data
      if (iterator?) then
	scan_form_data do |tag, value|
	  yield(tag, value)
	end
	return
      else
	if (@form_data) then
	  return @form_data
	else
	  return get_form_data
	end
      end
    end

    alias params form_data

    def cgi_var(cgi_var=Hash::new, script_name=nil, pass_auth=false)
      cgi_var['GATEWAY_INTERFACE'] = 'CGI/1.1'
      cgi_var['REQUEST_METHOD'] = self.method
      cgi_var['QUERY_STRING'] = self.query || ''
      cgi_var['REMOTE_ADDR'] = self.client.address
      if (self.client.name != self.client.address) then
	cgi_var['REMOTE_HOST'] = self.client.name
      end
      if (script_name) then
	cgi_var['SCRIPT_NAME'] = script_name
	if (script_name != self.normalized_path) then
	  path_info = subpath(script_name)
	  if (path_info && ! path_info.empty?) then
	    cgi_var['PATH_INFO'] = path_info
	    #cgi_var['PATH_TRANSLATED'] = ''
	  end
	end
      else
	cgi_var['SCRIPT_NAME'] = self.normalized_path
      end
      cgi_var['SERVER_NAME'] = self.host.name
      cgi_var['SERVER_PORT'] = self.host.port.to_s
      cgi_var['SERVER_PROTOCOL'] = self.version.to_s
      cgi_var['SERVER_SOFTWARE'] = WWWsrv::product_token
      if (self[:auth_userid]) then
	cgi_var['REMOTE_USER'] = self[:auth_userid]
      end
      for field_name, field_value in self.header
	case (field_name)
	when 'Content-Type'
	  cgi_var['CONTENT_TYPE'] = field_value
	when 'Content-Length'
	  cgi_var['CONTENT_LENGTH'] = field_value
	when /Authorization/
	  if (field_name == 'Authorization') then
	    cgi_var['AUTH_TYPE'] = self[:auth_type] || field_value.split(/\s+/, 2).first
	  end
	  if (pass_auth) then
	    name = field_name.upcase
	    name.gsub!(/-/, '_')
	    cgi_var['HTTP_' + name] = field_value
	  end
	else
	  name = field_name.upcase
	  name.gsub!(/-/, '_')
	  cgi_var['HTTP_' + name] = field_value
	end
      end

      cgi_var
    end

    def self::parse_line(input)
      while (request_line = input.gets(LF))
	request_line.chomp!(LF)
	request_line.chomp!(CR)
	request_line.empty? and next

	method, path, version = request_line.split(/\s+/, 3)
	(! method || method.empty?) and
	  raise RequestLineParseError, 'undefined HTTP request method'
	(! path || path.empty?) and
	  raise RequestLineParseError, 'undefined HTTP request path'
	if (version) then
	  http_version = HTTPversion::parse(version)
	else
	  http_version = HTTP_0_9
	end
	if (http_version >= HTTP_2_0) then
	  raise NotSupportedVersionError, "not supported HTTP version: #{http_version}"
	end

	return method, path, http_version
      end

      nil
    end

    def parse(input)
      if (request_line = Request::parse_line(input)) then
	@method, self.path, @version = request_line
	@header = Header::new unless @header
	if (@version >= HTTP_1_0) then
	  @header.parse(input)
	end

	return self
      end

      nil
    end

    CACHE = RecycleObjectCache::new(self)

    def self::fetch_cache(&block)
      CACHE.fetch(&block)
    end

    def self::store_cache(old_obj)
      CACHE.store(old_obj)
    end

    def self::parse(input, request=self::new)
      request.parse(input)
    end
  end

  module PassAuth
    def initialize(*args, &block)
      super(*args, &block)
      @pass_auth = false
    end

    attr_reader :pass_auth

    def pass_auth=(new_pass_auth)
      @pass_auth = new_pass_auth ? true : false
    end

    def fetch_cgi_var(request, cgi_var=Hash::new, script_name=nil)
      request.cgi_var(cgi_var, script_name, @pass_auth)
    end
    private :fetch_cgi_var
  end
end
