# $Id: response.rb,v 1.20.2.1 2002/12/27 15:18:09 toki Exp $
# -
# HTTP response

=begin
Copyright (c) 2001-2002
	TOKI Yoshinori <toki@freedom.ne.jp>. All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  1.  Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.

  2.  Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation
      and/or other materials provided with the distribution.

  3.  The name of the author may not be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
OF SUCH DAMAGE.
=end

require 'wwwsrv/const'
require 'wwwsrv/cache'
require 'wwwsrv/version'
require 'wwwsrv/header'
require 'wwwsrv/srvopt'

module WWWsrv
  # latest CVS commit date of core libraries for wwwsrv
  update(%q$Date: 2002/12/27 15:18:09 $)

  HTTP_MSG = {
    100 => 'Continue',
    101 => 'Switching Protocols',
    200 => 'OK',
    201 => 'Created',
    202 => 'Accepted',
    203 => 'Non-Authoritative Information',
    204 => 'No Content',
    205 => 'Reset Content',
    206 => 'Partial Content',
    300 => 'Multiple Choices',
    301 => 'Moved Permanently',
    302 => 'Found',
    303 => 'See Other',
    304 => 'Not Modified',
    305 => 'Use Proxy',
    307 => 'Temporary Redirect',
    400 => 'Bad Request',
    401 => 'Unauthorized',
    402 => 'Payment Required',
    403 => 'Forbidden',
    404 => 'Not Found',
    405 => 'Method Not Allowed',
    406 => 'Not Acceptable',
    407 => 'Proxy Authentication Required',
    408 => 'Request Timeout',
    409 => 'Conflict',
    410 => 'Gone',
    411 => 'Length Required',
    412 => 'Precondition Failed',
    413 => 'Request Entity Too Large',
    414 => 'Request-URI Too Long',
    415 => 'Unsupported Media Type',
    416 => 'Requested Range Not Satisfiable',
    417 => 'Expectation Failed',
    500 => 'Internal Server Error',
    501 => 'Not Implemented',
    502 => 'Bad Gateway',
    503 => 'Service Unavailable',
    504 => 'Gateway Timeout',
    505 => 'HTTP Version Not Supported'
  }

  class StatusLineParseError < ParseError; end

  class Response
    include Attributes

    def initialize(version=HTTP_1_1, status=200, header=Header::fetch_cache, trailer=Header::fetch_cache)
      @version = version
      self.status = status
      @header = header
      @trailer = trailer
      @option = {}
    end

    def clone
      new_response = super
      new_response.instance_eval do
	@header = @header.dup
	@trailer = @trailer.dup
	@option = @option.dup
      end

      new_response
    end

    def dup
      clone
    end

    def clear
      @version = HTTP_1_1
      self.status = 200
      @header.clear
      @trailer.clear
      @option.clear

      self
    end

    attr_accessor :version
    attr_reader   :status
    attr_accessor :reason
    attr_accessor :header
    attr_accessor :trailer

    # for hook of WWWsrv::Attributes module.
    attr_reader :option; private :option

    def status=(new_status)
      @status = new_status
      @reason = HTTP_MSG[@status]

      @status
    end

    def line
      "#{@version} #{@status} #{@reason}"
    end

    def self::chunked?(response_header)
      if (response_header.transfer_encoding) then
	if (response_header.transfer_encoding =~ /chunked/i) then
	  return true
	else
	  raise "unknown Transfer-Encoding type of response body: #{response_header.transfer_encoding.inspect}"
	end
      else
	return false
      end
    end

    def chunked?
      @header && (Response::chunked? @header)
    end

    def closed?
      if (@header && @header.connection && @header.connection =~ /close/i) then
	return true
      elsif (@version && @version < HTTP_1_0) then
	return true
      else
	return false
      end
    end

    def close
      if (@header) then
	if (! @header.connection || @header.connection !~ /close/i) then
	  @header.push('Connection', 'close')
	end
      end

      self
    end

    COND_HEADER_FIELDS = {
      Header::normalize('ETag') => true,
      Header::normalize('Last-Modified') => true
    }

    def erase_cond_header_fields
      if (@header) then
	@header.delete_if{ |field_name, field_value|
	  COND_HEADER_FIELDS.include? field_name
	}
      end

      nil
    end

    STICKY_ENTITY_HEADER_FIELDS = {
      Header::normalize('ETag') => true,
      Header::normalize('Last-Modified') => true,
      Header::normalize('Content-MD5') => true,
      Header::normalize('Content-Length') => true
    }

    def erase_sticky_entity_header_fields
      if (@header) then
	@header.delete_if{ |field_name, field_value|
	  STICKY_ENTITY_HEADER_FIELDS.include? field_name
	}
      end

      nil
    end

    CACHE = RecycleObjectCache::new(self)

    def self::fetch_cache(&block)
      CACHE.fetch(&block)
    end

    def self::store_cache(old_obj)
      CACHE.store(old_obj)
    end

    def self::parse_status(field_value)
      status, reason = field_value.split(/\s+/, 2)
      if (! status || status.empty? || status !~ /^\d\d\d$/) then
	raise StatusLineParseError, "failed to parse response status code: #{status.inspect}"
      end
      if (! reason || reason.empty?) then
	raise StatusLineParseError, "failed to parse response reason: #{reason.inspect}"
      end

      status_code = status.to_i
      if (status_code < 100 || 600 <= status_code) then
	raise StatusLineParseError, "out of range of status code: #{status_code}"
      end

      return status_code, reason
    end

    def self::parse_status_line(input)
      while (status_line = input.gets(LF))
	status_line.chomp!(LF)
	status_line.chomp!(CR)
	status_line.empty? and next

	version, others = status_line.split(/\s+/, 2)
	version or raise StatusLineParseError, "not found a HTTP version: #{status_line}"
	http_version = HTTPversion::parse(version)
	status_code, reason = parse_status(others)

	return http_version, status_code, reason
      end

      nil
    end
  end
end
