# $Id: server.rb,v 1.132.2.3 2003/04/11 14:09:09 toki Exp $
# -
# HTTP server

=begin
Copyright (c) 2001-2002
	TOKI Yoshinori <toki@freedom.ne.jp>. All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  1.  Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.

  2.  Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation
      and/or other materials provided with the distribution.

  3.  The name of the author may not be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
OF SUCH DAMAGE.
=end

require 'socket'
require 'wwwsrv/const'
require 'wwwsrv/thread'
require 'wwwsrv/process'
require 'wwwsrv/ioutil'
require 'wwwsrv/delegate'
require 'wwwsrv/htmlconv'
require 'wwwsrv/logger'
require 'wwwsrv/chunk'
require 'wwwsrv/reader'
require 'wwwsrv/request'
require 'wwwsrv/response'
require 'wwwsrv/document'
require 'wwwsrv/alert'

module WWWsrv
  # latest CVS commit date of core libraries for wwwsrv
  update(%q$Date: 2003/04/11 14:09:09 $)

  class WriteBuffer
    include Timeout
    include ConfirmedWrite

    def initialize(output, timeout, threshold)
      @output = output
      @timeout = timeout
      @threshold = threshold
      @buffer = ''
    end

    def flush
      wait(@timeout) do
	confirmed_write(@output, @buffer) unless @buffer.empty?
	@output.flush
      end
      @buffer = ''

      self
    end

    def write(fragment)
      if (fragment.length >= @threshold) then
	unless (@buffer.empty?) then
	  wait(@timeout) { confirmed_write(@output, @buffer) }
	  @buffer = ''
	end
	wait(@timeout) { confirmed_write(@output, fragment) }
      else
	@buffer.concat(fragment)
	if (@buffer.length >= @threshold) then
	  wait(@timeout) { confirmed_write(@output, @buffer) }
	  @buffer = ''
	end
      end

      fragment.length
    end
  end

  class ResponseWriter
    include HTMLconv
    include ChunkedTransferCoding

    # Note: state map to write a response.
    #                   chunk?
    # send_response --+--------> send_chunk -----> terminate_chunk
    #                 |
    #                 | spool?
    #                 +--------> send_spool -----> terminate_spool
    #                 |            |overflow?
    #                 | through?   V
    #                 +--------> send_through ---> terminate_through
    #                 |
    #                 | empty?
    #                 +--------> send_null ------> terminate_null

    def initialize(output, request, resposne_watcher, threshold=BUFFER_SIZE)
      @output = output
      @request = request
      @response = nil
      @used = false
      @threshold = threshold
      @send_length = nil
      @sender = method(:send_response)
      @terminator = nil
      @resposne_watcher = resposne_watcher
    end

    attr_reader :threshold
    attr_reader :response

    def used?
      @used
    end

    def set_timestamp(response)
      if (response.status < 500) then
	unless (response.header.date) then
	  response.header.date = http_date(Time::now)
	  response.header.set_first('Date')
	end
      end
    end
    private :set_timestamp

    def set_connection(request, response)
      unless (response.closed?) then
	if (request.closed?) then
	  response.close
	else
	  if (request.version == HTTP_1_0 || response.version == HTTP_1_0) then
	    if (response.header.connection !~ /Keep-Alive/i) then
	      response.header.push('Connection', 'Keep-Alive')
	    end
	  end
	end
      end
    end
    private :set_connection

    def set_send_length(response)
      if (response.header.content_length) then
	@send_length = response.header.content_length.to_i
      else
	@send_length = nil
      end
    end
    private :set_send_length

    def write_status_line(response)
      @output.write(response.line + CRLF)
    end
    private :write_status_line

    def write_header(header)
      head = header.to_s
      head.concat(CRLF)
      @output.write(head)
    end
    private :write_header

    def write_chunk_head(chunk_data, chunk_extension=nil)
      @output.write(chunk_head_line(chunk_data, chunk_extension))
    end
    private :write_chunk_head

    def write_message_head(response)
      write_status_line(response)
      write_header(response.header)
    end
    private :write_message_head

    EMPTY_STATUS = {
      204 => true,
      205 => true,
      304 => true
    }

    def send_response(sid, arg, opts=nil)
      @used = true
      @response = arg
      set_timestamp(@response)
      set_connection(@request, @response)
      set_send_length(@response)
      if (@request && @request.version >= HTTP_1_1 && @response.version >= HTTP_1_1) then
	unless (EMPTY_STATUS.include? @response.status) then
	  unless (@response.chunked?) then
	    @response.header.content_length = nil
	    @response.header.transfer_encoding = 'chunked'
	  end
	  write_message_head(@response)
	  @resposne_watcher.watch_response_head(sid, response)
	  if (@request.method != HTTP_HEAD) then
	    init_chunk
	  else
	    init_null
	  end
	else
	  if (@response.chunked?) then
	    # The response MUST NOT include a message-body, thus the
	    # resopnse is always terminated by the first empty line
	    # after the header fields.
	    @response.transfer_encoding = nil
	  end
	  if (@request.method != HTTP_HEAD) then
	    if (@response.header.content_length || @response.closed?) then
	      write_message_head(@response)
	      @resposne_watcher.watch_response_head(sid, response)
	      init_througth
	    else
	      # Head of response is delayed.
	      init_spool
	    end
	  else
	    write_message_head(@response)
	    @resposne_watcher.watch_response_head(sid, response)
	    init_null
	  end
	end
      elsif ((! @request || @request.version >= HTTP_1_0) && @response.version >= HTTP_1_0) then
	if (! @request || @request.method != HTTP_HEAD) then
	  if (@response.header.content_length || @response.closed?) then
	    write_message_head(@response)
	    @resposne_watcher.watch_response_head(sid, response)
	    init_througth
	  else
	    # Head of response is delayed.
	    init_spool
	  end
	else
	  write_message_head(@response)
	  @resposne_watcher.watch_response_head(sid, response)
	  init_null
	end
      else # HTTP/0.9
	@resposne_watcher.watch_response_head(sid, @response)
	init_througth
      end
    end
    private :send_response

    def init_chunk
      @last_chunk = false
      @sender = method(:send_chunk)
      @terminator = method(:terminate_chunk)
    end
    private :init_chunk

    def init_spool
      @message_body = ''
      @sender = method(:send_spool)
      @terminator = method(:terminate_spool)
    end
    private :init_spool

    def init_througth
      @sender = method(:send_through)
      @terminator = method(:terminate_through)
    end
    private :init_througth

    def init_null
      @send_length = nil
      @sender = method(:send_null)
      @terminator = method(:terminate_null)
    end
    private :init_null

    def send_chunk(sid, chunk_data, chunk_extension=nil)
      if (chunk_data && ! chunk_data.empty?) then
	@send_length -= chunk_data.length if @send_length
	write_chunk_head(chunk_data, chunk_extension)
	@output.write(chunk_data)
	@output.write(CRLF)
	@resposne_watcher.watch_response_body(sid, chunk_data, chunk_extension)
      elsif (! chunk_data) then
	@last_chunk = true
	write_chunk_head(nil, chunk_extension)
	@resposne_watcher.watch_response_body(sid, nil, chunk_extension)
      end
    end
    private :send_chunk

    def terminate_chunk(sid)
      unless (@last_chunk) then
	write_chunk_head(nil)
	@resposne_watcher.watch_response_body(sid, nil)
      end
      write_header(@response.trailer)
      @resposne_watcher.watch_response_trailer(sid, @response.trailer)
    end
    private :terminate_chunk

    def send_spool(sid, body_fragment, opts=nil)
      if (body_fragment) then
	@send_length -= body_fragment.length if @send_length
	if (@message_body.length + body_fragment.length < @threshold) then
	  @message_body.concat(body_fragment)
	else
	  @response.close
	  write_message_head(@response)
	  @resposne_watcher.watch_response_head(sid, @response)
	  unless (@message_body.empty?) then
	    @output.write(@message_body)
	    @resposne_watcher.watch_response_body(sid, @message_body)
	  end
	  @message_body = nil
	  @output.write(body_fragment)
	  @resposne_watcher.watch_response_body(sid, body_fragment)
	  init_througth
	end
      end
    end
    private :send_spool

    def terminate_spool(sid)
      @response.header.content_length = @message_body.length.to_s
      write_message_head(@response)
      @resposne_watcher.watch_response_head(sid, @response)
      unless (@message_body.empty?) then
	@output.write(@message_body)
	@resposne_watcher.watch_response_body(sid, @message_body)
      end
      @message_body = nil
    end
    private :terminate_spool

    def send_through(sid, body_fragment, opts=nil)
      if (body_fragment) then
	@send_length -= body_fragment.length if @send_length
	@output.write(body_fragment)
	@resposne_watcher.watch_response_body(sid, body_fragment)
      end
    end
    private :send_through

    def terminate_through(sid)
      # nothing to do.
    end
    private :terminate_through

    def send_null(sid, body_fragment, opts=nil)
      # nothing to do.
    end
    private :send_null

    def terminate_null(sid)
      # nothing to do.
    end
    private :terminate_null

    def write(sid, arg, opts=nil)
      @sender.call(sid, arg, opts)
      self
    end

    def terminate(sid)
      if (@send_length) then
	if (@send_length != 0) then
	  raise 'mismatch content length.'
	end
      end
      @terminator.call(sid)
      self
    end
  end

  class RequestError < StandardError; end

  class Server
    include Timeout
    include SocketManipulation

    SID_LOCK = Mutex::new

    @@sid = 'A'

    def initialize(srv_sock, alert=AlertDocument::new, document=PartialContent::new(DocumentFolder::new))
      @sid = SID_LOCK.synchronize{ sid = @@sid; @@sid = @@sid.succ; sid }

      @scheme = HTTP_SCHEME
      @srv_sock = srv_sock
      @closed = false
      @accept_thread = nil
      @accept_lock = Mutex::new
      @accept_cond = ConditionVariable::new
      @interruptable = true
      @interrupt_cond = ConditionVariable::new
      @connections = MAX_CONNECTIONS
      @curr_conns = 0
      @max_requests = MAX_REQUESTS
      @dispatch_threads = ThreadGroup::new

      @timeout = TIMEOUT
      @keep_alive = KEEP_ALIVE
      @buffer_size = BUFFER_SIZE

      if (@srv_sock.kind_of? IPsocket) then
	srv_family, srv_port, srv_name, srv_addr = @srv_sock.addr
	hostname, alias_list, addr_type, ip_addr = Socket.gethostbyname(Socket.gethostname)
	@host = Host::fetch_cache(hostname, srv_port)
      else
	@host = Host::fetch_cache('NOT-IP-SOCKET', 0)
      end
      @alert = alert
      @document = document

      @log = WWWsrv::MultiLogger::new
      @alias = nil
      @access = nil

      @connect_watcher = MultiConnectionWatcher::new
      @request_watcher = MultiRequestWatcher::new
      @response_watcher = MultiResponseWatcher::new
      @connect_logger = ConnectionLogger::new(@log)
      @request_logger = RequestLogger::new(@log)
      @response_logger = ResponseLogger::new(@log)
      @connect_watcher.add_connect_watcher(@connect_logger)
      @request_watcher.add_request_watcher(@request_logger)
      @response_watcher.add_response_watcher(@response_logger)

      # warpping server socket
      @srv_sock = ProtectedServerSocket::new(@srv_sock)
    end

    attr_accessor :scheme
    attr_accessor :timeout
    attr_accessor :keep_alive
    attr_accessor :connections
    attr_reader   :curr_conns
    attr_accessor :max_requests
    attr_accessor :buffer_size

    def hostname
      @host.name
    end

    def hostname=(new_name)
      @host = Host::fetch_cache(new_name, @host.port)
      new_name
    end

    def port
      @host.port
    end

    attr_accessor :alert
    attr_reader   :document

    def document=(new_doc)
      @alias = nil
      @access = nil
      @document = new_doc
    end

    def add_logger(logger)
      @log.add_logger(logger)
    end

    def add_connect_watcher(connect_watcher)
      @connect_watcher.add_connect_watcher(connect_watcher)
    end

    def add_request_watcher(request_watcher)
      @request_watcher.add_request_watcher(request_watcher)
    end

    def add_response_watcher(response_watcher)
      @response_watcher.add_response_watcher(response_watcher)
    end

    attr_reader :connect_logger
    attr_reader :request_logger
    attr_reader :response_logger

    # see WWWsrv::AccessControl#add_access_control method.
    def add_access_control(*args, &block)
      unless (@access) then
	@access = WWWsrv::AccessControl::new(@document)
	@document = @access
      end
      @access.add_access_control(*args, &block)

      self
    end

    def alias(dst_path, src_path)
      unless (@alias) then
	@alias = WWWsrv::AliasDocument::new(@document)
	@document = @alias
      end
      @alias.alias(dst_path, src_path)

      self
    end

    # delegation of document
    def method_missing(id, *args, &block)
      unless (Document::instance_methods.include? id.to_s) then
	if (@document.__delegatable__? id) then
	  ret_val = @document.__send__(id, *args, &block)
	  return (ret_val.__id__ == @document.__id__) ? self : ret_val
	end
      end

      raise NameError, 'undefined method ' + id.to_s
    end

    def make_server_option(sid)
      ServerOption::fetch_cache{ |srv_opt|
	srv_opt.sid = sid
	srv_opt.root = @document
	srv_opt.log = @log
	srv_opt.scheme = @scheme
      }
    end
    private :make_server_option

    def make_response(status=200, version=HTTP_1_1)
      Response::fetch_cache{ |response|
	response.version = version
	response.status = status
	response.header.server = WWWsrv::product_token
      }
    end
    private :make_response

    def make_fatal_error_response(status=500)
      response = make_response(status)
      response.close
      response
    end
    private :make_fatal_error_response

    def sending(wbuf, request)
      begin
	output = ResponseWriter::new(wbuf, request, @response_watcher, @buffer_size)
	yield(output)
      ensure
	wbuf.flush
      end
    end
    private :sending

    def receive_request(socket, count)
      request = Request::fetch_cache
      begin
	if (wait((count == 1) ? @timeout : @keep_alive) { request.parse(socket) }) then
	  return request
	else
	  Request::store_cache(request)
	  return nil
	end
      rescue TimeoutError
	Request::store_cache(request)
	return nil
      rescue RequestLineParseError, HeaderParseError
	raise RequestError, "#{$!} (#{$!.type})"
      end
    end
    private :receive_request

    def set_host(request)
      if (host_value = request.header.host) then
	name, port = host_value.split(/:/, 2)
	request.host = Host::fetch_cache(name, port ? port.to_i : @host.port)
      elsif (request.version <= HTTP_1_0) then
	request.host = @host
      else
	http_error(400, 'Not found a Host header field.', 'Connection' => 'close')
      end

      if (request.path !~ %r"^/") then
	if (request.path =~ %r"^https?://(.*?)(:(\d+))?(/|$)") then
	  name = $1
	  port = $3 ? $3.to_i : 80
	  abs_path = '/'+ $'
	  request.host = Host::fetch_cache(name, port)
	  request.path = abs_path
	end
      end
    end
    private :set_host

    def make_content_reader(sid, socket, srv_opt, request)
      begin
	if (request.chunked?) then
	  return ChunkedContentReader::new(socket, sid, @timeout, @request_watcher)
	elsif (request.header.content_length) then
	  content_length = request.header.content_length.to_i
	  if (content_length > 0) then
	    return LengthLimitedContentReader::new(socket, content_length, sid, @timeout, @request_watcher)
	  elsif (content_length == 0) then
	    return EmptyContentReader::instance
	  else
	    raise "negative Content-Length of request: #{content_length}"
	  end
	elsif (request.method == HTTP_POST) then
	  if (request.closed?) then
	    return ContentReader::new(socket, sid, @timeout, @request_watcher)
	  else
	    http_error(411, 'required Content-Length.', 'Connection' => 'close')
	  end
	else
	  return EmptyContentReader::instance
	end
      rescue
	http_error(400, "#{$!} (#{$!.type})", 'Connection' => 'close')
      end
    end
    private :make_content_reader

    def slight_alert(output, srv_opt, request, err_args)
      status, message, header, version = err_args
      if (status.kind_of? Response) then
	if ($WWWSRV_NOT_COMPAT) then
	  raise 'disabled backward compatibility'
	end
	response = status	# for backward compatibility
      else
	response = make_response(status, version)
	for field_name, field_value in header
	  response.header[field_name] = field_value
	end
      end

      @log.message(Logger::srv_format(srv_opt.sid, "document error: #{response.line}"))
      unless (output.used?) then
	srv_opt[:error] = message
	@alert.respond(srv_opt, '', request, response) do |arg, opts|
	  output.write(srv_opt.sid, arg, opts)
	end
	output.terminate(srv_opt.sid)
      end
    end
    private :slight_alert

    def fatal_alert(sid, wbuf, status, except, backtrace=false, not_send=false)
      response = make_fatal_error_response(status)
      @log.alert(Logger::srv_format(sid, "server error: #{response.line}: #{$!} (#{$!.type})"))
      Logger::backtrace(@log, Logger::WARNING, sid, except) if backtrace
      unless (not_send) then
	sending(wbuf, nil) do |output|
	  srv_opt = make_server_option(sid)
	  srv_opt[:error] = except
	  @alert.respond(srv_opt, '', nil, response) do |arg, opts|
	    output.write(srv_opt.sid, arg, opts)
	  end
	  output.terminate(srv_opt.sid)
	end
      end
    end
    private :fatal_alert

    def dispatch(sid, socket, wbuf)
      count = 1
      output = nil
      begin
	# socket may be warpped.
	if ((raw_sock = socket.to_io) && (raw_sock.kind_of? IPsocket)) then
	  cli_family, cli_port, cli_name, cli_addr = socket.peeraddr
	else
	  cli_name = "[not an IP socket: #{socket.inspect}]"
	  cli_addr = '0.0.0.0'
	end
	client = Client::fetch_cache(cli_name, cli_addr)
	begin
	  @connect_watcher.watch_connect_open(sid, socket)
	  loop do
	    srv_opt = make_server_option("#{sid}-#{count}")
	    request = receive_request(socket, count) or break
	    request.client = client
	    sending(wbuf, request) do |output|
	      doc_status, err_args = Document::catch_status{
		begin
		  set_host(request)
		  request.body = make_content_reader(srv_opt.sid, socket, srv_opt, request)
		  @request_watcher.watch_request_head(srv_opt.sid, request)
		  response = make_response
		  response.close if (count >= @max_requests)
		  @document.respond(srv_opt, '', request, response) do |arg, opts|
		    output.write(srv_opt.sid, arg, opts)
		  end
		  output.terminate(srv_opt.sid)
		rescue TimeoutError
		  @log.system(Logger::srv_format(srv_opt.sid, "request expired: #{$!}"))
		  http_error(408, "Request expired.", 'Connection' => 'close')
		end
	      }
	      slight_alert(output, srv_opt, request, err_args) if (doc_status == :error)
	    end

	    while (request.body.read(IO_UNIT_SIZE))
	      # nothing to do.
	    end

	    closed = request.closed? || output.response.closed?
	    ServerOption::store_cache(srv_opt)
	    Request::store_cache(request)
	    Response::store_cache(output.response)

	    closed and break
	    @connect_watcher.watch_connect_keep_alive(sid, socket)
	    count += 1
	  end
	rescue ConnectionError
	  raise
	rescue RequestError
	  fatal_alert(sid, wbuf, 400, $!)
	rescue NotSupportedVersionError
	  fatal_alert(sid, wbuf, 505, $!)
	rescue StandardError, ScriptError
	  fatal_alert(sid, wbuf, 500, $!, true, output && output.used?)
	ensure
	  begin
	    socket.shutdown
	  ensure
	    socket.close
	  end
	  @connect_watcher.watch_connect_close(sid, socket)
	end
      rescue ConnectionError
	@connect_watcher.watch_connect_error(sid, socket, $!)
      end
    end
    private :dispatch

    def scan_docs(document, mark={})
      unless (mark.include? document.id) then
	mark[document.id] = true
	yield(document)
	for doc in document.internal_documents
	  scan_docs(doc, mark) do |internal_document|
	    yield(internal_document)
	  end
	end
      end
    end
    private :scan_docs

    def _enable_interrupt
      @interruptable = true
      @interrupt_cond.signal
    end
    private :_enable_interrupt

    def _disable_interrupt
      @interruptable = false
    end
    private :_disable_interrupt

    def enable_interrupt
      @accept_lock.synchronize do
	_enable_interrupt
      end
    end
    private :enable_interrupt

    def disable_interrupt
      @accept_lock.synchronize do
	_disable_interrupt
      end
    end
    private :disable_interrupt

    def wait_interrupt
      @accept_lock.synchronize do
	until (@interruptable)
	  @interrupt_cond.wait(@accept_lock)
	end
	yield
      end
    end
    private :wait_interrupt

    def accept
      @connect_watcher.watch_accept_start(@sid, @srv_sock)
      begin
	@accept_thread = Thread::new{
	  Thread::current.priority = SYSTEM_PRIORITY
	  i = 1
	  begin
	    disable_interrupt
	    doc_mark = {}
	    @log.debug("debug: root document -> #{@document}")
	    scan_docs(@document, doc_mark) do |doc|
	      doc.warm_up(@log)
	    end
	    @log.debug("debug: alert document -> #{@alert}")
	    scan_docs(@alert, doc_mark) do |doc|
	      doc.warm_up(@log)
	    end
	    loop do
	      @closed and break
	      sid = "#{@sid}.#{i}"
	      begin
		begin
		  @accept_lock.synchronize do
		    _enable_interrupt
		    while (@curr_conns >= @connections)
		      @accept_cond.wait(@accept_lock)
		    end
		  end
		  socket = @srv_sock.accept
		  socket = ProtectedSocket::new(socket) if socket
		ensure
		  @accept_lock.synchronize do
		    begin
		      if (socket) then
			unless (socket.closed?) then
			  @curr_conns += 1
			else
			  socket = nil
			end
		      end
		    ensure
		      _disable_interrupt
		    end
		  end
		end
	      rescue
		@connect_watcher.watch_accept_error(sid, @srv_sock, $!)
		if (socket) then
		  begin
		    begin
		      socket.shutdown
		    ensure
		      socket.close
		    end
		  rescue
		    # nothing do do.
		  ensure
		    socket = nil
		  end
		end
	      end
	      next unless socket
	      @connect_watcher.watch_connect_count(sid, @srv_sock, curr_conns)

	      socket.binmode
	      set_tcp_no_delay(socket)
	      dispatch_thread = Thread::new{
		Thread::current.priority = NORMAL_PRIORITY
		wbuf = WriteBuffer::new(socket, @timeout, @buffer_size)
		begin
		  dispatch(sid, socket, wbuf)
		ensure
		  @accept_lock.synchronize do
		    @curr_conns -= 1
		    @accept_cond.signal
		  end
		end
	      }
	      dispatch_thread[:wwwsrv_sid] = sid
	      @dispatch_threads.add(dispatch_thread)

	      i += 1
	    end
	  ensure
	    enable_interrupt
	  end
	}
	@accept_thread.join
      ensure
	wait_thread = Thread::new{
	  Thread::current.priority = SYSTEM_PRIORITY
	  loop do
	    th_count = 0
	    for thread in @dispatch_threads.list
	      begin
		th_count += 1
		thread.join
	      rescue TimeoutError
		# nothing to do.
	      rescue
		sid = thread[:wwwsrv_sid] || @sid
		@log.alert(Logger::srv_format(sid, "shutdown error: $!"))
		Logger::backtrace(@log, Logger::WARNING, sid, $!)
	      end
	    end
	    (th_count == 0) and break
	  end
	}

	intercept_thread = Thread::new{
	  Thread::current.priority = SYSTEM_PRIORITY
	  while (wait_thread.alive?)
	    Timeout::intercept_all(@dispatch_threads)
	    sleep(0.1)
	  end

	  doc_mark = {}
	  scan_docs(@document, doc_mark) do |doc|
	    doc.cool_down(@log)
	  end
	  scan_docs(@alert, doc_mark) do |doc|
	    doc.cool_down(@log)
	  end
	}

	begin
	  wait_thread.join
	  intercept_thread.join
	ensure
	  @log.close
	end
      end

      self
    end

    def close
      wait_interrupt do
	unless (@closed) then
	  if (@accept_thread && @accept_thread.alive?) then
	    @accept_thread.exit
	  end
	  @closed = true
	end
      end

      self
    end

    def closed?
      @closed
    end
  end
end
