# $Id: ssi.rb,v 1.27.2.2 2003/04/11 14:09:10 toki Exp $
# -
# Server Side Inclusion

=begin
Copyright (c) 2001-2002
	TOKI Yoshinori <toki@freedom.ne.jp>. All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  1.  Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.

  2.  Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation
      and/or other materials provided with the distribution.

  3.  The name of the author may not be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
OF SUCH DAMAGE.
=end

require 'time'
require 'wwwsrv/const'
require 'wwwsrv/ioutil'
require 'wwwsrv/process'
require 'wwwsrv/document'
require 'wwwsrv/cgi'

module WWWsrv
  class SSIdocument < DocumentModifier
    include Timeout
    include IOunit
    include ProtectProcess

    CMD_TIMEOUT = 10

    def initialize(document, commands=nil)
      super(document)
      @enable = {}
      for cmd in (commands || %w[ config include echo fsize flastmod exec ])
	@enable[cmd] = true
      end
      @command_timeout = CMD_TIMEOUT
    end

    attr_accessor :command_timeout

    def protect_ssi_environ(local_path)
      if (local_path) then
	PROCESS_LOCK.synchronize do
	  protect_chdir(File::dirname(local_path)) do
	    return yield
	  end
	end
      else
	return yield
      end
    end
    private :protect_ssi_environ

    def virtual_request(srv_opt, method, path, request)
      if (%r"^/!" !~ path) then
	path = request.normalized_path.sub(%r"/.*?$", '/') + path
      end
      srv_opt.doc_log_debug(self, "virtual request: #{path}")
      virtual_request = Request::new(method, path, request.version, Header::new)

      %w[
	Host
	User-Agent
	Authorization
	Accept
	Accept-Charset
	Accept-Encoding
	Accept-Language
	From
	Pragma
	Referer
	Via
      ].each do |field_name|
	if (field_value = request.header[field_name]) then
	  virtual_request.header[field_name] = field_value
	end
      end

      virtual_request.body = EmptyContentReader::instance
      virtual_request.host = request.host
      virtual_request.client = request.client

      virtual_response = WWWsrv::Response::new
      virtual_response.header.server = self.class.name.gsub(/::/, '-') + '/' + VERSION

      return virtual_request, virtual_response
    end
    private :virtual_request

    def ssi_config(srv_opt, tag, value, ssi_conf, ssi_var, cgi_var, request, response)
      srv_opt.doc_log_debug(self, "config: #{tag}=#{value}")
      ssi_conf[tag] = value
      nil
    end
    private :ssi_config

    def ssi_include(srv_opt, tag, value, ssi_conf, ssi_var, cgi_var, request, response)
      srv_opt.doc_log_debug(self, "include #{tag}: #{value}")
      case (tag)
      when 'file'
	protect_ssi_environ(response[:local_path]) do
	  srv_opt.doc_log_debug(self, "local path: #{File::expand_path(value)}")
	  File::open(value) do |input|
	    input.binmode
	    return input.read
	  end
	end
      when 'virtual'
	message_body = ''
	doc_response = nil
	virtual_request, virtual_response = virtual_request(srv_opt, HTTP_GET, value, request)
	doc_status, err_args = Document::catch_status{
	  srv_opt.root.respond(srv_opt, '', virtual_request, virtual_response) do |arg, opts|
	    unless (doc_response) then
	      doc_response = arg
	    else
	      message_body.concat(arg)
	    end
	  end
	}

	case (doc_status)
	when :success
	  return message_body
	when :error
	  return ssi_conf['errmsg']
	else
	  raise "invalid internal status: #{doc_status}"
	end
      else
	srv_opt.doc_log_warning(self, "include: invalid tag: #{tag}")
	return ssi_conf['errmsg']
      end
    end
    private :ssi_include

    def ssi_echo(srv_opt, tag, value, ssi_conf, ssi_var, cgi_var, request, response)
      srv_opt.doc_log_debug(self, "echo #{tag}: #{value}")
      case (tag)
      when 'var'
	case (value)
	when 'PATH'
	  PROCESS_LOCK.synchronize do
	    srv_opt.doc_log_debug(self, "echo system variable: PATH: #{ENV['PATH']}")
	    return ENV['PATH']
	  end
	when 'DOCUMENT_NAME', 'DOCUMENT_URI', 'QUERY_STRING_UNESCAPED'
	  srv_opt.doc_log_debug(self, "echo SSI variable: #{value}: #{ssi_var[value]}")
	  return ssi_var[value]
	when 'DATE_LOCAL', 'DATE_GMT', 'LAST_MODIFIED'
	  if (timefmt = ssi_conf['timefmt']) then
	    time_info = ssi_var[value].strftime(timefmt)
	  else
	    time_info = ssi_var[value].to_s
	  end
	  srv_opt.doc_log_debug(self, "echo SSI time variable: #{value}: #{time_info}")
	  return time_info
	else
	  srv_opt.doc_log_debug(self, "echo CGI variable: #{value}: #{cgi_var[value]}")
	  return cgi_var[value]
	end
      else
	srv_opt.doc_log_warning(self, "echo: invalid tag: #{tag}")
	return ssi_conf['errmsg']
      end
    end
    private :ssi_echo

    def ssi_fsize(srv_opt, tag, value, ssi_conf, ssi_var, cgi_var, request, response)
      srv_opt.doc_log_debug(self, "fsize #{tag}: #{value}")
      fsize_format = proc{ |fsize|
	case (ssi_conf['sizefmt'])
	when 'bytes'
	  fsize_info = fsize.to_s
	  srv_opt.doc_log_debug(self, "fsize #{tag}: #{value}: bytes: #{fsize_info}")
	when 'abbrev'
	  case (fsize)
	  when 0...1024
	    fsize_info = fsize.to_s
	    srv_opt.doc_log_debug(self, "fsize #{tag}: #{value}: abbrev: Bytes: #{fsize_info}")
	  when 1024...(1024**2)
	    fsize_info = format('%f', (fsize.to_f / 1024).round)
	    srv_opt.doc_log_debug(self, "fsize #{tag}: #{value}: abbrev: KBytes: #{fsize_info}")
	  else
	    fsize_info = format('%f', (fsize.to_f / 1024**2).round)
	    srv_opt.doc_log_debug(self, "fsize #{tag}: #{value}: abbrev: MBytes: #{fsize_info}")
	  end
	else
	  fsize_info = ssi_conf['errmsg']
	  srv_opt.doc_log_warning(self, "fsize: invalid sizefmt value: #{sizefmt}")
	end

	fsize_info
      }

      case (tag)
      when 'file'
	stat = protect_ssi_environ(response[:local_path]) {
	  File::stat(value)
	}
	return fsize_format.call(stat.size)
      when 'virtual'
	doc_response = nil
	virtual_request, virtual_response = virtual_request(srv_opt, HTTP_HEAD, value, request)
	doc_status, err_args = Document::catch_status{
	  srv_opt.root.respond(srv_opt, '', virtual_request, virtual_response) do |arg, opts|
	    unless (doc_response) then
	      doc_response = arg
	    end
	  end
	}

	if (doc_status == :success) then
	  if (doc_response.header.content_length) then
	    content_length = doc_response.header.content_length.to_i
	  else
	    content_length = 0
	    doc_response = nil
	    virtual_request, virtual_response = virtual_request(srv_opt, HTTP_GET, value, request)
	    doc_status, err_args = Document::catch_status{
	      srv_opt.root.respond(srv_opt, '', virtual_request, virtual_response) do |arg, opts|
		unless (doc_response) then
		  doc_response = arg
		else
		  content_length += arg.length
		end
	      end
	    }
	  end
	end

	case (doc_status)
	when :success
	  return fsize_format.call(content_length)
	when :error
	  return ssi_conf['errmsg']
	else
	  raise "invalid internal status: #{doc_status}"
	end
      else
	srv_opt.doc_log_warning(self, "fsize: invalid tag: #{tag}")
	return ssi_conf['errmsg']
      end
    end
    private :ssi_fsize

    def ssi_flastmod(srv_opt, tag, value, ssi_conf, ssi_var, cgi_var, request, response)
      srv_opt.doc_log_debug(self, "flastmod #{tag}: #{value}")
      flastmod_format = proc{ |mtime|
	if (timefmt = ssi_conf['timefmt']) then
	  mtime_info = mtime.strftime(timefmt)
	else
	  mtime_info = mtime.to_s
	end
	srv_opt.doc_log_debug(self, "flastmod #{tag}: #{value} #{mtime_info}")

	mtime_info
      }

      case (tag)
      when 'file'
	stat = protect_ssi_environ(response[:local_path]) {
	  File::stat(value)
	}
	mtime = stat.mtime.gmtime
	return flastmod_format.call(mtime)
      when 'virtual'
	doc_response = nil
	virtual_request, virtual_response = virtual_request(srv_opt, HTTP_HEAD, value, request)
	doc_status, err_args = Document::catch_status{
	  srv_opt.root.respond(srv_opt, '', virtual_request, virtual_response) do |arg, opts|
	    unless (doc_response) then
	      doc_response = arg
	    end
	  end
	}

	case (doc_status)
	when :success
	  if (doc_response.header.last_modified) then
	    mtime = Time::parse(doc_response.header.last_modified)
	    return flastmod_format.call(mtime)
	  else
	    return ssi_conf['errmsg']
	  end
	when :error
	  return ssi_conf['errmsg']
	else
	  raise "invalid internal status: #{doc_status}"
	end
      else
	srv_opt.doc_log_warning(self, "flastmod: invalid tag: #{tag}")
	return ssi_conf['errmsg']
      end
    end
    private :ssi_flastmod

    def ssi_exec(srv_opt, tag, value, ssi_conf, ssi_var, cgi_var, request, response)
      srv_opt.doc_log_debug(self, "exec #{tag}: #{value}")
      case (tag)
      when 'cmd'
	env_var = cgi_var.to_a + ssi_var.map{ |var_name, var_value|
	  case (var_name)
	  when 'DATE_LOCAL', 'DATE_GMT', 'LAST_MODIFIED'
	    if (timefmt = ssi_conf['timefmt']) then
	      [ var_name, var_value.strftime(timefmt) ]
	    else
	      [ var_name, var_value.to_s ]
	    end
	  else
	    [ var_name, var_value ]
	  end
	}

	cmd_io = nil
	PROCESS_LOCK.synchronize do
	  protect_environ(env_var) do
	    if (response.has_key? :local_path) then
	      protect_chdir(File::dirname(response[:local_path])) do
		cmd_io = IO::popen(value)
	      end
	    else
	      cmd_io = IO::popen(value)
	    end
	  end
	end

	begin
	  cmd_result = ''
	  cmd_output_fragment = nil
	  block_size = io_unit(cmd_io.stat)
	  while (true)
	    cmd_output_fragment = wait(@command_timeout) { cmd_io.read(block_size) }
	    cmd_output_fragment or break
	    cmd_result.concat(cmd_output_fragment)
	  end
	ensure
	  cmd_io.close
	end

	return cmd_result
      when 'cgi'
	message_body = ''
	doc_response = nil
	virtual_request, virtual_response = virtual_request(srv_opt, HTTP_GET, value, request)
	doc_status, err_args = Document::catch_status{
	  srv_opt.root.respond(srv_opt, '', virtual_request, virtual_response) do |arg, opts|
	    unless (doc_response) then
	      doc_response = arg
	    else
	      message_body.concat(arg)
	    end
	  end
	}

	case (doc_status)
	when :success
	  return message_body
	when :error
	  return ssi_conf['errmsg']
	else
	  raise "invalid internal status: #{doc_status}"
	end
      else
	srv_opt.doc_log_warning(self, "exec: invalid tag: #{tag}")
	return ssi_conf['errmsg']
      end
    end
    private :ssi_exec

    def modify_request(srv_opt, context, prefix, request, response)
      request.erase_cond_header_fields
      context[:request] = request

      nil
    end

    def modify_response(srv_opt, context, prefix, request, response)
      case (response.status)
      when 200, 201, 203
	context[:state] = :filter
	context[:response] = response
	context[:message_body] = ''
      else
	context[:state] = :through
      end

      nil
    end

    def modify_message_body(srv_opt, context, body_fragment, opts=nil)
      case (context[:state])
      when :filter
	context[:message_body].concat(body_fragment)
      when :through
	yield(body_fragment, opts)
      else
	raise "unknown internal state: #{context[:state].inspect}"
      end

      nil
    end

    def modify_terminal(srv_opt, context)
      case (context[:state])
      when :filter
	# continue
      when :through
	return
      else
	raise "unknown internal state: #{context[:state].inspect}"
      end

      request = context[:request]
      response = context[:response]
      message_body = context[:message_body]

      ssi_conf = {
	'errmsg' => '[SSI error]',
	'timefmt' => nil,
	'sizefmt' => 'bytes'
      }
      ssi_var = {
	'DOCUMENT_NAME' => response[:local_path],
	'DOCUMENT_URI' => request.normalized_path
      }
      ssi_var['QUERY_STRING_UNESCAPED'] = decodeURL(request.query) if request.query
      ssi_var['DATE_LOCAL'] = Time::now
      ssi_var['DATE_GMT'] = ssi_var['DATE_LOCAL'].dup.gmtime
      ssi_var['LAST_MODIFIED'] =
	File::stat(response[:local_path]).mtime.gmtime if (response.has_key? :local_path)
      cgi_var = request.cgi_var

      message_body.gsub!(/<!--\s*#(\S+)\s+(\S+)\s*=\s*"([^"]*)"\s*-->/) { |src|
	command, tag, value = $1.downcase, $2.downcase, unquoteHTML($3)
	srv_opt.doc_log_debug(self, "SSI command: #{command} #{tag}=\"#{value}\"")
	if (@enable.include? command) then
	  begin
	    case (command)
	    when 'config'
	      fragment = ssi_config(srv_opt, tag, value,
				    ssi_conf, ssi_var, cgi_var,
				    request, response)
	    when 'include'
	      fragment = ssi_include(srv_opt, tag, value,
				     ssi_conf, ssi_var, cgi_var,
				     request, response)
	    when 'echo'
	      fragment = ssi_echo(srv_opt, tag, value,
				  ssi_conf, ssi_var, cgi_var,
				  request, response)
	    when 'fsize'
	      fragment = ssi_fsize(srv_opt, tag, value,
				   ssi_conf, ssi_var, cgi_var,
				   request, response)
	    when 'flastmod'
	      fragment = ssi_flastmod(srv_opt, tag, value,
				      ssi_conf, ssi_var, cgi_var,
				      request, response)
	    when 'exec'
	      fragment = ssi_exec(srv_opt, tag, value,
				  ssi_conf, ssi_var, cgi_var,
				  request, response)
	    else
	      srv_opt.doc_log_warning(self, "unknown SSI command: #{command}")
	      fragment = ssi_conf['errmsg']
	    end
	  rescue
	    srv_opt.doc_log_alert(self, "SSI error: #{$!}")
	    fragment = ssi_conf['errmsg']
	  end

	  fragment
	else
	  ssi_conf['errmsg']
	end
      }

      response.erase_sticky_entity_header_fields
      response.header.content_length = message_body.length.to_s

      yield(response)
      yield(message_body)

      nil
    end
  end

  # alias
  SSIDocument = SSIdocument
end
